function c = gmmake_cyl(subdiv)
% gmmake_cyl: Make a bezier approximation to a cylinder
%  c = gmmake_cyl(subdiv)
% Makes a cubic bezier patch approximation to a cylinder
% of radius 1 and height 2, centered at the origin and
% with vertical direction aligned with the z axis.
% The argument is the number of rectangular patches to use around
% the side of the cylinder.

if subdiv < 3
  error('Number of subdivisions must be at least 3')
end
[vertex,code] = gm_circ_approx(0, 2*pi, subdiv);
vertex = vertex(1:3*subdiv,:);

control_points = zeros(3,12*subdiv+2);

% top and bottom layers of control points on edges

control_points(:,1:2:6*subdiv) = [vertex';-ones(1,3*subdiv)];
control_points(:,2:2:6*subdiv) = [vertex';ones(1,3*subdiv)];

global GM_BREP_TYPE_CODE;
c = zba({GM_BREP_TYPE_CODE; 3; 3; {}; []; {}; cell(5,2); cell(5,3); cell(5,1)});



% Make side surface.

c{7}{0,0} = 'side_face';
c{7}{1,0} = {};
c{7}{2,0} = {'bottom_edge', 'top_edge'};
c{7}{3,0} = {};
c{7}{4,0} = cell(3,subdiv);
for k = 0 : subdiv - 1
  c{7}{4,0}{0,k} = 'bezier_quad';
  c{7}{4,0}{1,k} = [3,1]; 
  kp1 = rem(k + 1, subdiv);
  c{7}{4,0}{2,k} = [k*6,k*6+2,k*6+4, kp1*6, k*6+1, k*6+3, k*6+5, kp1*6+1];
end

% do the top and bottom


for s = 0 : 1
  % generate the topological edge
  if s == 0
    c{6}{0,s} = 'bottom_edge';
  else
    c{6}{0,s} = 'top_edge';
  end
  c{6}{1,s} = {};
  c{6}{2,s} = {};
  c{6}{3,s} = {};
  c{6}{4,s} = cell(3,subdiv);
  for k = 0 : subdiv - 1
    c{6}{4,s}{0,k} = 'bezier_curve';
    c{6}{4,s}{1,k} = 3;
    kp1 = rem(k + 1, subdiv);
    c{6}{4,s}{2,k} = [k*6+s, k*6+2+s, k*6+4+s, kp1*6+s];
  end
  cpbase1 = 6*subdiv + (3 * subdiv + 1) * s;
  control_points(:,cpbase1 + 1 : cpbase1 + subdiv) = ...
        [.6*cos(2*pi*(0:(subdiv - 1))/subdiv); ...
         .6*sin(2*pi*(0:(subdiv - 1))/subdiv); ...
         (2 * s - 1) * ones(1,subdiv)];
  cpbase2 = 7 * subdiv + (3 * subdiv + 1) * s;
  control_points(:,cpbase2 + 1: cpbase2 + subdiv) = ...
        [.6*cos(2*pi*(0.5:(subdiv - 0.5))/subdiv); ...
         .6*sin(2*pi*(0.5:(subdiv - 0.5))/subdiv); ...
         (2 * s - 1) * ones(1,subdiv)];
  cpbase3 = 8 * subdiv + (3 * subdiv + 1) * s;
  control_points(:,cpbase3 + 1 : cpbase3 + subdiv) = ...
        [.3*cos(2*pi*(0:(subdiv - 1))/subdiv); ...
         .3*sin(2*pi*(0:(subdiv - 1))/subdiv); ...
         (2 * s - 1) * ones(1,subdiv)];
  centercp = 9 * subdiv + (3 * subdiv + 1) * s;
  control_points(:,centercp + 1) = [0;0;2*s-1];

  % Generate the triangular patches making up this face
  % (either top or bottom).

  if s == 0
    c{7}{0,s+1} = 'bottom_face';
    c{7}{2,s+1} = {'bottom_edge'};
  else
    c{7}{0,s+1} = 'top_face';
    c{7}{2,s+1} = {'top_edge'};
  end
  c{7}{1,s+1} = {};
  c{7}{3,s+1} = {};
  c{7}{4,s+1} = cell(3,subdiv);
  for k = 0 : subdiv - 1
    c{7}{4,s+1}{0,k} = 'bezier_triangle';
    c{7}{4,s+1}{1,k} = [3];
    kp1 = rem(k + 1, subdiv);
    c{7}{4,s+1}{2,k} = [centercp,cpbase3+k,cpbase3+kp1,cpbase1+k, ...
            cpbase2+k,cpbase1+kp1, ...
            k*6+s, k*6+2+s, k*6+4+s, kp1*6+s];
  end
end

c{8}{0,0} = 'cylinder';
c{8}{1,0} = {};
c{8}{2,0} = {'side_face','bottom_face','top_face'};
c{8}{3,0} = {};
c{8}{4,0} = {};
c{4} = control_points;



% ------------------------------------------------------------------
% Copyright (c) 1999 by Cornell University.  All rights reserved
% See the accompanying file 'Copyright' for authorship information,
% the terms of the license governing this software, and disclaimers
% concerning this software.
% ------------------------------------------------------------------
% This file is part of the QMG software.  
% Version 2.0 of QMG, release date September 3, 1999
% ------------------------------------------------------------------
