// ------------------------------------------------------------------
// qsimpcomp.cpp
//
// This file contains member functions the class SimpComp and SimpComp_
// Under_Construction, which are for simplicial complexes (meshes).
// This version of qsimpcomp.cpp is for the Matlab front end.
// ------------------------------------------------------------------
// Author: Stephen A. Vavasis
// Copyright (c) 1999 by Cornell University.  All rights reserved.
// 
// See the accompanying file 'Copyright' for authorship information,
// the terms of the license governing this software, and disclaimers
// concerning this software.
// ------------------------------------------------------------------
// This file is part of the QMG software.  
// Version 2.0 of QMG, release date September 3, 1999.
// ------------------------------------------------------------------


#include "qsimpcomp.h"
#include "string.h"


namespace {
  void pause(const char* s) {
    mexPrintf("%s\nPress any key to continue\n", s);
    mexEvalString("pause");
  }
  void dump610(const mxArray* a) {
    mexPrintf("Dumping m{6}{1,0}\n");
    mxArray* a0 = mxGetCell(a,6);
    mxArray* a1 = mxGetCell(a0,1);
    int m = mxGetM(a1);
    int n = mxGetN(a1);
    mexPrintf("size = %d-by-%d\n", m,n);
    for (int i = 0; i < m; ++i) {
      for (int j = 0; j < n; ++j)
        mexPrintf("%10.5f ", mxGetPr(a1)[i + j * m]);
      mexPrintf("\n");
    }
  }
}



QMG::SimpComplex::VertexOrdinalIndex 
QMG::SimpComplex::process_ordinal_vnum_(VertexOrdinalIndex vnumo) const {
  if (vnumo < 0 || vnumo >= mxGetN(mxGetCell(c_entries_, 4))) {
    mexPrintf("%d %d\n", vnumo, mxGetN(mxGetCell(c_entries_, 4)));
    throw_error("ordinal vnum out of range");
  }
  return vnumo;
}

QMG::SimpComplex::SimpComplex(const SimpComplex& sc) :
  mesh_(sc.mesh_),
  c_embedded_dim_(sc.c_embedded_dim_),
  c_gdim_(sc.c_gdim_),
  c_entries_(sc.c_entries_),
  c_nodearray_(sc.c_nodearray_),
  global_to_ordinal_(sc.global_to_ordinal_)
{
 }


int 
QMG::SimpComplex::num_nodes() const {
  return mxGetN(mxGetCell(c_entries_, 4));
}


int
QMG::SimpComplex::num_elements() const {
  const mxArray* toplev = mxGetCell(c_entries_, c_gdim_ + 5);
  int numface = mxGetN(toplev);
  int nume = 0;
  for (Brep::FaceIndex faceind = 0; faceind < numface; ++faceind)
    nume += mxGetN(mxGetCell(toplev, 2 * faceind + 1));
  return nume;
}

int
QMG::SimpComplex::brep_levelsize(int fdim) const {
#ifdef RANGECHECK
  if (fdim < 0 || fdim > c_gdim_)
    throw_error("fdim out of range");
#endif
  return mxGetN(mxGetCell(c_entries_, fdim + 5));
}

QMG::string
QMG::SimpComplex::lookup_prop_val(const string& prop) const {
  const mxArray* pvlist = mxGetCell(c_entries_, 3);
  int npv = mxGetN(pvlist);
  for (int k = 0; k < npv; ++k) {
    const mxArray* propa = mxGetCell(pvlist, 2 * k);
    if (propa == 0 || !mxIsChar(propa) || mxGetM(propa) != 1
      || mxIsSparse(propa))
      throw_error("Illegal data type stored in property slot");
    int sz = mxGetN(propa);
    if (sz != prop.length()) continue;
    char* buf = new char[sz + 2];
    mxGetString(propa, buf, sz + 1);
    string props(buf);
    delete[] buf;
    if (compare_nocase(props, prop) == 0) {
      const mxArray* vala = mxGetCell(pvlist, 2 * k + 1);
      if (vala == 0 || !mxIsChar(vala) || mxGetM(vala) != 1
          || mxIsSparse(vala))
        throw_error("Illegal data type stored in value slot");
      sz = mxGetN(vala);
      buf = new char[sz + 2];
      mxGetString(vala, buf, sz + 1);
      string rval(buf);
      delete[] buf;
      return rval;
    }
  }
  return string("");
}

int
QMG::SimpComplex::num_prop_val() const {
  return mxGetN(mxGetCell(c_entries_, 3));
}

QMG::string 
QMG::SimpComplex::prop(int i) const {
  const mxArray* pvlist = mxGetCell(c_entries_, 3);
  int npv = mxGetN(pvlist);
#ifdef RANGECHECK
  if (i < 0 || i >= npv)
    throw_error("Index out of range in prop");
#endif
  const mxArray* propa = mxGetCell(pvlist, 2 * i);
  if (propa == 0 || !mxIsChar(propa) || mxGetM(propa) != 1
      || mxIsSparse(propa))
    throw_error("Illegal data type stored in property slot");
  int sz = mxGetN(propa);
  char* buf = new char[sz + 2];
  mxGetString(propa, buf, sz + 1);
  string prop_s(buf);
  delete[] buf;
  return prop_s;
}

QMG::string 
QMG::SimpComplex::val(int i) const {
  const mxArray* pvlist = mxGetCell(c_entries_, 3);
  int npv = mxGetN(pvlist);
#ifdef RANGECHECK
  if (i < 0 || i >= npv)
    throw_error("Index out of range in val");
#endif
  const mxArray* vala = mxGetCell(pvlist, 2 * i + 1);
  if (vala == 0 || !mxIsChar(vala) || mxGetM(vala) != 1
      || mxIsSparse(vala))
    throw_error("Illegal data type stored in value slot");
  int sz = mxGetN(vala);
  char* buf = new char[sz + 2];
  mxGetString(vala, buf, sz + 1);
  string val_s(buf);
  delete[] buf;
  return val_s;
}

QMG::SimpComplex::VertexGlobalIndex 
QMG::SimpComplex::ordinal_to_global(VertexOrdinalIndex vnumo) const {
  // mexPrintf("Reached ordinal to global vnumo = %d\n", vnumo);
  double vnumr = c_nodearray_[process_ordinal_vnum_(vnumo) *
                             (c_embedded_dim_ + 1)];
  VertexGlobalIndex vnum = static_cast<int>(vnumr);
  if (vnumr != static_cast<Real>(vnum))
    throw_error("Noninteger data stored as vertex global index");
  // mexPrintf("return\n");
  return vnum;
}

QMG::Real 
QMG::SimpComplex::real_coord_o(VertexOrdinalIndex vnumo, int jj) const {
#ifdef RANGECHECK
  if (jj < 0 || jj >= c_embedded_dim_) 
    throw_error("Range err in real_coord_o");
#endif
  // mexPrintf("reached real_coord_o vnumo = %d\n", vnumo);
  VertexOrdinalIndex vnumo1 = process_ordinal_vnum_(vnumo);
  // mexPrintf("returning\n");
  return c_nodearray_[vnumo1 *
                     (c_embedded_dim_ + 1) + jj + 1];
}

int 
QMG::SimpComplex::num_node_on_face(const Brep::Face_Spec& fspec) const {
  int fdim = fspec.fdim();
  int faceind = fspec.faceind();
#ifdef RANGECHECK
  if (fdim < 0 || fdim > c_gdim_)
    throw_error("fdim out of range");
#endif
  const mxArray* levarray = mxGetCell(c_entries_, 5+fdim);
#ifdef RANGECHECK
  if (faceind < 0 || faceind >= mxGetN(levarray))
    throw_error("faceind out of range");
#endif
  return mxGetN(mxGetCell(levarray, 2 * faceind));
}

QMG::SimpComplex::VertexGlobalIndex
QMG::SimpComplex::node_on_face(const Brep::Face_Spec& fspec,
                               int seqno) const {
  int fdim = fspec.fdim();
  int faceind = fspec.faceind();
#ifdef RANGECHECK
  if (fdim < 0 || fdim > c_gdim_)
    throw_error("fdim out of range");
#endif
  const mxArray* levarray = mxGetCell(c_entries_, 5+fdim);
#ifdef RANGECHECK
  if (faceind < 0 || faceind >= mxGetN(levarray))
    throw_error("faceind out of range");
#endif
  const mxArray* nodear = mxGetCell(levarray, 2 * faceind);
#ifdef RANGECHECK
  if (seqno < 0 || seqno >= mxGetN(nodear))
    throw_error("seqno out of range");
#endif
  double vnumr = mxGetPr(nodear)[seqno * mxGetM(nodear)];
  VertexGlobalIndex vnum = static_cast<int>(vnumr);
  if (static_cast<Real>(vnum) != vnumr)
    throw_error("Nonintegral data stored in node number slot");
  return vnum;
}

QMG::Brep::PatchIndex
QMG::SimpComplex::patchind_on_face(const Brep::Face_Spec& fspec,
                                   int seqno) const {
  int fdim = fspec.fdim();
  if (fdim == 0)
    return 0;
  int faceind = fspec.faceind();
#ifdef RANGECHECK
  if (fdim < 0 || fdim > c_gdim_)
    throw_error("fdim out of range");
#endif
  const mxArray* levarray = mxGetCell(c_entries_, 5+fdim);
#ifdef RANGECHECK
  if (faceind < 0 || faceind >= mxGetN(levarray))
    throw_error("faceind out of range");
#endif
  const mxArray* nodear = mxGetCell(levarray, 2 * faceind);
#ifdef RANGECHECK
  if (seqno < 0 || seqno >= mxGetN(nodear))
    throw_error("seqno out of range");
#endif
  double patchindr = mxGetPr(nodear)[seqno * mxGetM(nodear) + 1];
  Brep::PatchIndex patchind = static_cast<int>(patchindr);
  if (static_cast<Real>(patchind) != patchindr)
    throw_error("Nonintegral data stored in patch number slot");
  return patchind;
}

QMG::Real 
QMG::SimpComplex::param_coord_on_face(const Brep::Face_Spec& fspec, 
                                      int seqno, 
                                      int jj) const {
  int fdim = fspec.fdim();
  if (fdim == 0)
    return 0;
  int faceind = fspec.faceind();
#ifdef RANGECHECK
  if (fdim < 0 || fdim > c_gdim_)
    throw_error("fdim out of range");
  if (jj < 0 || jj >= fdim)
    throw_error("jj out of range");
#endif
  const mxArray* levarray = mxGetCell(c_entries_, 5+fdim);
#ifdef RANGECHECK
  if (faceind < 0 || faceind >= mxGetN(levarray))
    throw_error("faceind out of range");
#endif
  const mxArray* nodear = mxGetCell(levarray, 2 * faceind);
#ifdef RANGECHECK
  if (seqno < 0 || seqno >= mxGetN(nodear))
    throw_error("seqno out of range");
#endif
  return mxGetPr(nodear)[seqno * mxGetM(nodear) + jj + 2];
}


int 
QMG::SimpComplex::num_meshface_on_face(const Brep::Face_Spec& fspec) const {
  int fdim = fspec.fdim();
  int faceind = fspec.faceind();
#ifdef RANGECHECK
  if (fdim < 0 || fdim > c_gdim_)
    throw_error("fdim out of range");
#endif
  const mxArray* levarray = mxGetCell(c_entries_, 5+fdim);
#ifdef RANGECHECK
  if (faceind < 0 || faceind >= mxGetN(levarray))
    throw_error("faceind out of range");
#endif
  if (fdim == 0)
    return mxGetN(mxGetCell(levarray, 2 * faceind));
  return mxGetN(mxGetCell(levarray, 2 * faceind + 1));
}

QMG::SimpComplex::VertexGlobalIndex 
QMG::SimpComplex::node_of_meshface_on_face(const Brep::Face_Spec& fspec, 
                                           int seqno, 
                                           int jj) const {
  int fdim = fspec.fdim();
  int faceind = fspec.faceind();
#ifdef RANGECHECK
  if (fdim < 0 || fdim > c_gdim_)
    throw_error("fdim out of range");
  if (jj < 0 || jj > fdim)
    throw_error("jj  out of range");
#endif
  const mxArray* levarray = mxGetCell(c_entries_, 5+fdim);
#ifdef RANGECHECK
  if (faceind < 0 || faceind >= mxGetN(levarray))
    throw_error("faceind out of range");
#endif
  const mxArray* simparray = (fdim == 0)?
    mxGetCell(levarray, 2 * faceind) :
    mxGetCell(levarray, 2 * faceind + 1);
    
#ifdef RANGECHECK
  if (seqno < 0 || seqno >= mxGetN(simparray))
    throw_error("Seqno out of range");
#endif
  double vnumr = mxGetPr(simparray)[seqno * (fdim + 1) + jj];
  VertexGlobalIndex vnum = static_cast<int>(vnumr);
  if (static_cast<Real>(vnum) != vnumr) {
    mexPrintf("%d %e\n", vnum, vnumr);
    throw_error("Nonintegral data stored in vnum slot");
  }
  return vnum;
}


QMG::SimpComplex_From_FrontEnd::
SimpComplex_From_FrontEnd(const mxArray* mesh) {
  mesh_ = mesh;
  if (strcmp(mxGetClassName(mesh), "zba") != 0)
    throw_error("Attempt to use non-zba array as simplicial complex");
  c_entries_ = mxGetFieldByNumber(mesh,0,0);
  if (c_entries_ == 0)
    throw_error("Input array has no field 0 (not a simplicial complex)");
  if (!mxIsCell(c_entries_) || mxGetM(c_entries_) < 6 || 
      mxGetN(c_entries_) != 1)
    throw_error("Input array should be columnvec cell zba with at least 6 entries for simplicial complex");
  
  const mxArray* typecode = mxGetCell(c_entries_, 0);
  if (typecode == 0)
    throw_error("Input array has no cell 0 (not a simplicial complex)");
  int lth1 = mxGetM(typecode) * mxGetN(typecode);
  if (lth1 != strlen(io_header_code()))
    throw_error("Typecode has wrong size -- object not a simplicial complex");
  char buf[100];
  if (mxGetString(typecode, buf, 100) ||
      compare_nocase(string(buf), string(io_header_code())) != 0)
    throw_error("Incorrect typecode -- object not a simplicial complex");
  const mxArray* gdima = mxGetCell(c_entries_, 1);
  if (gdima == 0 || !mxIsDouble(gdima) || mxIsSparse(gdima) || mxGetPi(gdima)
      || mxGetM(gdima) * mxGetN(gdima) != 1)
    throw_error("Missing or ill-formatted intrinsic dim field");
  Real gdimr = mxGetScalar(gdima);
  c_gdim_ = static_cast<int>(gdimr);
  if (static_cast<Real>(c_gdim_) != gdimr)
    throw_error("Nonintegral data for intrinsic dim");
  const mxArray* edima = mxGetCell(c_entries_, 2);
  if (edima == 0 || !mxIsDouble(edima) || mxIsSparse(gdima) || mxGetPi(gdima)
      || mxGetM(edima) * mxGetN(edima) != 1)
    throw_error("Missing or ill formatted embedded dim field");
  Real edimr = mxGetScalar(edima);
  c_embedded_dim_ = static_cast<int>(edimr);
  if (static_cast<Real>(c_embedded_dim_) != edimr)
    throw_error("Nonintegral data for embedded dim");
  if (c_embedded_dim_ != 2 && c_embedded_dim_ != 3)
    throw_error("Embedded dim must be 2 or 3");
  if (c_gdim_ < 1 || c_gdim_ > c_embedded_dim_)
    throw_error("Intrinsic dimension out of range");
  if (mxGetM(c_entries_) != 6 + c_gdim_) {
    mexPrintf(" c_gdim_ = %d mxGetM(c_entries_) = %d\n",
              c_gdim_, mxGetM(c_entries_));
    throw_error("Mesh cell array has wrong number of entries");
  }
  const mxArray* vertices = mxGetCell(c_entries_, 4);
  if (vertices == 0 || !mxIsDouble(vertices) || mxIsSparse(vertices)
      || mxGetPi(vertices))
    throw_error("Missing or illformatted vertices");
  c_nodearray_ = mxGetPr(vertices);
  if (mxGetN(vertices) > 0 && mxGetM(vertices) != 1 + c_embedded_dim_)
    throw_error("Node array has incorrect number of rows");
  if (mxGetN(vertices) > 0 && c_nodearray_ == 0)
    throw_error("Unable to retrieve node array");
  const mxArray* pvlist = mxGetCell(c_entries_, 3);
  if (pvlist == 0 || !mxIsCell(pvlist) || 
      (mxGetN(pvlist) > 0 && mxGetM(pvlist) != 2))
    throw_error("Property-value list must be a 2-by-k cell array");
  for (int facedim = 0; facedim <= c_gdim_; ++facedim) {
    const mxArray* levarray = mxGetCell(c_entries_, 5 + facedim);
    if (levarray == 0 || !mxIsCell(levarray))
      throw_error("Level array of mesh must be cell array");
    int levsize = mxGetN(levarray);
    if (levsize > 0 && mxGetM(levarray) != 2)
      throw_error("Level array of mesh must be 2-by-k cell array");
    for (Brep::FaceIndex faceind = 0; faceind < levsize; ++faceind) {
      const mxArray* vtxlist = mxGetCell(levarray, 2 * faceind);
      if (vtxlist == 0 || !mxIsDouble(vtxlist) || mxIsSparse(vtxlist))
        throw_error("Vertex list of brep face must be a matrix");
      int numv = mxGetN(vtxlist);
      if (numv > 0 && mxGetM(vtxlist) != 
          1 + ((facedim == 0)? 0 : 1) + facedim)
        throw_error("Vertex list of brep face has wrong number of rows");
      const mxArray* smplist = mxGetCell(levarray, 2 * faceind + 1);
      if (smplist == 0 || !mxIsDouble(smplist) || mxIsSparse(smplist))
        throw_error("Simplex list of brep face must be a matrix");
      int nums = mxGetN(smplist);
      if (nums > 0 && mxGetM(smplist) != facedim + 1)
        throw_error("Vertex list of brep face has wrong number of rows");
    }
  }
  global_to_ordinal_ = 0;
}

void
QMG::SimpComplex_From_FrontEnd::make_global_to_ordinal_() const {
  if (global_to_ordinal_) 
    throw_error("Make global to ordinal called twice");
  global_to_ordinal_ = new GlobalToOrdinalMapType;
  int nn = mxGetN(mxGetCell(c_entries_, 4));
  for (VertexOrdinalIndex vnumo = 0;
       vnumo < nn;
       ++vnumo) {
    double vnumr = c_nodearray_[vnumo * (c_embedded_dim_ + 1)];
    VertexGlobalIndex vnum = static_cast<int>(vnumr);
    if (vnumr != static_cast<Real>(vnum))
      throw_error("Nonintegral data as vertex global index");
    if (vnum < 0)
      throw_error("global vertex id out of range");
    pair<GlobalToOrdinalMapType::iterator, bool> rval =
      global_to_ordinal_ -> 
      insert(pair<VertexGlobalIndex, VertexOrdinalIndex>(vnum,vnumo));
    if (!rval.second)
      throw_error("global vertex id occurs twice in input");
  }
}

QMG::SimpComplex_From_FrontEnd::~SimpComplex_From_FrontEnd() {
  delete global_to_ordinal_;
}


QMG::SimpComplex_From_FrontEnd::
SimpComplex_From_FrontEnd(SimpComplex_From_FrontEnd& other) :
  SimpComplex(other) {
  other.global_to_ordinal_ = 0;
}


QMG::SimpComplex_From_FrontEnd::
SimpComplex_From_FrontEnd(const SimpComplex_From_FrontEnd_Returnval& sucr) :
SimpComplex(sucr.sref_) {
  sucr.sref_.global_to_ordinal_ = 0;
}


bool 
QMG::SimpComplex_Under_Construction::Face_Bdry_Inc_::
operator<(const Face_Bdry_Inc_& other) const {
  if (meshv[0] < other.meshv[0]) return true;
  if (meshv[0] > other.meshv[0]) return false;
  if (meshv[1] < other.meshv[1]) return true;
  if (meshv[1] > other.meshv[1]) return false;
  return meshv[2] < other.meshv[2];
}

bool
QMG::SimpComplex_Under_Construction::Vertex_Bdry_Inc_::
operator<(const Vertex_Bdry_Inc_& other) const {
  if (vnum < other.vnum) return true;
  if (vnum > other.vnum) return false;
  return fspec < other.fspec;
}

void 
QMG::SimpComplex_Under_Construction::clobber_() {
  mesh_ = 0;
  global_to_ordinal_ = 0;
  vertinclookup_ = 0;  
  faceinclookup_ = 0;
  max_elems_per_face_ = 0;
  max_nodes_per_face_ = 0;
}

  
QMG::SimpComplex::VertexOrdinalIndex 
QMG::SimpComplex_Under_Construction::add_vertex(const vector<Real>& realcoord, 
                                                VertexGlobalIndex vnum) {
#ifdef RANGECHECK
  if (realcoord.size() != c_embedded_dim_)
    throw_error("Real coord wrong size in addvertex");
#endif

  mxArray* node_array = mxGetCell(c_entries_, 4); 
  int numnode = mxGetN(node_array);
  int p = numnode * (c_embedded_dim_ + 1);
  if (numnode == max_nodes_) {
    int newsize = numnode * 2 + 2;
    double* newpr = 
      static_cast<double*>(mxMalloc(newsize * (c_embedded_dim_ + 1) * sizeof(double)));
    for (int j = 0; j < p; ++j)
      newpr[j] = c_nodearray_[j];
    mxFree(const_cast<double*>(c_nodearray_));
    c_nodearray_ = newpr;
    mxSetPr(node_array, newpr);
    max_nodes_ = newsize;
  }
  double* nodearray1 = const_cast<double*>(c_nodearray_);
  nodearray1[p] = static_cast<double>(vnum);
  mxSetN(node_array, numnode + 1);
  for (int j1 = 0; j1 < c_embedded_dim_; ++j1)
    nodearray1[++p] = realcoord[j1];
  pair<GlobalToOrdinalMapType::iterator, bool> rval =
    global_to_ordinal_ -> 
    insert(pair<VertexGlobalIndex, VertexOrdinalIndex>(vnum,numnode));
  if (!rval.second)
    throw_error("global vertex id inserted twice");
  return numnode;
}


#ifndef BUG_IN_USING
using QMG::pair;
#endif

pair<QMG::SimpComplex::VertexOrdinalIndex, QMG::SimpComplex::VertexGlobalIndex> 
QMG::SimpComplex_Under_Construction::
add_vertex(const vector<Real>& realcoord) {
  VertexGlobalIndex thisid = next_vnum_;
  VertexOrdinalIndex vnum_o = add_vertex(realcoord, thisid);
  ++next_vnum_;
  return pair<VertexOrdinalIndex,VertexGlobalIndex>(vnum_o, thisid);
}


void 
QMG::SimpComplex_Under_Construction::add_fspec(const Brep::Face_Spec& fspec) {
  int fdim = fspec.fdim();
  if (fdim < 0 || fdim > c_gdim_)
    throw_error("Face dim out of range");
  const mxArray* lev1 = mxGetCell(c_entries_, 5 + fdim);
  mxArray* nclev1 = const_cast<mxArray*>(lev1);
  Brep::FaceIndex faceind = fspec.faceind();
  if (mxGetN(lev1) != faceind)
    throw_error("Faces added out of order");
  if (faceind == max_lev_size_[fdim]) {
    int newsize = faceind * 2 + 2;
    mxArray** oldcells = static_cast<mxArray**>(mxGetData(lev1));
    mxArray** newcells = 
      static_cast<mxArray**>(mxMalloc(2 * newsize * sizeof(mxArray*)));
    for (int j = 0; j < 2 * faceind; ++j)
      newcells[j] = oldcells[j];
    mxFree(oldcells);
    mxSetData(nclev1, static_cast<void*>(newcells));
    max_lev_size_[fdim] = newsize;
  }
  int numrows = 1 + ( (fdim == 0)? 0 : 1) + fdim;
  mxArray* newv = mxCreateDoubleMatrix(numrows, 1, mxREAL);
  mxSetN(newv, 0);
  (*max_nodes_per_face_)[fspec] = 1;
  mxSetN(nclev1, faceind + 1);
  mxSetCell(nclev1, faceind * 2, newv);

  mxArray* newe = mxCreateDoubleMatrix(fdim + 1, 1,  mxREAL);
  mxSetN(newe, 0);
  (*max_elems_per_face_)[fspec] = 1;
  mxSetCell(nclev1, faceind * 2 + 1, newe);
}


void 
QMG::SimpComplex_Under_Construction::
add_vertex_bdryinc(VertexGlobalIndex vnum,
                   const Brep::Face_Spec& fspec,
                   Brep::PatchIndex patchind,
                   const vector<Real>& param) {
  int fdim = fspec.fdim();
  Brep::FaceIndex faceind = fspec.faceind();
  if (global_to_ordinal_ -> find(vnum) == global_to_ordinal_ -> end())
    throw_error("Global vertex id not found");
  const mxArray* lev1 = mxGetCell(c_entries_, 5 + fdim);
#ifdef RANGECHECK
  if (fdim < 0 || fdim >= c_embedded_dim_ ||
      fdim > c_gdim_)
    throw_error("Face dim out of range");
  if (faceind < 0 || faceind >= mxGetN(lev1))
    throw_error("Face index out of range");
#endif
  if (param.size() != fdim) 
    throw_error("Wrong length for param coords");
  if (patchind < 0)
    throw_error("Patch index out of range");
  Vertex_Bdry_Inc_ inc;
  Vertex_Pos_ pos;
  inc.vnum = vnum;
  inc.fspec = fspec;
  for (int k1 = 0; k1 < fdim; ++k1)
    pos.param[k1] = param[k1];
  pos.patchind = patchind;
  pair<Vert_Bdry_Lookup_::iterator,bool> rval = 
    vertinclookup_ -> insert(pair<Vertex_Bdry_Inc_, Vertex_Pos_>(inc,pos));
  if (!rval.second) {
    return;
  }
  const mxArray* vtab1 = mxGetCell(lev1, 2 * faceind);
  mxArray* ncvtab1 = const_cast<mxArray*>(vtab1);
  int nn = mxGetN(vtab1);
  int numrow = 1 + ((fdim == 0)? 0 : 1) + fdim;
  int p = numrow * nn;
  int oldmax = (*max_nodes_per_face_)[fspec];
  if (nn == oldmax) {
    int newsize = oldmax * 2 + 2;
    double* newpr = 
      static_cast<double*>(mxMalloc(newsize * numrow * sizeof(double)));
    double* oldpr = mxGetPr(vtab1);
    for (int i = 0; i < p; ++i)
      newpr[i] = oldpr[i];
    mxFree(oldpr);
    mxSetPr(ncvtab1, newpr);
    (*max_nodes_per_face_)[fspec] = newsize;
  }
  mxSetN(ncvtab1, nn + 1);
  double* pr = mxGetPr(vtab1);
  pr[p] = static_cast<double>(vnum);
  if (fdim > 0)
    pr[++p] = static_cast<double>(patchind);
  for (int j = 0; j < fdim; ++j)
    pr[++p] = param[j];
}



QMG::Brep::PatchIndex 
QMG::SimpComplex_Under_Construction::
retrieve_vertex_bdryinc_patchind(VertexGlobalIndex vnum,
                                 const Brep::Face_Spec& fspec) const {
  Vertex_Bdry_Inc_ vp;
  vp.vnum = vnum;
  vp.fspec = fspec;
  Vert_Bdry_Lookup_::const_iterator it = 
    vertinclookup_ -> find(vp);
  if (it == vertinclookup_ -> end())
    throw_error("Vertex incidence not found");
  return it -> second.patchind;
}


QMG::Point
QMG::SimpComplex_Under_Construction::
retrieve_vertex_bdryinc_param(VertexGlobalIndex vnum,
                              const Brep::Face_Spec& fspec) const {
  Vertex_Bdry_Inc_ vp;
  vp.vnum = vnum;
  vp.fspec = fspec;
  Vert_Bdry_Lookup_::const_iterator it = 
    vertinclookup_ -> find(vp);
  if (it == vertinclookup_ -> end())
    throw_error("Vertex incidence not found");
  Point pt;
  pt[0] = it -> second.param[0];
  pt[1] = it -> second.param[1];
  return pt;
}


void
QMG::SimpComplex_Under_Construction::
add_simplex_face(const vector<VertexGlobalIndex>& verts,
                 const Brep::Face_Spec& fspec) { 
  int fdim = fspec.fdim();
  Brep::FaceIndex faceind = fspec.faceind();

#ifdef RANGECHECK
  if (fdim <= 0 || fdim > c_gdim_)
    throw_error("Face dim out of range");
#endif

  const mxArray* lev1 = mxGetCell(c_entries_, 5 + fdim);

#ifdef RANGECHECK
  if (faceind < 0 || faceind >= mxGetN(lev1))
    throw_error("Face index out of range");
  if (verts.size() != fdim + 1) 
    throw_error("Wrong length for vertex list");
#endif

  for (int jj = 0; jj < fdim + 1; ++jj) {
    if (global_to_ordinal_ -> find(verts[jj]) == global_to_ordinal_ -> end())
      throw_error("Global vertex id not found");
  }
  Face_Bdry_Inc_ inc;
  if (fdim < c_embedded_dim_) {
    {
      for (int j1 = 0; j1 < fdim + 1; ++j1)
        inc.meshv[j1] = verts[j1];
    }
    {
      for (int j1 = fdim + 1; j1 < 3; ++j1)
        inc.meshv[j1] =  -1;
    }
    {
      for (int j1 = 0; j1 < fdim; ++j1) {
        for (int j2 = j1 + 1; j2 < fdim + 1; ++j2) {
          if (inc.meshv[j2] < inc.meshv[j1]) {
            VertexGlobalIndex tmp = inc.meshv[j2];
            inc.meshv[j2] = inc.meshv[j1];
            inc.meshv[j1] = tmp;
          }
        }
      }
    }
    pair<Face_Bdry_Lookup_::iterator, bool> rval =
      faceinclookup_ -> insert(pair<Face_Bdry_Inc_,Brep::Face_Spec>(inc,fspec));
    if (!rval.second) {
      if (rval.first -> second != fspec) {
        throw_error("Attempt to add simplex face with mismatched fspecs");
      }
      return;
    }
  }
  const mxArray* etab1 = mxGetCell(lev1, 2 * faceind + 1);
  mxArray* ncetab1 = const_cast<mxArray*>(etab1);
  int ne = mxGetN(etab1);
  int numrow = fdim + 1;
  int p = numrow * ne;
  int oldmax = (*max_elems_per_face_)[fspec];
  if (ne == oldmax) {
    int newsize = oldmax * 2 + 2;
    double* newpr = 
      static_cast<double*>(mxMalloc(newsize * numrow * sizeof(double)));
    double* oldpr = mxGetPr(etab1);
    for (int i = 0; i < p; ++i)
      newpr[i] = oldpr[i];
    mxFree(oldpr);
    mxSetPr(ncetab1, newpr);
    (*max_elems_per_face_)[fspec] = newsize;
  }
  mxSetN(ncetab1, ne + 1);
  double* pr = mxGetPr(etab1);
  for (int j = 0; j < fdim + 1; ++j)
    pr[p++] = static_cast<double>(verts[j]);
  //dump610(c_entries_);
}


void
QMG::SimpComplex_Under_Construction::remove_unused_vertices() {
  map<VertexGlobalIndex, int> lookupnode;
  int levsize = mxGetN(mxGetCell(c_entries_, 5 + c_gdim_));
  {
    for (Brep::FaceIndex faceind = 0; faceind < levsize; ++faceind) {
      Brep::Face_Spec fspec(c_gdim_, faceind);
      int nf = this -> num_meshface_on_face(fspec);
      for (int seqno = 0; seqno < nf; ++seqno) {
        for (int l = 0; l < c_gdim_ + 1; ++l) {
          VertexGlobalIndex vnum = 
            this -> node_of_meshface_on_face(fspec, seqno, l);
          lookupnode[vnum] = 1;
        }
      }
    }
  }
  VertexOrdinalIndex dest_vnumo = 0;
  mxArray* node_array = mxGetCell(c_entries_, 4);
  int nn = mxGetN(node_array);
  double* nodepr = mxGetPr(node_array);
  int numrows = c_embedded_dim_ + 1;
  for (VertexOrdinalIndex source_vnumo = 0;
  source_vnumo < nn;
  ++source_vnumo) {
    VertexGlobalIndex vnum = 
      static_cast<int>(nodepr[source_vnumo * numrows]);
    if (lookupnode[vnum]) {
      for (int jj = 0; jj < numrows; ++jj)
        nodepr[dest_vnumo * numrows + jj] = 
          nodepr[source_vnumo * numrows + jj];
      ++dest_vnumo;
    }
  }
  mxSetN(node_array, dest_vnumo);


  for (int fdim = 0; fdim < c_gdim_; ++fdim) {
    const mxArray* lev1 = mxGetCell(c_entries_, fdim + 5);
    int levsize = mxGetN(lev1);
    for (Brep::FaceIndex faceind = 0; faceind < levsize; ++faceind) {
      Brep::Face_Spec fspec(fdim, faceind);
      mxArray* facenodes = mxGetCell(lev1, 2 * faceind);
      double* fnr = mxGetPr(facenodes);
      int nnf = mxGetN(facenodes);
      int numr1 = 1 + ((fdim == 0)? 0:1) + fdim;
      int destseqno = 0;
      for (int seqno = 0; seqno < nnf; ++seqno) {
        VertexGlobalIndex vnum = 
          static_cast<int>(fnr[seqno * numr1]);
        if (lookupnode[vnum]) {
          for (int j = 0; j < numr1; ++j)
            fnr[destseqno * numr1 + j] = fnr[seqno * numr1 + j];
          ++destseqno;
        }
      }
      mxSetN(facenodes, destseqno);
      if (fdim == 0) continue;
      mxArray* simpnodes = mxGetCell(lev1, 2 * faceind + 1);
      double* snr = mxGetPr(simpnodes);
      int nmf = mxGetN(simpnodes);
      numr1 = fdim + 1;
      int destseqno2 = 0;
      for (int seqno2 = 0; seqno2 < nmf; ++seqno2) {
        bool all_still_there = true;
        for (int jj = 0; jj < fdim + 1; ++jj) {
          VertexGlobalIndex vnum = static_cast<int>(snr[seqno2 * numr1 + jj]);
          if (!lookupnode[vnum]) {
            all_still_there = false;
            break;
          }
        }
        if (all_still_there) {
          for (int j2 = 0; j2 < fdim + 1; ++j2)
            snr[destseqno2 * numr1 + j2] = snr[seqno2 * numr1 + j2];
          ++destseqno2;
        }
      }
      mxSetN(simpnodes, destseqno2);
    }
  }
}

void 
QMG::SimpComplex_Under_Construction::add_prop_val(const string& prop, 
                                                  const string& val) {
  mxArray* pvlist = mxGetCell(c_entries_, 3);
  int npv = mxGetN(pvlist);
  if (npv == max_prop_val_) {
    int newsize = npv * 2 + 2;
    mxArray** oldcells = static_cast<mxArray**>(mxGetData(pvlist));
    mxArray** newcells = 
      static_cast<mxArray**>(mxMalloc(2 * newsize * sizeof(mxArray*)));
    for (int j = 0; j < 2 * npv; ++j)
      newcells[j] = oldcells[j];
    mxFree(oldcells);
    mxSetData(pvlist, static_cast<void*>(newcells));
    max_prop_val_ = newsize;
  }
  mxSetN(pvlist, npv + 1);
  mxSetCell(pvlist, 2 * npv, mxCreateString(prop.c_str()));
  mxSetCell(pvlist, 2 * npv + 1, mxCreateString(val.c_str()));
}



QMG::SimpComplex_Under_Construction::
SimpComplex_Under_Construction(int intrinsic_dim, 
			       int embedded_dim, 
                               const vector<int>& level_size) {
  if (embedded_dim != 2 && embedded_dim != 3)
    throw_error("Illegal embedded dimension");
  if (intrinsic_dim < 1 || intrinsic_dim > embedded_dim)
    throw_error("Illegal intrinsic dimension");
  int brep_gdim = level_size.size() - 1;
  if (brep_gdim < intrinsic_dim || brep_gdim > embedded_dim)
    throw_error("illegal brep intrinsic dimension");


  mxArray* plhsa[1];
  mxArray* prhsa[1];
  mexCallMATLAB(1,plhsa,0, prhsa, "make_empty_zba");
  mesh_ = plhsa[0];
  mxArray* oldentries = mxGetFieldByNumber(mesh_,0, 0);
  if (oldentries)
    mxFree(oldentries);

  c_gdim_ = intrinsic_dim;
  c_embedded_dim_ = embedded_dim;

  mxArray* entries = mxCreateCellMatrix(6+intrinsic_dim,1);
  c_entries_ = entries;
  mxSetFieldByNumber(const_cast<mxArray*>(mesh_),0,0,entries);


  mxArray* typecode = mxCreateString(io_header_code());
  mxSetCell(entries, 0, typecode);


  mxArray* gdima = mxCreateDoubleMatrix(1,1, mxREAL);
  *mxGetPr(gdima) = static_cast<double>(intrinsic_dim);
  mxSetCell(entries, 1, gdima);

  mxArray* edima = mxCreateDoubleMatrix(1,1,mxREAL);
  *mxGetPr(edima) = static_cast<double>(embedded_dim);
  mxSetCell(entries, 2, edima);

  mxArray* pvlist = mxCreateCellMatrix(2,1);
  mxSetN(pvlist, 0);
  max_prop_val_ = 1;
  mxSetCell(entries, 3, pvlist);

  mxArray* nodearray = mxCreateDoubleMatrix(1 + c_embedded_dim_, 1, mxREAL);
  mxSetN(nodearray, 0);
  max_nodes_ = 1;
  mxSetCell(entries, 4, nodearray);
  c_nodearray_ = mxGetPr(nodearray);


  max_lev_size_.resize(c_gdim_ + 1);
  max_nodes_per_face_ = new map<Brep::Face_Spec, int>;
  max_elems_per_face_ = new map<Brep::Face_Spec, int>;
  for (int fdim = 0; fdim <= c_gdim_; ++fdim) {
    int lsize = level_size[fdim];
    if (lsize < 0)
      throw_error("Level size out of bounds");
    int maxlevsize = (lsize == 0)? 1 : lsize;
    max_lev_size_[fdim] = maxlevsize;
    mxArray* lev = mxCreateCellMatrix(2, maxlevsize);
    mxSetN(lev, lsize);
    for (Brep::FaceIndex faceind = 0; faceind < lsize; ++faceind) {
      Brep::Face_Spec fspec(fdim, faceind);
      mxArray* nodes = 
        mxCreateDoubleMatrix(1 + ((fdim == 0)? 0 : 1) + fdim, 1, mxREAL);
      mxSetN(nodes, 0);
      mxSetCell(lev, 2 * faceind, nodes);
      (*max_nodes_per_face_)[fspec] = 1;
      mxArray* simps = 
        mxCreateDoubleMatrix(fdim + 1, 1, mxREAL);
      mxSetN(simps, 0);
      mxSetCell(lev, 2 * faceind + 1, simps);
      (*max_elems_per_face_)[fspec] = 1;
    }
    mxSetCell(entries, 5 + fdim, lev);
  }
  vertinclookup_ = new Vert_Bdry_Lookup_;
  faceinclookup_ = new Face_Bdry_Lookup_;
  global_to_ordinal_ = new GlobalToOrdinalMapType;
  next_vnum_ = 0;
}

mxArray*
QMG::SimpComplex_Under_Construction::release() {
  mxArray* tmp = const_cast<mxArray*>(mesh_);
  delete faceinclookup_;
  delete vertinclookup_;
  delete global_to_ordinal_;
  delete max_elems_per_face_;
  delete max_nodes_per_face_;
  clobber_();
  return tmp;
}


QMG::SimpComplex_Under_Construction::~SimpComplex_Under_Construction() {
  if (mesh_)
    mxDestroyArray(const_cast<mxArray*>(mesh_));
  delete faceinclookup_;
  delete vertinclookup_;
  delete global_to_ordinal_;
  delete max_elems_per_face_;
  delete max_nodes_per_face_;
}
  

QMG::SimpComplex_Under_Construction::
SimpComplex_Under_Construction(SimpComplex_Under_Construction& other) :
SimpComplex(other),
max_nodes_per_face_(other.max_nodes_per_face_),
max_elems_per_face_(other.max_elems_per_face_),
max_lev_size_(other.max_lev_size_),
max_nodes_(other.max_nodes_),
max_prop_val_(other.max_prop_val_),
next_vnum_(other.next_vnum_),
vertinclookup_(other.vertinclookup_),
faceinclookup_(other.faceinclookup_) {
  other.clobber_();
}


QMG::SimpComplex_Under_Construction::
SimpComplex_Under_Construction(const SimpComplex_Under_Construction_Returnval& sucr) :
SimpComplex(sucr.sref_),
max_nodes_per_face_(sucr.sref_.max_nodes_per_face_),
max_elems_per_face_(sucr.sref_.max_elems_per_face_),
max_lev_size_(sucr.sref_.max_lev_size_),
max_nodes_(sucr.sref_.max_nodes_),
max_prop_val_(sucr.sref_.max_prop_val_),
next_vnum_(sucr.sref_.next_vnum_),
vertinclookup_(sucr.sref_.vertinclookup_),
faceinclookup_(sucr.sref_.faceinclookup_) {
  sucr.sref_.clobber_();
}



