// ------------------------------------------------------------------
// crtsubbox.cpp
//
// This file contains member functions from qboxstack.h for creating
// subboxes. 
// ------------------------------------------------------------------
// Author: Stephen A. Vavasis
// Copyright (c) 1999 by Cornell University.  All rights reserved.
// 
// See the accompanying file 'Copyright' for authorship information,
// the terms of the license governing this software, and disclaimers
// concerning this software.
// ------------------------------------------------------------------
// This file is part of the QMG software.  
// Version 2.0 of QMG, release date September 3, 1999.
// ------------------------------------------------------------------


#ifdef _MSC_VER
#if _MSC_VER == 1200
#include "qmatvec.h"
#endif
#endif


#include "qboxstack.h"
#include "qunionfind.h"

namespace QMG {
  namespace MG {
    using namespace QMG;

    
    //global function in this file


    void boxstack_init_static_data(int di); 
  }
}

// File scope functions.


#ifdef DEBUGGING
namespace QMG {
  namespace Local_to_PatchMath_CPP {
    using namespace QMG;
    extern ostream* debug_str;
    extern bool dump_everything;
  }
}
#endif


namespace {

  using namespace QMG;
  using namespace QMG::MG;


  // ------------------------------------------------------------------
  // Combinatorial merge: merge each patch with its child patches.  This leaves
  // all bounding faces in the same connected component with the same label.
  // The return value is left in workspace variable wkspa.uf.

  void combinatorial_merge(const ActiveBox& b,
    ActiveBoxVec::Create_Subbox_Workspace& wkspa) {

 
    wkspa.map_to_relnum.reset(-1);
  
    for (int relnum = 0; relnum < b.num_patch(); relnum++) {
      PatchTable::Index p = b.patch(relnum).patchind();
      wkspa.map_to_relnum[p] = relnum;
      // Loop over the children patches.
      for (PatchTable::Loop_over_children loop(wkspa.patchtable, p);
      loop.notdone();
      ++loop) {
        PatchTable::Index c = loop.child_index();
        int relnum2 = wkspa.map_to_relnum[c];
        if (relnum2 >= 0) {
          wkspa.uf.merge(relnum, relnum2);
        }
      }
    }
  }  
  
  // Merge patches that belong to the same brep face.  Necessary else
  // alignment won't work correctly. Leave results in uf.

  void combinatorial_merge2(const ActiveBox& b,
    ActiveBoxVec::Create_Subbox_Workspace& wkspa) {
    {
      for (int relnum = 0; relnum < b.num_patch(); ++relnum) {
        PatchTable::Index p = b.patch(relnum).patchind();
        wkspa.example_patchrelnum[wkspa.patchtable.owner(p)] = relnum;
      }
    }

    {
      for (int relnum = 0; relnum < b.num_patch(); ++relnum) {
        PatchTable::Index p = b.patch(relnum).patchind();
        int relnum2 = wkspa.example_patchrelnum[wkspa.patchtable.owner(p)];
        wkspa.uf.merge(relnum, relnum2);
      }
    }
  }




    // ------------------------------------------------------------------
  // Draw a segment through a box; return the sorted list of hits on patches.
  // Used for geometric merging of connected components.  Direct return
  // value is true if there is a degeneracy else false.  The sorted list
  // is returned in workspace variable wkspa.hitlist.

  
  bool draw_segment_thru_box(const ActiveBox& b,
    IncidenceTable::Index startpoint_incind,
    int startpoint_relnum,
    const Point& endpoint,
    ActiveBoxVec::Create_Subbox_Workspace& wkspa) {

    wkspa.hitlist.resize(0, ActiveBoxVec::RayHitRecord());

    PatchTable::Index startpoint_patchind = wkspa.inc_table.patchind(startpoint_incind);
    Point startpoint = wkspa.inc_table.real_coord(startpoint_incind);
    int di = wkspa.patchtable.embedded_dim();

    Point normal;
    Point segment_tangent = Point::subtract(endpoint, startpoint);
    
    Real farthest_hit_dist = 0.0;
    pair<PatchTable::Index, PatchTable::ConeIntersectCode> rval1 =
      wkspa.patchtable.determine_ray_front_face(startpoint_patchind, 
      wkspa.inc_table.param_coord(startpoint_incind),
      segment_tangent, wkspa.tol);
    if (rval1.second == PatchTable::DEGEN) {
      if (wkspa.logstr.verbosity() >= 8) {
        wkspa.logstr.str() << " Degeneracy in determine_ray_front_face\n";
      }
      return true;
    }

    PatchTable::Index patchind_s = rval1.first;
    Brep::Face_Spec base_fspec = (rval1.second == PatchTable::HITS_PATCH_FRONT)? 
      wkspa.patchtable.front_side(patchind_s) : wkspa.patchtable.back_side(patchind_s);

    if (wkspa.logstr.verbosity() >= 8)
      wkspa.logstr.str() <<  "reached shootray tangent = " << segment_tangent 
      << " start point = " << startpoint << "= incd#" << wkspa.inc_table.seqno(startpoint_incind)
      << "#"
      <<   "\n basefspec = " << base_fspec << " startpoint_patchind = pat#" 
      << startpoint_patchind << "#\n"
      << " endpoint = " << endpoint << '\n'; 

    ActiveBoxVec::RayHitRecord rayhit;
    rayhit.relnum = startpoint_relnum;
    rayhit.front_hit = base_fspec;
    rayhit.dist = 0.0; 
    rayhit.is_front = rval1.second == PatchTable::HITS_PATCH_FRONT;
    wkspa.hitlist.push_back(rayhit);

    for (int relnum = 0; relnum < b.num_patch(); ++relnum) {
      PatchTable::Index patchind = b.patch(relnum).patchind();      
      if (wkspa.patchtable.gdim(patchind) < di - 1)
        continue;

      if (wkspa.logstr.verbosity() >= 8) {
        wkspa.logstr.str() << "loop relnum = " << relnum
          << "pat#" << patchind << "#" << " orig = " 
          << wkspa.patchtable.owner(patchind) << " # " 
          << wkspa.patchtable.orig_index(patchind) << '\n';
      }

      bool require_startpoint_degen = 
        wkspa.patchtable.is_ancestor(startpoint_patchind, patchind);

      wkspa.output_stack.resize(0, PatchMath::IntersectRecord());
      wkspa.patchtable.patchmath(patchind).intersect_seg(wkspa.output_stack, wkspa.rworkspace, 
        startpoint, endpoint, wkspa.scfac, wkspa.tol, false);      
      if (wkspa.logstr.verbosity() >= 8)
        wkspa.logstr.str() << "require startpoint degen = " << require_startpoint_degen 
        << " size = " << wkspa.output_stack.size() << '\n';

      int sz = wkspa.output_stack.size();
      int degen_count = 0;
      for (int ii = 0; ii < sz; ++ii) {
        if (wkspa.logstr.verbosity() >= 8) {
          wkspa.logstr.str() << "hit " << ii << " at real point " 
            << wkspa.output_stack[ii].realcoord << " param " << wkspa.output_stack[ii].paramcoord
            << "\n degen = " << wkspa.output_stack[ii].degen
            << " uncer = " << wkspa.output_stack[ii].param_uncertainty
            << " seguncer = " << wkspa.output_stack[ii].param_uncertainty_seg
            << " dist = " << wkspa.output_stack[ii].dist << '\n';
        }
        if (wkspa.output_stack[ii].degen) {
          ++degen_count;
          continue;
        }
        normal = 
          wkspa.patchtable.patchmath(patchind).normal(wkspa.output_stack[ii].paramcoord);
        Real ip = Point::inner_product(normal, segment_tangent);
        rayhit.relnum = relnum;
        rayhit.dist = wkspa.output_stack[ii].dist;        
        if (ip > 0) {
          rayhit.front_hit = wkspa.patchtable.front_side(patchind);
          rayhit.back_hit = wkspa.patchtable.back_side(patchind);
          rayhit.is_front = true;
        }
        else {
          rayhit.front_hit = wkspa.patchtable.back_side(patchind);
          rayhit.back_hit = wkspa.patchtable.front_side(patchind);
          rayhit.is_front = false;
        }
        if (wkspa.logstr.verbosity() >= 8) {
          wkspa.logstr.str() << " front = " << rayhit.front_hit << " back = " << rayhit.back_hit << '\n';
        }
        wkspa.hitlist.push_back(rayhit);
      }
      
      if (degen_count > 1 || (degen_count > 0 && !require_startpoint_degen))
        return true;
      if (degen_count == 0 && require_startpoint_degen) {

        wkspa.logstr.str() << "Error! Output stack size = " << sz << '\n';
        {
          for (int jj = 0; jj < sz; ++jj) {
            wkspa.logstr.str() << "output_stack[" << jj << "] dist = "
              << wkspa.output_stack[jj].dist << " degen = "
              << wkspa.output_stack[jj].degen << " paramcoord = "
              << wkspa.output_stack[jj].paramcoord << '\n';
          }
        }
        QMG::Local_to_PatchMath_CPP::debug_str = &wkspa.logstr.str();
        QMG::Local_to_PatchMath_CPP::dump_everything = true;
        wkspa.output_stack.resize(0, PatchMath::IntersectRecord());
        wkspa.patchtable.patchmath(patchind).intersect_seg(wkspa.output_stack, wkspa.rworkspace,
          startpoint, endpoint,
          wkspa.scfac, wkspa.tol, false);
        
        int sz = wkspa.output_stack.size();
        wkspa.logstr.str() << " Second time Output stack size = " << sz << '\n';
        {
          for (int jj = 0; jj < sz; ++jj) {
            wkspa.logstr.str() << "wkspa.output_stack[" << jj << "] dist = "
              << wkspa.output_stack[jj].dist << " degen = "
              << wkspa.output_stack[jj].degen << " paramcoord = "
              << wkspa.output_stack[jj].paramcoord << '\n';
          }
        }
        throw_error("Ray missed starting patch?  This message may indicate a tolerance error");
      }
    }
    sort(wkspa.hitlist.begin(), wkspa.hitlist.end());


#ifdef DEBUGGING
    for (int hitlistind = 0; 
    hitlistind < wkspa.hitlist.size() - 1; 
    hitlistind++) {
      if (wkspa.hitlist[hitlistind].front_hit != wkspa.hitlist[hitlistind+1].back_hit) {
        ostringstream os;
        os << "Problem 2 in draw_segment.\n"
          << "This error may indicate a problem with illegally crossing faces\n"
          << "in the input brep. The two surfaces causing the error are\n"
          << wkspa.patchtable.owner(b.patch(wkspa.hitlist[hitlistind].relnum).patchind())
          << " and "
          << wkspa.patchtable.owner(b.patch(wkspa.hitlist[hitlistind+1].relnum).patchind());
        throw_error(os.str());
      }
    }
#endif
    return false;
  }

    // ------------------------------------------------------------------
  // Merge connected components of the content of a box by shooting rays.
  // Return variable is wkspa.uf (union find).
  
  void segshoot_merge(const ActiveBox& b,
    ActiveBoxVec::Create_Subbox_Workspace& wkspa) {
    
    int di = wkspa.patchtable.embedded_dim();
    int ufsize = wkspa.uf.size();     
    wkspa.ray_shot_from_label.resize(ufsize);
    for (int jj = 0; jj < ufsize; ++jj)
      wkspa.ray_shot_from_label[jj] = false;

    bool degenhit = true;
    int trycount = 0;
    if (wkspa.logstr.verbosity() >= 8) {
      wkspa.logstr.str() << " reached segshoot merge box = " << b << '\n';
    }

    int oldverb = wkspa.logstr.verbosity();

    while (degenhit && ++trycount < 400) {
      Point randpoint = b.select_rand_point();
      if (trycount == 398) {        
        wkspa.logstr.change_verbosity(8);
        wkspa.logstr.str() << "trycount at 398 b = ";
        b.full_dump(wkspa.logstr.str());
      }
      if (wkspa.logstr.verbosity() >= 8) {
        wkspa.logstr.str() << "trycount = " << trycount 
          << " randpoint = " << randpoint << "\n";
      }

      degenhit = false;
      int root_component_patchrelnum = -1;

      for (int relnum1 = 0; relnum1 < b.num_patch(); ++relnum1) 
        wkspa.ray_shot_from_label[relnum1] = false;


      for (int relnum = 0; relnum < b.num_patch(); ++relnum) {
        PatchTable::Index patchind = b.patch(relnum).patchind();
        int c = wkspa.uf.get_label(relnum);
        if (wkspa.logstr.verbosity() >= 8) {
          wkspa.logstr.str() << " in loop relnum = " << relnum << " c = " << c
            << " rsfl = " << wkspa.ray_shot_from_label[c] << '\n';
        }      

        if (!wkspa.ray_shot_from_label[c]) {

          // Select starting point on patch relnum.

          Patch_in_ActiveBox ptc = b.patch(relnum);
          if (ptc.num_incidence() == 0) 
            throw_error("No incidences on patch of component?");
          IncidenceTable::Index incind = ptc.incidence(0);
          if (wkspa.logstr.verbosity() >= 8) {
            wkspa.logstr.str() << " calling draw seg from inc incd#" 
              << wkspa.inc_table.seqno(incind) << "#\n";
          }

          degenhit = draw_segment_thru_box(b, incind, relnum, randpoint, wkspa);
          
          // Last call puts return list in wkspa.hitlist.

          if (degenhit) {
            if (wkspa.logstr.verbosity() >= 8) {
               wkspa.logstr.str() << "degenhit = true\n";
            }
            break;
          }
#ifdef DEBUGGING
          // First hit should be on p.
          if (wkspa.hitlist.size() == 0 || wkspa.hitlist[0].relnum != relnum) {
            throw_error("Problem 1 in rayshoot_merge");
            return;
          }
#endif        

          if (wkspa.logstr.verbosity() >= 8) {
            wkspa.logstr.str() << " hitlist size = " << wkspa.hitlist.size() << '\n';
            for (int iii = 0; iii < wkspa.hitlist.size(); ++iii) {
              wkspa.logstr.str() << "hit list " << iii << " dist = " << wkspa.hitlist[iii].dist 
                << " relnum = " <<  wkspa.hitlist[iii].relnum
                << " front = " << wkspa.hitlist[iii].front_hit << " back = " 
                << wkspa.hitlist[iii].back_hit << '\n';
              
            }
          }
          // Merge each item in the hit list with its successor.

          for (int hitlistind = 0; 
          hitlistind < wkspa.hitlist.size() - 1; 
          hitlistind++) {

            int relnum1 = wkspa.hitlist[hitlistind].relnum;
            int relnum2 = wkspa.hitlist[hitlistind + 1].relnum;
            
            if (wkspa.logstr.verbosity() >= 8) {              
              wkspa.logstr.str() << "hitlistind = " << hitlistind <<  " relnum1 = " << relnum1
                << " relnum2 = " << relnum2 << " fronthit = "
                << wkspa.hitlist[hitlistind].front_hit.fdim() << ":" 
                << wkspa.hitlist[hitlistind].front_hit.faceind() << " backhit = "
                << wkspa.hitlist[hitlistind+1].back_hit.fdim() << ':' 
                << wkspa.hitlist[hitlistind+1].back_hit.faceind()
                << '\n';
            }


            wkspa.ray_shot_from_label[wkspa.uf.get_label(relnum1)] = true;
            
            if (wkspa.hitlist[hitlistind].front_hit.faceind() >= 0) {
              if (wkspa.logstr.verbosity() >= 8) {
                wkspa.logstr.str() << " merging\n";
              }
              wkspa.uf.merge(relnum1, relnum2);
            }
          }
            
            // Now merge the last item with the root if the segment is inside.
            
          int lastind = wkspa.hitlist.size() - 1;
          int relnum1 = wkspa.hitlist[lastind].relnum;
          wkspa.ray_shot_from_label[wkspa.uf.get_label(relnum1)] = true;
          
          if (wkspa.hitlist[lastind].front_hit.faceind() >= 0) {
            if (wkspa.logstr.verbosity() >= 8) {
              wkspa.logstr.str() << "last hit face " 
                << wkspa.hitlist[lastind].front_hit.fdim() << ":" 
                << wkspa.hitlist[lastind].front_hit.faceind() 
                << " root compo relnum = " << root_component_patchrelnum << '\n';
              
            }
            if (root_component_patchrelnum < 0)
              root_component_patchrelnum = relnum1;
            else
              wkspa.uf.merge(relnum1, root_component_patchrelnum);
          }
        }
      }
    }
    wkspa.logstr.change_verbosity(oldverb);
    if (degenhit) {
      throw_error("unable to resolve degeneracy in ray_shoot_component");
    }
 }


  // ------------------------------------------------------------------
  // Find connected components of a box's content.    Return # of components;
  // leave labels in the wkspa.uf argument.

  int mark_components(const ActiveBox& b, 
    ActiveBoxVec::Create_Subbox_Workspace& wkspa) {

    if (b.completely_interior())
      return 1;

    int numpatch = b.num_patch();
    wkspa.uf.initialize(numpatch);

    combinatorial_merge(b, wkspa);
    combinatorial_merge2(b, wkspa);
    int numcompo = wkspa.uf.pack_labels();

    if (wkspa.logstr.verbosity() >= 6) {
      wkspa.logstr.str() << "Mark components: after comb merge on " << numpatch
        << " patches, #components = " << numcompo << '\n';
      for (int relnum1 = 0; relnum1 < numpatch; ++relnum1) {
        wkspa.logstr.str() << "pat#" << b.patch(relnum1).patchind() << "#lab = "
          << wkspa.uf.get_label(relnum1) << " ";
      }
      wkspa.logstr.str() << '\n';
    }



    if (numcompo > 1) {
      segshoot_merge(b, wkspa);
      numcompo = wkspa.uf.pack_labels();
      if (wkspa.logstr.verbosity() >= 6) {
        wkspa.logstr.str() << "Mark components: after segshoot merge on " << numpatch
          << " patches, #components = " << numcompo << '\n';
        for (int relnum1 = 0; relnum1 < numpatch; ++relnum1) {
          wkspa.logstr.str() << "pat#" << b.patch(relnum1).patchind() << "#lab = "
            << wkspa.uf.get_label(relnum1) << " ";
        }
        wkspa.logstr.str() << '\n';
      }
    }
    return numcompo;
  }
}

bool QMG::MG::ActiveBoxVec::RayHitRecord::operator<(const RayHitRecord& o) const {
  return dist < o.dist;
}


// ------------------------------------------------------------------
// If the new box doesn't meet any patches, then we have to decide
// whether it is outside the brep or completely interior.  This requires
//  a ray-shoot test.  Shoot a ray from any point in the new box to
// a patch in the old box; see if we hit the front or back of a patch.
// This is a virtual function that is overridden for certain types
// of box parents.
// The return value is the face-spec and a bool which is true if the
// hit is on the front of a face.

#ifndef BUG_IN_USING
using QMG::pair;
#endif

pair<QMG::Brep::Face_Spec, bool> 
QMG::MG::ActiveBox_As_Parent::
chamber_containing_subbox(const ActiveBox& testbox,
                          ActiveBoxVec::Create_Subbox_Workspace& wkspa) const {
  
  // Select a starting point

  bool startpoint_selected = false;
  IncidenceTable::Index startpt;
  int startpoint_relnum;

  for (int relnum = 0; relnum < this -> num_patch(); ++relnum) {
    this -> incloop_init(relnum);
    for (int incind = 0; incind < this -> incloop_numinc(); ++incind) {
      startpt = this -> incloop_getinc(incind);
      startpoint_selected = true;
      startpoint_relnum = relnum;
      break;
    }
    if (startpoint_selected) break;
  }
  if (!startpoint_selected)
    throw_error("box has no incidences");

  

  if (wkspa.logstr.verbosity() >= 8) {
    wkspa.logstr.str() << "reached chamber containing box src box is ";
    this -> full_dump(wkspa.logstr.str());
    wkspa.logstr.str() << " target box is " << testbox << "\n startpoint = "
      << "incd#" << wkspa.inc_table.seqno(startpt) << "#\n";
  }
  
  int trycount = 0;
  Point randpoint;
  int oldverb = wkspa.logstr.verbosity();

  while (++trycount < 400) {
    randpoint =  testbox.select_rand_point();
    
#ifdef DEBUGGING
    if (trycount == 398) {
      wkspa.logstr.change_verbosity(8);
      Local_to_PatchMath_CPP::dump_everything = true;
    }

    if (wkspa.logstr.verbosity() >= 8) {
      wkspa.logstr.str() << "shootray randpoint = " << randpoint << '\n';
    }
#endif


    // Shoot a ray from the starting point.

    
    bool degenhit = draw_segment_thru_box(*this, startpt, startpoint_relnum,
      randpoint, wkspa);
    
    
#ifdef DEBUGGING
    if (wkspa.logstr.verbosity() >= 8) {
      wkspa.logstr.str() << "result = " << degenhit << '\n';
    }
#endif

    if (!degenhit) {
      wkspa.logstr.change_verbosity(oldverb);
      Local_to_PatchMath_CPP::dump_everything = false;
      return pair<Brep::Face_Spec,bool>(wkspa.hitlist.back().front_hit,
        wkspa.hitlist.back().is_front);
    }
  }
  
#ifdef DEBUGGING
  this -> full_dump(wkspa.logstr.str());
  wkspa.logstr.str() << " New box = " << testbox;
  wkspa.logstr.str() <<  " randpoint = " << randpoint << " newexpa = " << this -> expa() << '\n';
  wkspa.logstr.str() << " startpoint = "
      << "incd#" << wkspa.inc_table.seqno(startpt) << "#\n";
#endif
  throw_error("Unrecoverable degeneracy in interior test");
  return pair<Brep::Face_Spec,bool>(Brep::Face_Spec(0,0), false);
}


// ------------------------------------------------------------------
// Creates a new box on a specified vector; doesn't deal with weights;
// Does not create more than one box.  Returns the lowest-dimensional
// patch in the box.

int 
QMG::MG::ActiveBoxVec::
create_new_subbox_(const ActiveBox_As_Parent& b,  //parent box
                   const BoxAddress& bxa,
                   int bwidth,
                   BoxFaceDag::Index parentind,
                   int facetval,
                   int facetdim,
                   const PatchTableIndexMap<int>& componum,
                   int current_componum,
                   Create_Subbox_Workspace& wkspa) {

  
  Index newboxind = this -> boxvec_.size();
  this -> boxvec_.push_back(Box_Data());
  Box_Data& newboxd = this -> boxvec_.back();
  newboxd.bxa = bxa;
  newboxd.link_parent = parentind;
  newboxd.link_parent_side = facetval;
  newboxd.link_parent_dim = facetdim;
  newboxd.box_completely_interior = b.completely_interior();
  newboxd.size_function_verified = b.size_function_verified();
  newboxd.has_close_face = false;
  newboxd.killed = false;
  newboxd.is_front_surface = b.is_front_surface();
  if (b.completely_interior()) {
    newboxd.num_patch = b.owner();
    newboxd.curvature_verified = true;
    return di_;
  }
  newboxd.curvature_verified = b.curvature_verified();
  int lowest_patch_dim = di_ + 1;
  int patchvec_initsize = this -> patchvec_.size();
  int incidencevec_initsize = this -> incidencevec_.size();
  newboxd.patchvec_begin = patchvec_initsize;
  newboxd.num_patch = 0;
  ActiveBox temp_b(*this, newboxind);
  
  if (wkspa.logstr.verbosity() >= 2) { ////
    wkspa.logstr.str() << "create new box, parent = " << b << " new box = " << temp_b << '\n';
  }
  
  Patch_Data newpatchd;
  wkspa.patch_was_included.reset(My_bool(false));
  
  for (int relnum = 0; relnum < b.num_patch(); ++relnum) {
    PatchTable::Index patchind = b.patch(relnum).patchind();
    if (componum[patchind] != current_componum) continue;
    newpatchd.inc_begin = this -> incidencevec_.size();
    newpatchd.num_inc = 0;
    b.incloop_init(relnum);
    for (int incind = 0; incind < b.incloop_numinc(); ++incind) {
      IncidenceTable::Index inc = b.incloop_getinc(incind);
      if (temp_b.contains_point_in_ex(wkspa.inc_table.real_coord(inc), 
        wkspa.inc_table.param_uncertainty(inc),
        wkspa.scfac)) {
        this -> incidencevec_.push_back(inc);
        ++newpatchd.num_inc;
      }
    }
    
    bool insert_patch = (newpatchd.num_inc > 0);
    if (!insert_patch) {
      for (PatchTable::Loop_over_children cloop(wkspa.patchtable, patchind);
      cloop.notdone();
      ++cloop) {
        if (wkspa.patch_was_included[cloop.child_index()]) {
          insert_patch = true;
          break;
        }
      }
    }
    
    if (insert_patch) {
      if (wkspa.logstr.verbosity() >= 7)
        wkspa.logstr.str() << "inserting pat#" << patchind << "#\n";
      newpatchd.patchnum = patchind;
      wkspa.patch_was_included[patchind] = true;
      this -> patchvec_.push_back(newpatchd);
      ++newboxd.num_patch;
      int patch_owner_dim = wkspa.patchtable.owner(patchind).fdim();
      if (patch_owner_dim < lowest_patch_dim) {
        lowest_patch_dim = patch_owner_dim;
      }
    }
  }
  
  
  // If the new box doesn't meet any patches, then we have to decide
  // whether it is outside the brep or completely interior.  This requires
  //  a ray-shoot test.  Shoot a ray from any point in the new box to
  // a patch in the old box; see if we hit the front or back of a patch.

  if (wkspa.logstr.verbosity() >= 5)
    wkspa.logstr.str() << " #patches = " << newboxd.num_patch << '\n';
  

  
  if (newboxd.num_patch == 0) {
    pair<Brep::Face_Spec,bool> rval = b.chamber_containing_subbox(temp_b, wkspa);
    Brep::Face_Spec owner = rval.first;
    if (owner.faceind() >= 0) {
      newboxd.box_completely_interior = true;
      newboxd.curvature_verified = true;
      newboxd.num_patch = owner.faceind();
      newboxd.is_front_surface = (rval.second)? 1 : 0;
      lowest_patch_dim = di_;
    }
  }
  if (wkspa.logstr.verbosity() >= 5) {
    wkspa.logstr.str() << "returning from create_new_subbox val = "
      << lowest_patch_dim << '\n';
  }
  return lowest_patch_dim;
}
  


#ifdef DEBUGGING
namespace QMG {
  namespace Local_to_BoxStack_CPP {
    BoxAddress bxa_check;
  }
}
#endif


#ifdef NSF_DEMO
namespace QMG {
  namespace Nsf_Demo {
    using namespace QMG;
    using namespace QMG::MG;
    extern void show_box(const ActiveBox& b, int code);
  }
}
#endif




// ------------------------------------------------------------------
// create_subbox
// This is the main routine for creating new boxes from parent boxes
// that are not completely interior.
//  The parent box is b, and the address of the child box
// is bxa.  The width of the child box is bwidth.
// The argument cutoff_dim determines the disposition of the
// box: if the new box contains topological faces of dimension
// cutoff_dim or lower, then the box is placed in low_dim_dest
// else it is placed in high_dim_dest.
// This routine splits the new box into connected components.


void
QMG::MG::ActiveBoxVec::create_subbox(const ActiveBox_As_Parent& b,
                                     const BoxAddress& bxa,
                                     int bwidth,
                                     int cutoff_dim,
                                     BoxFaceDag::Index parentind,
                                     int facetval,
                                     int facetdim,
                                     ActiveBoxVec& low_dim_dest,
                                     ActiveBoxVec& high_dim_dest,
                                     Create_Subbox_Workspace& wkspa) {


  if (wkspa.logstr.verbosity() >= 2) { ////
    wkspa.logstr.str() << "Creating subbox(es) of " << b << " cutoffdim = " << cutoff_dim << '\n';
  }
  // Create one box on the low-dimensional destination.

  wkspa.patchcomponum.reset(0);

  int lowest_patch_dim = low_dim_dest.create_new_subbox_(b, bxa,  bwidth, 
    parentind, facetval, facetdim, wkspa.patchcomponum, 0, wkspa);


  // If box is complete outside (doesn't exist) return.

  if (lowest_patch_dim == di_ + 1) {
    if (wkspa.logstr.verbosity() >= 2) { //// ////
      wkspa.logstr.str() << "No content -- returning\n";
    }
    low_dim_dest.delete_last_box_();
    return;
  }

  // Name the new box tmpb.

  ActiveBox tmpb = low_dim_dest.get_last_box_();

#ifdef DEBUGGING
  if (!tmpb.completely_interior() && tmpb.num_patch() == 0)
    throw_error(" box has no patches?");
#endif


  // If the box meets some patches, must split it into connected components.

  int num_components = mark_components(tmpb, wkspa);  
  if (wkspa.logstr.verbosity() >= 2) { ////
    wkspa.logstr.str() << "Lowest patch dim = " << lowest_patch_dim
      << " Num components = " << num_components << '\n';
  }

  // If the number of components is 1, and the lowest dimensional patch
  // agrees with the cutoff, then we are done -- postprocess the box and quit.

  if (num_components == 1 && lowest_patch_dim <= cutoff_dim) {
    low_dim_dest.post_process_box_(b, tmpb, lowest_patch_dim, wkspa.logstr);
    if (wkspa.logstr.verbosity() >= 2) ////
      wkspa.logstr.str() << "One box, low dim dest\n";
    if (wkspa.logstr.verbosity() >= 6)
      tmpb.full_dump2(wkspa.logstr.str(), wkspa.patchtable, wkspa.inc_table);
    return;
  }


  // Mark the patches to indicate which component they go with.
  // For each component,  figure out its lowest-dimensional
  // patch.

  wkspa.lowest_patchdim_of_component.resize(num_components);
  for (int jj = 0; jj < num_components; ++jj)
    wkspa.lowest_patchdim_of_component[jj] = di_;
  if (num_components == 1) {
    wkspa.lowest_patchdim_of_component[0] = lowest_patch_dim;
  }
  else {
    for (int relnum = 0; relnum < tmpb.num_patch(); ++relnum) {
      int c = wkspa.uf.get_label(relnum);
      PatchTable::Index patchind = tmpb.patch(relnum).patchind();
      wkspa.patchcomponum[patchind] = c;
      int ownerdim = wkspa.patchtable.owner(patchind).fdim();
      if (ownerdim < wkspa.lowest_patchdim_of_component[c])
        wkspa.lowest_patchdim_of_component[c] = ownerdim;
    }
  } 

  // Delete the box we just created and try again.

  if (wkspa.logstr.verbosity() >= 2) ////
    wkspa.logstr.str() << "Numcompo > 1 or dimension too high; deleting\n";
  low_dim_dest.delete_last_box_();  

  // Loop over connected components to find the ones that should go on the
  // high-dimensional destination.

  // Note: loop bound num_components2 is changed within the loop!
  
  int num_components2 = num_components;

  for (int this_componum = 0; this_componum < num_components2; ++this_componum) {

    // Find the lowest-dimensional patch of the component.

    int lowest_patch_dim_this_compo = wkspa.lowest_patchdim_of_component[this_componum];
    if (wkspa.logstr.verbosity() >= 2) { ////
      wkspa.logstr.str() << "Component " << this_componum << " out of "
        << num_components2 << " has lowest dimension " << lowest_patch_dim_this_compo << '\n';
    }
    
#ifdef DEBUGGING
    if (lowest_patch_dim_this_compo >= di_ && this_componum > 0)
      throw_error("No patch found in component?");
#endif

    if (lowest_patch_dim_this_compo <= cutoff_dim) {
      // Get here if the component is for the low-dimensional
      // destination.  Create the new box on the low-dimensional
      // destination and postprocess.
      
      int lpd0 = low_dim_dest.create_new_subbox_(b, bxa,  bwidth, 
        parentind, facetval, facetdim, wkspa.patchcomponum, this_componum, wkspa);
#ifdef DEBUGGING
      if (lpd0 > cutoff_dim)
        throw_error("create_new_subbox problem");
#endif
      ActiveBox tmpb2 = low_dim_dest.get_last_box_();
      low_dim_dest.post_process_box_(b, tmpb2, lowest_patch_dim_this_compo, wkspa.logstr);
      if (wkspa.logstr.verbosity() >= 6) {
        wkspa.logstr.str() << "Full dump of lo0 dim box just created:\n";
        tmpb2.full_dump2(wkspa.logstr.str(), wkspa.patchtable, wkspa.inc_table);
      }
    }
    else {
      // Get here if the component is for the high-dimensional
      // destination.
      

      // First create the box in the temporary vector.


      int lpd1;
      if (num_components2 > 1) {
        wkspa.tmp_boxvec.iwidth_ = low_dim_dest.iwidth_;
        wkspa.tmp_boxvec.expa_ = low_dim_dest.expa_;
        int lpd0 = wkspa.tmp_boxvec.create_new_subbox_(b, bxa, bwidth,
          parentind, facetval, facetdim, wkspa.patchcomponum, this_componum, wkspa);
#ifdef DEBUGGING
        if (lpd0 >= di_)
          throw_error("create_new_subbox problem");
#endif
        
      // Now create the box on the hi-dimensional destination.

        lpd1 = 
          high_dim_dest.create_new_subbox_(ActiveBox_As_Parent(wkspa.tmp_boxvec.get_last_box_()), 
           bxa, bwidth,
           parentind, facetval, facetdim, wkspa.patchcomponum, this_componum, wkspa);
        wkspa.tmp_boxvec.delete_last_box_();
      }
      else {
        lpd1 = 
          high_dim_dest.create_new_subbox_(b, bxa, bwidth, parentind, facetval,
          facetdim, wkspa.patchcomponum, this_componum, wkspa);
      }


      if (lpd1 > di_) {
        high_dim_dest.delete_last_box_();
        if (wkspa.logstr.verbosity() >= 2) ////
          wkspa.logstr.str() << "Hi dim component no created (no content)\n";
      }
      else {
        
        ActiveBox tmpb2 = high_dim_dest.get_last_box_();
        
        // Now we have to check whether tmpb2 is connected.
        // It might be disconnected because the higher-dimensional
        // vector has a small ex() size.
        
        int num_components3 = mark_components(tmpb2, wkspa);
#ifdef DEBUGGING
        if (num_components3 == 0)
          throw_error("Numcomponents3 = 0");
#endif
        
        if (wkspa.logstr.verbosity() >= 2) ////
          wkspa.logstr.str() << "Hi-dim component split; # components = " 
          << num_components3 << '\n';
        
        
        // If the high-dimensional box has more than one component,
        // we have to kill the box we just created and 
        // renumber the components again, and change the bound
        // for this loop (that we are in).
        // Nasty!
        
        if (num_components3 > 1) {
          int base_componum = num_components2;
          num_components2 += num_components3;
          wkspa.lowest_patchdim_of_component.resize(num_components2, di_);
          for (int relnum = 0; relnum < tmpb2.num_patch(); ++relnum) {
            PatchTable::Index patchind = tmpb2.patch(relnum).patchind();
            int c = base_componum + wkspa.uf.get_label(relnum);
            wkspa.patchcomponum[patchind] = c;
            int ownerdim = wkspa.patchtable.owner(patchind).fdim();
            if (ownerdim < wkspa.lowest_patchdim_of_component[c])
              wkspa.lowest_patchdim_of_component[c] = ownerdim;
          }
          high_dim_dest.delete_last_box_();
        }
        else  {
          high_dim_dest.post_process_box_(b, tmpb2, lowest_patch_dim_this_compo, wkspa.logstr);
          if (wkspa.logstr.verbosity() >= 6) {
            wkspa.logstr.str() << "Full dump of hi dim box just created:\n";
            tmpb2.full_dump2(wkspa.logstr.str(), wkspa.patchtable, wkspa.inc_table);
          }
        }
      }
    }
  }
}

// --------------------------------------------------------------
// After a box is created with create_new_subbox_, this routine
// must also be called for postprocessing the new box.  Call
// this routine only if the box is not discarded.
// This routine assumes the new box is at the end of its stack.


void
QMG::MG::ActiveBoxVec::post_process_box_(const ActiveBox_As_Parent& b,
                                         const ActiveBox& newb,
                                         int lowest_patch_dim,
                                         Logstream& logstr) {
  
  // Set weights for the last box created.
  
  // There are 3^{boxdim} weights.  No weights stored
  // for full-dimensional boxes.
  // Weights computed by calling the virtual function get_subbox_weight
  // in the parent box.



#ifdef NSF_DEMO
  QMG::Nsf_Demo::show_box(newb, lowest_patch_dim);
#endif

  ++ActiveBox::active_box_count_;

  Box_Data& newboxd = this -> boxvec_.back();
  int bdim = newboxd.bxa.dim();
  if (bdim == di_) return;

  int bdimpower3 = Base3::power3(bdim);

  vector<Weight_t>& wtvec = this -> weights_;
  int wtsz = wtvec.size();
  newboxd.weights_begin = wtsz;
  wtvec.resize(wtsz + bdimpower3, 0.0);
  for (Base3 ri = 0; ri < bdimpower3; ++ri) {
    wtvec[wtsz + ri] = b.get_subbox_weight(ri);
  }
  if (logstr.verbosity() >= 2) { ////
    if (logstr.verbosity() >= 4) {
      logstr.str() << "Subbox ";
      newb.full_dump(logstr.str());
      logstr.str() << " created ";
    }
    else {
      logstr.str() << "Subbox " << newb << " created ";
    }
    logstr.str() << " on dim " << lowest_patch_dim << " dest\n";
  }

  // Call the specialized postprocessing of the parent box.

  b.specialized_postprocess_subbox(newb, lowest_patch_dim);

}
  

// ------------------------------------------------------------------
// create_subbox_completely_interior
// This is similar to create_subbox, except the parent box is
// completely interior so there is no cutoffdim, and only one destvec,
// and no need to split components.

void
QMG::MG::ActiveBoxVec::create_subbox_completely_interior(const ActiveBox_As_Parent& b,
                                                         const BoxAddress& bxa,
                                                         int bwidth,
                                                         ActiveBoxVec& destvec,
                                                         BoxFaceDag::Index parentind,
                                                         int facetval,
                                                         int facetdim,
                                                         Logstream& logstr) {

#ifdef DEBUGGING
  if (!b.completely_interior()) 
    throw_error("push completely interior called on noninterior box");
#endif

  //  Curvature is verified in that case.

  Index newboxind = destvec.boxvec_.size();
  destvec.boxvec_.push_back(Box_Data());
  Box_Data& newboxd = destvec.boxvec_.back();
  newboxd.bxa = bxa;
  newboxd.link_parent = parentind;
  newboxd.link_parent_side = facetval;
  newboxd.link_parent_dim = facetdim;
  newboxd.box_completely_interior = true;
  newboxd.size_function_verified = b.size_function_verified();
  newboxd.has_close_face = false;
  newboxd.curvature_verified = true;
  newboxd.killed = false;
  newboxd.num_patch = b.owner();
  newboxd.is_front_surface = b.is_front_surface();
  ActiveBox tmpb = destvec.get_last_box_();
  destvec.post_process_box_(b, tmpb, di_, logstr);
}


// ------------------------------------------------------------------
// push_box_completely_interior
// This routine pushes a completely interior box onto a stack.

void
QMG::MG::ActiveBoxVec::push_box_completely_interior(const ActiveBox& b,
                                                    Logstream& logstr) {

#ifdef DEBUGGING
  if (!b.completely_interior()) 
    throw_error("push completely interior called on noninterior box");
#endif
 
  ActiveBox_As_Parent b_ex(b);

  create_subbox_completely_interior(b_ex, b.boxaddress(),
      b.iwidth(), *this, b.link_parent(),
      b.link_parent_side(), b.link_parent_dim(), logstr);
}



// ------------------------------------------------------------------
// push_box_shrink_ex
// This routine pushes a box onto a stack, but first shrinks its
// expanded neighborhood and therefore must recompute the patches it contains.

void
QMG::MG::ActiveBoxVec::push_box_shrink_ex(const ActiveBox& b,
                                          Create_Subbox_Workspace& wkspa) {
  

  if (wkspa.logstr.verbosity() >= 2)  ////
    wkspa.logstr.str() << "push box shrink ex on " << b << '\n';

  if (b.completely_interior()) {
    push_box_completely_interior(b, wkspa.logstr);
  }
  else {
    ActiveBox_As_Parent b_ex(b);
    create_subbox(b_ex, b.boxaddress(), b.iwidth(), di_ + 1, b.link_parent(),
      b.link_parent_side(), b.link_parent_dim(), *this, *this, wkspa);
  }
}



// ------------------------------------------------------------------
// Deletes the last box of a vector.  Doesn't delete the weights --
// assume weights aren't yet created. This should be called on a box
// for which create_new_box_ was called, but not postprocess_subbox_.

void QMG::MG::ActiveBoxVec::delete_last_box_() {
  Box_Data& lastboxd = this -> boxvec_.back();
  if (!lastboxd.box_completely_interior) {
    int patchvec_initsize = lastboxd.patchvec_begin;
    if (lastboxd.num_patch > 0) {
      int incvec_initsize = this -> patchvec_[patchvec_initsize].inc_begin;
      this -> incidencevec_.resize(incvec_initsize, IncidenceTable::Index());
    }
    this -> patchvec_.resize(patchvec_initsize, Patch_Data());
  }
  this -> boxvec_.pop_back();
}

