// ------------------------------------------------------------------
// gmcoarsentopo.cpp
//
// This file contains a function for coarsening the topology of
// a brep.
// ------------------------------------------------------------------
// Author: Stephen A. Vavasis
// Copyright (c) 1999 by Cornell University.  All rights reserved.
// 
// See the accompanying file 'Copyright' for authorship information,
// the terms of the license governing this software, and disclaimers
// concerning this software.
// ------------------------------------------------------------------
// This file is part of the QMG software.  
// Version 2.0 of QMG, release date September 3, 1999.
// ------------------------------------------------------------------


#include "qbrep_constr.h"
#include "qpatchtab.h"
#include "qfrontend.h"
#include "qunionfind.h"
#include "qancestor.h"
#include "qsmall.h"

#define GM_ROUTINE_NAME gmcoarsetopo
#define GM_ROUTINE_NAME_Q "gmcoarsetopo"


namespace {
  using namespace QMG;


  // Duplicate of stuff from qseparation.cpp.  Fix later.

  const int NUM_TEST_DIRECTIONS_2D = 4;
  const int NUM_CURVATURE_TEST_DIRECTION = 13;

  const Real test_direction_init[] = {0,1,0, 1,0,0, 1,1,0, 1,-1,0,
                                      0,0,1,  0,1,1, 1,0,1, 1,1,1,
                                      0,-1,1, -1,0,1, 1,-1,1, -1,1,1,
                                      -1,-1,1};

 
  
  

  // Find the deviations along all edge points.  Similar to gmchecknormals.cpp
  
  Real maxip[NUM_CURVATURE_TEST_DIRECTION];
  Real minip[NUM_CURVATURE_TEST_DIRECTION];
  Point test_directions[NUM_CURVATURE_TEST_DIRECTION];
  int num_test_directions;

  void init_test_directions(int di) {
    num_test_directions = (di == 2)?    
      NUM_TEST_DIRECTIONS_2D : NUM_CURVATURE_TEST_DIRECTION;
    for (int j1 = 0; j1 < num_test_directions; ++j1) {
      for (int k = 0; k < di; ++k) 
        test_directions[j1][k] = test_direction_init[j1 * 3 + k];
      test_directions[j1].normalize();
    }
  }
  
  void init_maxmin_() {         
    for (int j = 0; j < num_test_directions; ++j) {
      maxip[j] = -100;
      minip[j] = 100;
    }
  }

              // Update maxip and minip with normal_or_tan.
              
  bool update_maxmin_(const Point& normal_or_tan, double allowable_curvature) {
    for (int j = 0; j < num_test_directions; ++j) {
      Real ip = Point::inner_product(test_directions[j], normal_or_tan);
      if (ip > maxip[j])
        maxip[j] = ip;
      if (ip < minip[j])
        minip[j] = ip;
      if (maxip[j] - minip[j] > allowable_curvature) {
        return false;
      }
    }
    return true;
  }

  bool twovec_ok_for_curve_bound(const Point& nrml1,
    const Point& nrml2,
    double allowable_curvature) {
    init_maxmin_();
    update_maxmin_(nrml1, allowable_curvature);
    return update_maxmin_(nrml2, allowable_curvature);
  }


  // Copied from qpatchtab.cpp.  Fix later.

  bool lookup_orientation_[33];
  void init_lookup_orientation() {
    lookup_orientation_[0] = true;
    lookup_orientation_[2] = true;
    lookup_orientation_[3] = false;
    lookup_orientation_[10] = false;
    lookup_orientation_[12] = false;
    lookup_orientation_[13] = true;
    lookup_orientation_[20] = false;
    lookup_orientation_[21] = true;
    lookup_orientation_[23] = true;
    lookup_orientation_[30] = true;
    lookup_orientation_[31] = false;
    lookup_orientation_[32] = false;
  }


  // New iterator: loop over descendents of a brep face, generally with
  // with repeats.

  class Fspec_loop_over_descendents : public Brep::Face_Spec {
  private:
    const Brep& b_;
    vector<Brep::Face_Spec> stack_;
    void set_this_();
    bool done_;
    // No copying, no assignment
    Fspec_loop_over_descendents(const Fspec_loop_over_descendents& o) :
    b_(o.b_) { }
    void operator=(const Fspec_loop_over_descendents&) { }
    Brep::Face_Spec& operator=(const Brep::Face_Spec& fspec) {return *this;}
  public:
    Fspec_loop_over_descendents(const Brep& b) :
        b_(b)  {set_this_();}
    void init(const Brep::Face_Spec& fspec);
    bool notdone();
    void operator++();
    ~Fspec_loop_over_descendents() {}
  };

  void Fspec_loop_over_descendents::set_this_() {
    if (stack_.size() == 0) {
      done_ = true;
      fdim_ = -1;
      faceind_ = -1;     
    } 
    else {
      fdim_ = stack_.back().fdim();
      faceind_ = stack_.back().faceind();
    }
  }


  void Fspec_loop_over_descendents::init(const Brep::Face_Spec& fspec) {
    stack_.resize(1,fspec);
    done_ = false;
    set_this_();
  }

  bool Fspec_loop_over_descendents::notdone() {
    set_this_();
    return !done_;
  }

  void Fspec_loop_over_descendents::operator++() {
    Brep::Face_Spec stacktop = stack_.back();
    stack_.pop_back();
    for (Brep::Face_Spec_Loop_Over_Face_Subfaces subfspec(b_, stacktop);
    subfspec.notdone();
    ++subfspec) {
      stack_.push_back(subfspec);
    }
  }
  

  Brep_Under_Construction coarsen_topo(const Brep& b, 
    double allowable_curvature) {


    // std::ofstream ctlog("ctlog");

    int di = b.embedded_dim();
    Point::InitStaticData(di);

    int gdim = b.gdim();
    if (gdim != di)
      throw_error("Brep must be full dimensional");
    ostringstream nullstr;

    using QMG::MG::Logstream;
    using QMG::MG::PatchTable;
    using QMG::MG::PatchTableIndexMap;
    using QMG::MG::PatchInBrepAddress;
    using QMG::MG::My_bool;

    Logstream null_log(nullstr,0);
    // Make a map to determine which parent-child patch relationships.
    PatchTable patchtable(b, null_log, -1, -1);
    PatchTableIndexMap<My_bool> already_seen(patchtable, My_bool(false));
    Brep::Ancestor_Lookup anc_lookup(b);
  
    init_test_directions(di);
    init_lookup_orientation();

    
    map<PatchInBrepAddress, PatchTable::Index> lookup_patchind; 
    {
      for (PatchTable::Index patchind = 0; 
      patchind < patchtable.numpatch(); 
      ++patchind) {
        Brep::Face_Spec owner = patchtable.owner(patchind);
        if (patchtable.gdim(patchind) != owner.fdim()) continue;
        Brep::PatchIndex patchind1 = patchtable.orig_index(patchind);
        lookup_patchind[PatchInBrepAddress(owner,patchind1)] = patchind;
      }
    }

    // Labeling to keep track of faces to delete.

    Brep::Face_Labeling<bool> ok_to_delete(b, true);

    // Union-find for merging faces.

    UnionFind uf[MAXDIM + 1];

    // Not ok to delete faces of dimension di or di - 1.

    {
      for (int fdim = di; fdim >= di - 1; --fdim) {
        uf[fdim].initialize(b.level_size(fdim));
        for (Brep::Face_Spec_Loop_Over_Faces_Of_Dim fspec(b, fdim);
        fspec.notdone();
        ++fspec) {
          ok_to_delete(fspec) = false;
        }
      }
    }

    vector<int> parent1_parentlabels;
    vector<int> parent2_parentlabels;

    // Loop over dimensions of faces that might be ok to delete.

    {
      for (int fdim = di - 2; fdim >= 0; --fdim) {
        uf[fdim].initialize(b.level_size(fdim));

        for (Brep::Face_Spec_Loop_Over_Faces_Of_Dim fspec(b, fdim);
        fspec.notdone();
        ++fspec) {

          // ctlog << " considering whether to delete " << fspec << " name " 
            // << b.face_name(fspec) << '\n';

          // Don't delete a face with prop-val pairs.

          int numpv = 0;
          {
            for (Brep::Loop_over_propvals_of_face pvloop1(b, fspec);
            pvloop1.notdone();
            ++pvloop1) {
              ++numpv;
            }
          }


          if (numpv > 0) { 
            ok_to_delete(fspec) = false;

            // ctlog << " not deleting; face has property-values\n";
            continue;
          }     
          
          // Count the number of parents

          int anc_count = 0;
          Brep::Face_Spec parent1, parent2;

          for (Brep::Ancestor_Lookup::Loop_over_ancestors ancloop(anc_lookup, fspec);
          ancloop.notdone();
          ++ancloop) {
            Brep::Face_Spec afspec = ancloop.ancestor_fspec();
            if (afspec.fdim() == fdim + 1 && !ok_to_delete(afspec)) {
              if (anc_count == 0)
                parent1 = afspec;
              if (ancloop.occurrence_count() > 1)
                ++anc_count;
              if (anc_count > 0)
                parent2 = afspec;
              ++anc_count;
            }
          }

          // If the number of parents is not 0 and not 2, can't delete,

          if (anc_count != 0 && anc_count != 2) {
            ok_to_delete(fspec) = false;
            // ctlog << " not deleting; # ancestors = " << anc_count << "\n";
            continue;
          }
          if (anc_count == 2) {

            // ctlog << " two parents " << parent1 << " (" << b.face_name(parent1)
             // << ") and " << parent2 << " (" << b.face_name(parent2) << ")\n";

            // Check that the two parents have the same p-v-pairs.
            
            for (Brep::Loop_over_propvals_of_face pvloop(b, parent1);
            pvloop.notdone();
            ++pvloop) {
              string otherval = b.face_lookup_prop_val(parent2, pvloop.prop());
              if (pvloop.val() != otherval) {
                // ctlog << " not deleting; parents differ in property " << pvloop.prop()
                //  << " values = " << pvloop.val() << "," << otherval << '\n';
                ok_to_delete(fspec) = false;
                break;
              }
            }
            
            if (!ok_to_delete(fspec)) continue;
            for (Brep::Loop_over_propvals_of_face pvloop2(b, parent2);
            pvloop2.notdone();
            ++pvloop2) {
              string otherval = b.face_lookup_prop_val(parent1, pvloop2.prop());
              if (pvloop2.val() != otherval) {
                //ctlog << " not deleting; parents differ in property " << pvloop2.prop()
                //  << " values = " << pvloop2.val() << "," << otherval << '\n';
                ok_to_delete(fspec) = false;
                break;
              }
            }
            if (!ok_to_delete(fspec)) continue;

            // Check that the two parents have the same parents.

            parent1_parentlabels.resize(0);

            for (Brep::Ancestor_Lookup::Loop_over_ancestors ancloop1(anc_lookup, parent1);
            ancloop1.notdone();
            ++ancloop1) {
              Brep::Face_Spec afspec = ancloop1.ancestor_fspec();
              if (afspec.fdim() != fdim + 2) continue;
              if (ok_to_delete(afspec)) continue;
              for (int j = 0; j < ancloop1.occurrence_count(); ++j) 
                parent1_parentlabels.push_back(uf[fdim+1].get_label(afspec.faceind()));
            }

            parent2_parentlabels.resize(0);
            
            for (Brep::Ancestor_Lookup::Loop_over_ancestors ancloop2(anc_lookup, parent2);
            ancloop2.notdone();
            ++ancloop2) {
              Brep::Face_Spec afspec = ancloop2.ancestor_fspec();
              if (afspec.fdim() != fdim + 2) continue;
              if (ok_to_delete(afspec)) continue;
              for (int j = 0; j < ancloop2.occurrence_count(); ++j) 
                parent2_parentlabels.push_back(uf[fdim+1].get_label(afspec.faceind()));
            }
/*
            {
              ctlog << " parent labels of parent1 = ";
              for (int jj = 0; jj < parent1_parentlabels.size(); ++jj) {
                ctlog << parent1_parentlabels[jj] << ' ';
              }
              ctlog << " of parent2 = ";
              for (int jj2 = 0; jj2 < parent2_parentlabels.size(); ++jj2) {
                ctlog << parent2_parentlabels[jj2] << ' ';
              }
              ctlog << '\n';
            }
            */


            if (parent1_parentlabels.size() != parent2_parentlabels.size()) {
              ok_to_delete(fspec) = false;
              // ctlog << " not deleting; different length parent labels\n";
              continue;
            }

            sort(parent1_parentlabels.begin(), parent1_parentlabels.end());
            sort(parent2_parentlabels.begin(), parent2_parentlabels.end());
            for (int jj = 0; jj < parent1_parentlabels.size(); ++jj) {
              if (parent1_parentlabels[jj] != parent2_parentlabels[jj]) {
                ok_to_delete(fspec) = false;
                // ctlog << " not deleting; different parents\n";
                break;
              }
            }
            if (!ok_to_delete(fspec)) continue;


            // If the number of ancestors is 2, loop over patches of
            // the face to see if the parent patches have sufficiently
            // close normals..


            for (Brep::Loop_over_patches_of_face patchloop(b, fspec);
            patchloop.notdone();
            ++patchloop) {
              Brep::PatchIndex ind = patchloop.index();
              PatchTable::Index patchind = 
                lookup_patchind[PatchInBrepAddress(fspec,ind)];
              int numends = fdim + 1;
              for (int whichend = 0; whichend < numends; ++whichend) {
                // Find the two ancestors of this patch.
                int pt_anc_count = 0;
                Point prev_normal_or_tan;
                Point param;
                param[0] = static_cast<double>(whichend);
                Point paramdirec;
                paramdirec[0] = 1.0;
                for (PatchTable::Loop_over_parents parloop(patchtable, patchind);
                parloop.notdone();
                ++parloop) {
                  PatchTable::Index ppatchind = parloop.parent_index();
                  Brep::PatchIndex thisparentind = patchtable.orig_index(ppatchind);
                  Brep::Face_Spec thisparent = patchtable.owner(ppatchind);
                  if (patchtable.gdim(ppatchind) != fdim + 1 ||
                    thisparent.fdim() != fdim + 1)
                    throw_error("patch loop inconsistency");
                  if (ok_to_delete(thisparent)) continue;
                  if (thisparent != parent1 && thisparent != parent2)
                    throw_error("patch loop problem 2");
                  // Lift the coordinates, and compute the normal.
                  // Do this differently depending on the type of patch.
                  int lift0 = parloop.lift_index(0);
                  int lift1 = (fdim == 1)? parloop.lift_index(1) : 0;
                  Point liftparam = patchtable.lift_parameters(param, 
                    patchind, parloop.listind());
                  Point normal_or_tan;
                  if (fdim == 0) {
                    normal_or_tan = b.patchmath(thisparent, thisparentind).
                      direc_deriv(liftparam, paramdirec);
                    if (normal_or_tan.l2norm() == 0)
                      throw_error("Vanishing tangent on curve");
                    normal_or_tan.normalize();
                  }
                  else {
                    normal_or_tan = b.patchmath(thisparent, thisparentind).                        
                      normal(liftparam);
                  }

                  // Flip sign if necessary.

                  if (lookup_orientation_[lift0*10+lift1] != (pt_anc_count == 0)) {
                    for (int j = 0; j < di; ++j)
                      normal_or_tan[j] = -normal_or_tan[j];
                  }
                  if (pt_anc_count == 0)
                    prev_normal_or_tan = normal_or_tan;
                  else {
                    if (!twovec_ok_for_curve_bound(prev_normal_or_tan, 
                      normal_or_tan, allowable_curvature))
                      ok_to_delete(fspec) = false;
                  }
                  /*
                  ctlog << " normal on patch " << patchtable.orig_index(patchind) 
                    <<  " end " << whichend
                    << " parent patch " << thisparent << ':' << thisparentind
                    << " (count = " << pt_anc_count << ") is"
                    << normal_or_tan << '\n';
                   
                  if (!ok_to_delete(fspec))
                    ctlog << " not ok to delete --curvature\n";
                    */

                  ++pt_anc_count;
                }
                if (pt_anc_count != 2) {
                  throw_error("Patch table problem 3");
                }
                if (!ok_to_delete(fspec)) break;
              }
              if (!ok_to_delete(fspec)) break;
            }
          }
          else { // anc_count == 0

            // ctlog << " no parents\n";
            if (fdim == 0 && di == 3) {
              // In this special case, must loop over patches 
              // containing this node to check if they
              // line up.

              init_maxmin_();
              PatchTable::Index patchind = 
                lookup_patchind[PatchInBrepAddress(fspec,0)];
              int parcount1 = 0;
              int parcount2 = 0;
              for (PatchTable::Loop_over_parents parloop(patchtable, patchind);
              parloop.notdone();
              ++parloop) {
                ++parcount1;
                if (parloop.listind() > 0) continue;
                PatchTable::Index firstcurve = parloop.parent_index();
                PatchTable::Index prev2D = -1;
                PatchTable::Index currcurve = firstcurve;
                int currendp = parloop.lift_index(0);
                already_seen.reset(My_bool(false));

                while (!already_seen[currcurve]) {
/*
                  ctlog << " looping over parent patch " << currcurve
                    << " dim " << patchtable.gdim(currcurve) << " owner "
                    << patchtable.owner(currcurve);
                   
                  if (patchtable.owner(currcurve).fdim() == patchtable.gdim(currcurve))
                    ctlog << ':' << patchtable.orig_index(currcurve);
                  ctlog << '\n';
                  */

                  // Loop over parents of currcurve; find a parent
                  // not equal to prev2D;
                  
                  already_seen[currcurve] = true;
                  int parcount3 = 0;
                  PatchTable::Index nextcurve = -1;
                  PatchTable::Index nextprev2d;
                  int nextendp;
                  for (PatchTable::Loop_over_parents parloop2(patchtable, currcurve);
                  parloop2.notdone();
                  ++parloop2) {
                    ++parcount3;
                    PatchTable::Index gpindex = parloop2.parent_index();
                    if (gpindex == prev2D)
                      continue;
                    for (PatchTable::Loop_over_children chloop(patchtable, gpindex);
                    chloop.notdone();
                    ++chloop) {
                      PatchTable::Index uncleindex = chloop.child_index();
                      if (uncleindex == currcurve) continue;
                      for (PatchTable::Loop_over_children chloop2(patchtable, uncleindex);
                      chloop2.notdone();
                      ++chloop2) {
                        if (chloop2.child_index() == patchind) {
                          nextprev2d = gpindex;
                          nextcurve = uncleindex;
                          nextendp = chloop2.lift_index(0);
                        }
                      }
                    }
                  }
                  if (parcount3 != 2) {
                    ok_to_delete(fspec) = false;
                    // ctlog << " parent patch does not have two parents; not OK to delete\n"; 
                    break;
                  }
                  if (nextcurve < 0)
                    throw_error("Nextcurv not found?");
                  Point paramcoord, direc;
                  paramcoord[0] = static_cast<double>(currendp);
                  direc[0] = (currendp)? -1.0 : 1.0;
                  Point tan1 = patchtable.patchmath(currcurve).
                    direc_deriv(paramcoord, direc);
                  paramcoord[0] = static_cast<double>(nextendp);
                  direc[0] = (nextendp)? -1.0 : 1.0;
                  Point tan2 = patchtable.patchmath(nextcurve).
                    direc_deriv(paramcoord, direc);
                  Point nrml = Point::cross_product(tan1, tan2);
                  if (nrml.l2norm() == 0)
                    throw_error("Zero normal?");
                  nrml.normalize();
                  // ctlog << " crossprod with nextcurve " << nextcurve << " tan yields " << nrml
                  //  << '\n';
                  bool ok1 = update_maxmin_(nrml, allowable_curvature);
                  if (!ok1) {
                    ok_to_delete(fspec) = false;
                    // ctlog << " curvature not met; not OK to delete\n";
                    break;
                  }
                  currcurve = nextcurve;
                  currendp = nextendp;
                  prev2D = nextprev2d;
                  ++parcount2;
                }
                if (currcurve != firstcurve) {
                  // ctlog << " didn't return to first patch currcurve = " << currcurve
                  //  << " firstcurve = " << firstcurve << "; not ok to delete\n";
                  ok_to_delete(fspec) = false;
                }
                if (!ok_to_delete(fspec)) break;
              }
              if (parcount1 != parcount2)
                ok_to_delete(fspec) = false;
            }
          }
          if (!ok_to_delete(fspec)) continue;
          // ctlog << "OK to delete " << fspec;
          if (anc_count == 2) {
            uf[fdim + 1].merge(parent1.faceind(), parent2.faceind());
            // ctlog << " merged " << parent1 << "," << parent2 << " under label "
            //  << uf[fdim+1].get_label(parent1.faceind());
          }
          // ctlog << '\n';
        }
      }
    }



    // uf gives numbers to indicate which faces are merged.
    // Now make a map from old faces to new.
    // We renumber the uf labels to skip over deleted faces.


    Brep::Face_Labeling<Brep::FaceIndex> old_to_new_map(b,0);
    vector<int> new_level_size(di + 1);

    {
      vector<Brep::FaceIndex> renumuf;
      for (int fdim = 0; fdim <= di; ++fdim) {
        int numlabels = uf[fdim].pack_labels();
        renumuf.resize(numlabels);
        for (int j = 0; j < numlabels; ++j)
          renumuf[j] = -1;
        Brep::FaceIndex nextnewfaceind = 0;
        for (Brep::Face_Spec_Loop_Over_Faces_Of_Dim fspec(b, fdim);
        fspec.notdone();
        ++fspec) {
          if (ok_to_delete(fspec)) continue;
          Brep::FaceIndex oldfaceind = fspec.faceind();
          int lab1 = uf[fdim].get_label(oldfaceind);
          Brep::FaceIndex lab2 = renumuf[lab1];
          if (lab2 < 0) {
            renumuf[lab1] = nextnewfaceind;
            lab2 = nextnewfaceind;
            ++nextnewfaceind;
          }
          old_to_new_map(fspec) = lab2;
          // ctlog << "old " << fspec << " to new " << Brep::Face_Spec(fdim, lab2) << '\n';
        }
        new_level_size[fdim] = nextnewfaceind;
      }
    }

    // Now invert old_to_new_map with a multimap new_to_old_map.

    typedef multimap<Brep::Face_Spec, Brep::FaceIndex> FspecMM;
    FspecMM new_to_old_map;
    {
      for (int fdim = 0; fdim <= di; ++fdim) {
        for (Brep::FaceIndex oldfaceind = 0; oldfaceind < b.level_size(fdim); ++oldfaceind) {
          Brep::Face_Spec oldfspec(fdim, oldfaceind);
          if (ok_to_delete(oldfspec)) continue;
          Brep::Face_Spec newfspec(fdim, old_to_new_map(oldfspec));
          new_to_old_map.insert(pair<Brep::Face_Spec, Brep::FaceIndex>
            (newfspec, oldfaceind));
          // ctlog << "new " << newfspec << " to old " << oldfspec << '\n';
        }
      }
    }

    Brep::Face_Labeling<int> occurrence_count(b,0);
    Brep::Face_Labeling<int> occurrence_count_last_update(b,0);
    Brep::Face_Labeling<int> descendent_last_update(b,0);
    Fspec_loop_over_descendents desc_fspec(b);

    
    Brep_Under_Construction newb(di, di);

    Fspec_loop_over_descendents desc_fspec_newb(newb);
    int passcount = 0;


    // Copy property-values from b to newb.

    Brep_Under_Construction::Propval_inserter pvi(newb);
    for (Brep::Loop_over_propvals_of_brep pvloop(b);
    pvloop.notdone();
    ++pvloop) {
      if (compare_nocase(pvloop.prop(), Brep::global_id_propname()) == 0) {
        ostringstream os;
        os << "(" << pvloop.val() << "_with_faces_merged_allowable_curvature="
          << allowable_curvature << ")";
        pvi.insert_propval(Brep::global_id_propname(), os.str());
      }
      else {
        pvi.insert_propval(pvloop.prop(), pvloop.val());
      }
    }

    vector<Real> realcoord(di);

    Brep_Under_Construction::Control_point_inserter cpi(newb);

    // Copy the control points.
    for (Brep::ControlPointIndex cpind = 0; 
    cpind < b.num_control_points();
    ++cpind) {
      for (int d = 0; d < di; ++d)
        realcoord[d] = b.control_point(cpind, d);
      cpi.insert_control_point(realcoord);
    }


    vector<Brep::ControlPointIndex> controlpts;
    // Insert the faces
    for (int fdim = 0; fdim <= di; ++fdim) {
      Brep_Under_Construction::Top_face_inserter tfi(newb, fdim);
      int newfacecount = new_level_size[fdim];
      for (Brep::FaceIndex newfaceind = 0; newfaceind < newfacecount; ++newfaceind) {
        ++passcount;
        Brep::Face_Spec newfspec(fdim, newfaceind);
        pair<FspecMM::iterator, FspecMM::iterator> rval = 
          new_to_old_map.equal_range(newfspec);
        FspecMM::iterator startloop = rval.first;
        FspecMM::iterator endloop = rval.second;
        int mergersize = 0;
        {
          for (FspecMM::iterator it = startloop; it != endloop; ++it) {
            ++mergersize;
            if (ok_to_delete(Brep::Face_Spec(fdim, it -> second)))
              throw_error("Deleted face part of merge?");
          }
        }
      
        if (mergersize == 0)
          throw_error("New face has no antecedents?");
        string facename;
        Brep::Face_Spec oldfspec(fdim, startloop -> second);
        if (mergersize == 1) {
          facename = b.face_name(oldfspec);
        }
        else {
          facename = "Merge_of";
          for (FspecMM::iterator it = startloop; it != endloop; ++it) {
            facename += "_$_";
            facename += b.face_name(Brep::Face_Spec(fdim, it -> second));
          }
        }
        
        tfi.insert_top_face(facename);
        // ctlog << "created new face " << newfspec << " named " << facename << '\n';
        
        // Insert the face pv pairs.    
        Brep_Under_Construction::Top_face_propval_inserter tfpvi(newb, newfspec);
        for (Brep::Loop_over_propvals_of_face pvloop2(b, oldfspec);
        pvloop2.notdone();
        ++pvloop2) {
          tfpvi.insert_propval(pvloop2.prop(), pvloop2.val());
        }
      
    

      
        // Add child faces.
        // Insert internal boundaries.

        Brep_Under_Construction::Top_face_bound_inserter tfbi(newb, newfspec);
        Brep_Under_Construction::Top_face_ib_inserter tfii(newb, newfspec);


        // Track down the old subfaces of all faces being merged.
        {
          for (FspecMM::iterator it = startloop; it != endloop; ++it) {
            Brep::Face_Spec oldfspec(fdim, it -> second);
            for (Brep::Face_Spec_Loop_Over_Face_Subfaces subfspec(b, oldfspec);
            subfspec.notdone();
            ++subfspec) {
              if (occurrence_count_last_update(subfspec) < passcount) {
                occurrence_count_last_update(subfspec) = passcount;
                occurrence_count(subfspec) = 0;
              }
              ++occurrence_count(subfspec);
              // ctlog << " occurrence count for old fspec " << oldfspec << " child " << subfspec
              //  << " incremented to " << occurrence_count(subfspec) << '\n';
            }
          }
        }

        // Now insert the subfaces.

        {
          for (FspecMM::iterator it = startloop; it != endloop; ++it) {
            Brep::Face_Spec oldfspec(fdim, it -> second);
            for (Brep::Face_Spec_Loop_Over_Face_Subfaces subfspec(b, oldfspec);
            subfspec.notdone();
            ++subfspec) {
              if (ok_to_delete(subfspec)) continue;
              int occ_cnt = occurrence_count(subfspec);
              if (occ_cnt == 0) continue;

              // Find what face this face maps to, and then check
              // that the occurrence counts are all consistent.

              Brep::Face_Spec newsubfspec(subfspec.fdim(), old_to_new_map(subfspec));
              pair<FspecMM::iterator, FspecMM::iterator> rval =
                new_to_old_map.equal_range(newsubfspec);
              // ctlog << "inserting subfspec " << newsubfspec << " from oldfspec "
              //  << oldfspec << " subface " << subfspec << " occcnt = " << occ_cnt << '\n';
              for (FspecMM::iterator it2 = rval.first; it2 != rval.second; ++it2) {
                Brep::Face_Spec oldsubfspec2(subfspec.fdim(), it2 -> second);
                if (occurrence_count(oldsubfspec2) != occ_cnt)
                  throw_error("Inconsistent occurrence of child faces");
                occurrence_count(oldsubfspec2) = 0;
              }
              if (newsubfspec.fdim() == fdim - 1) {
                int icount = 1;
                if (occ_cnt % 2 == 0)
                  icount = 2;
                for (int i = 0; i < icount; ++i) {
                  tfbi.insert_boundary(newsubfspec, subfspec.orientation());
                }
              }
              else {
                tfii.insert_ib(newsubfspec);
              }
            }
          }
        }

        // Now comes a tricky section: because of deleted faces, we may have
        // some new internal boundaries to insert.

        // Mark the descendents of the face.

        {          
          for (FspecMM::iterator it = startloop; it != endloop; ++it) {
            Brep::Face_Spec oldfspec(fdim, it -> second);
            for (desc_fspec.init(oldfspec); desc_fspec.notdone(); ++desc_fspec) {
              descendent_last_update(desc_fspec) = passcount;
            }
          }
        }

        // Unmark the descendents that occur in the new the face.

        {
          for (desc_fspec_newb.init(newfspec); desc_fspec_newb.notdone(); ++desc_fspec_newb) {
            pair<FspecMM::iterator,FspecMM::iterator> rval = 
              new_to_old_map.equal_range(desc_fspec_newb);
            for (FspecMM::iterator it = rval.first; it != rval.second; ++it) {
              Brep::Face_Spec oldsubfspec(desc_fspec_newb.fdim(), it -> second);
              descendent_last_update(oldsubfspec) = passcount - 1;
            }
          }
        }

        // Find descendents still marked and add them as internal boundaries.
   

        {          
          for (FspecMM::iterator it = startloop; it != endloop; ++it) {
            Brep::Face_Spec oldfspec(fdim, it -> second);
            for (desc_fspec.init(oldfspec); desc_fspec.notdone(); ++desc_fspec) {
              if (ok_to_delete(desc_fspec)) continue;
              if (descendent_last_update(desc_fspec) == passcount) {
                if (desc_fspec.fdim() >= fdim - 2)
                  throw_error("Some inconsistency with ib's");
                Brep::Face_Spec new_descfspec(desc_fspec.fdim(), old_to_new_map(desc_fspec));
                tfii.insert_ib(new_descfspec);
                pair<FspecMM::iterator,FspecMM::iterator> rval = 
                  new_to_old_map.equal_range(new_descfspec);
                for (FspecMM::iterator it2 = rval.first; it2 != rval.second; ++it2) {
                  descendent_last_update(Brep::Face_Spec(new_descfspec.fdim(),
                    it2 -> second)) = passcount - 1;
                }
              }
            }
          }
        }

        // Now insert the patches.

                
        Brep_Under_Construction::Top_face_patch_inserter 
          tfpi(newb, newfspec);
        {

          for (FspecMM::iterator it = startloop; it != endloop; ++it) {
            Brep::Face_Spec oldfspec(fdim, it -> second);
            for (Brep::Loop_over_patches_of_face ptloop(b, oldfspec);
            ptloop.notdone();
            ++ptloop) {
              
              PatchType ptype = ptloop.ptype();
              int deg1 = ptloop.degree1();
              int ncp, deg2;
              
              if (fdim == 0) {
                ncp = 1;
              }
              else if (fdim == 1) {
                ncp = deg1  + 1;
              }
              else  { // if (gdim == 2) 
                if (ptype == BEZIER_TRIANGLE) {
                  ncp = (deg1 + 1) * (deg1 + 2) / 2;
                } 
                else { // ptype == BEZIER_QUAD
                  deg2 = ptloop.degree2();
                  ncp = (deg1 + 1) * (deg2 + 1);
                }
              }            
              controlpts.resize(ncp);
              for (int j = 0; j < ncp; ++j)
                controlpts[j] = ptloop.control_point(j);                
              tfpi.insert_patch(deg1, deg2, ptype, controlpts);
            }
          }
        }
      }
    }
    return newb;
  }


  static void worker(const QMG::FrontEnd::ArgValType& argvalhandle,
    QMG::FrontEnd::ReturnValType& returnvalhandle,
    QMG::FrontEnd::Interpreter& interp) {
    
    using namespace QMG;
    using namespace QMG::FrontEnd;
    
    argvalhandle.verify_nargin(2, 2, GM_ROUTINE_NAME_Q);
    returnvalhandle.verify_nargout(1, 1, GM_ROUTINE_NAME_Q);
    double allow_curve = argvalhandle.get_double(1);
    Brep_Under_Construction newb = 
      coarsen_topo(argvalhandle.get_brep(0), allow_curve);
    returnvalhandle.put_brep(0, newb);
  }
}

#include "gm_entrypoint.cpp"

