// ------------------------------------------------------------------
// qboxstack.cpp
//
// This file contains member functions qboxstack.h.  These are classes
// for boxes in the mesh generator, and vectors of boxes.
// ------------------------------------------------------------------
// Author: Stephen A. Vavasis
// Copyright (c) 1999 by Cornell University.  All rights reserved.
// 
// See the accompanying file 'Copyright' for authorship information,
// the terms of the license governing this software, and disclaimers
// concerning this software.
// ------------------------------------------------------------------
// This file is part of the QMG software.  
// Version 2.0 of QMG, release date September 3, 1999.
// ------------------------------------------------------------------


#ifdef _MSC_VER
#if _MSC_VER == 1200
#include "qmatvec.h"
#endif
#endif

#include "qboxstack.h"


namespace QMG {
  namespace MG {
    using namespace QMG;

    
    //global function in this file


    void boxstack_init_static_data(int di); 
  }
}

// File scope functions.

namespace {

  using namespace QMG;
  using namespace QMG::MG;

#ifdef DEBUGGING
  ostream* debug_ostream;  
  bool dump_everything;
  int create_subbox_invocation_count;
#endif



  // ------------------------------------------------------------------
  // Routine to compute a random point in the unit octahedron.
  // Selection is not uniform -- I don't know an easy way to do this uniformly.
  Point select_rand_point_in_unit_octahedron(int di) {
    Point p;
    {
      for (int i = 0; i < di; ++i)
        p[i] = random_real() * 2.0 - 1.0;
    }
    Real t = random_real();
    {
      for (int i = 0; i < di; ++i)
        t += fabs(p[i]);
    }
    {
      for (int i = 0; i < di; ++i)
        p[i] /= t;
    }
    return p;
  }

}


// ------------------------------------------------------------------
// contains_point_in_ex
// This routine determines whether a point is contained in the expanded
// version of a box.

bool 
QMG::MG::ActiveBox::contains_point_in_ex(const Point& pt, 
                                         Real param_uncertainty,
                                         Real scfac) const { 

  Real l1dist = 0.0;
  Real mult = BoxAddress::multiplier();
  Real rwidth = mult / (1<<iwidth());
  for (int actualdim = 0; actualdim < di_; ++actualdim) {
    Real lower = ((Real) this -> boxaddress().lowerleft(actualdim)) * mult / 
      (1 << MAXLEV) + BoxAddress::global_coordinate_origin(actualdim);
    Real upper = lower + ( (flatdim(actualdim)) ? 0 : rwidth);
    if (pt[actualdim] < lower)
      l1dist += lower - pt[actualdim];
    else if (pt[actualdim] > upper)
      l1dist += pt[actualdim] - upper;
  }
  return (l1dist <= rwidth * expa() - scfac * param_uncertainty);
}


// ------------------------------------------------------------------
// This routine selects a random point in the expanded neighborhood
// of a box.

QMG::MG::Point 
QMG::MG::ActiveBox::select_rand_point() const {

  Real mult = BoxAddress::multiplier();
  Real rwidth = mult / (1<<iwidth());
  Point pt =  select_rand_point_in_unit_octahedron(di_);

#ifdef DEBUGGING
  if (dump_everything) {
    *debug_ostream << " in select rand initalpt = " << pt << "expa = " << expa() 
      << " mult = " << mult << " rwidth = " << rwidth << '\n';
  }
#endif

  for (int actualdim = 0; actualdim < di_; ++actualdim) {
    Real lower = ((Real) this -> boxaddress().lowerleft(actualdim)) * mult / 
      (1 << MAXLEV) + BoxAddress::global_coordinate_origin(actualdim);
    Real thisw =  (flatdim(actualdim)) ? 0.0: rwidth;
    pt[actualdim] = pt[actualdim] * rwidth * expa() + 
      random_real() * thisw + lower;
  }
  return pt;
}

// ------------------------------------------------------------------
// This routine computes the real coordinates of the lower left corner
// of the box.

QMG::MG::Point
QMG::MG::BoxAddress::real_lowerleft() const {
  Point p;
  for (int ii = 0; ii < di_; ++ii) {
    p[ii] = lowerleft_[ii] * multiplier_ / (1 << MAXLEV) +
      global_coordinate_origin_[ii];
  }
  return p;
}



// ------------------------------------------------------------------
// This routine computes the real coordinates of the lower left
// corner of a face of a box.

QMG::MG::Point 
QMG::MG::ActiveBox::face_real_lowerleft(Base3 rindex) const {
  int bdim = this -> dim();
  int actual_dim = -1;
  BoxAddress bxa = this -> boxaddress();
  int iwidth = this -> iwidth();

  for (int ii = 0; ii < bdim; ++ii) {
    while (this -> boxaddress().flatdim(++actual_dim));
    if (rindex[ii] == 1)
      bxa.lowerleft().increment(actual_dim,iwidth);
  }
  return bxa.real_lowerleft();
}



// ------------------------------------------------------------------
// This routine computes the weight of a box face.  Use the static
// weights for fulldimensional boxes, properly adjusted if the box
// lies at the boundary of the top box.  Else use the weights stored
// in the box vector.

QMG::MG::Weight_t
QMG::MG::ActiveBox::weight(Base3 rindex) const {
#ifdef RANGECHECK
  if ((UInt32) rindex >= Base3::power3(this -> dim())) {
    throw_error("Arg out of range in weight()");
  }
#endif
  if (this -> dim() == di_) {
    Weight_t w = fulldim_weights_[rindex];
    UInt32 rightmost = (1 << MAXLEV) - (1 << (MAXLEV - this -> iwidth()));
    for (int j = 0; j < di_; ++j) {
      int ri = rindex[j];
      if (ri == 0 && this -> boxaddress().lowerleft(j) == 0)
        w *= 2;
      else if (ri == 1 && this -> boxaddress().lowerleft(j) == rightmost)
        w *= 2;
    }
    return w;

  }
  else {
    return (this -> owner_).weights_[(this -> owner_).boxvec_[this -> idx_].weights_begin + 
      (UInt32) rindex];
  }
}


// ------------------------------------------------------------------
// Initialize static data in this file.


void QMG::MG::ActiveBox::init_static_data(int di) {
  di_ = di;
  active_box_count_ = 0;
  for (Base3 rindex = 0; rindex < Base3::power3(di); ++rindex) {
    int d = rindex.dim();
    fulldim_weights_[rindex] = 1.0 / ((Weight_t) (1 << (di - d)));
  }
}



// ------------------------------------------------------------------
// Returns the weight of a box face as a nonconstant reference (so
// that it can be updated).  Valid only for non-full-dimensional boxes.

QMG::MG::Weight_t& 
QMG::MG::NonConstActiveBox::weight_nonconst(Base3 rindex) {
#ifdef RANGECHECK
  if ((UInt32) rindex >= Base3::power3(this -> dim())) {
    throw_error("Arg out of range in weight_nonconst()");
  }
#endif
#ifdef DEBUGGING
  if (this -> dim() == di_) {
    throw_error("Attempt to set weight for fulldimensional box");
  }
#endif
  return (const_cast<ActiveBoxVec&>(this -> owner_)).
    weights_[(this -> owner_).boxvec_[this -> idx_].weights_begin + 
    (UInt32) rindex];
}

 

// ------------------------------------------------------------------
// Loop_over_vertices: class for looping over the vertices of a box.

void 
QMG::MG::ActiveBox::Loop_over_vertices::operator++() {
  this -> ind_++;
  int reldim = 0;
  for (int j = 0; j < this -> di_; ++j) {
    int bit = (this -> bflatdim_[j])? 0 : ((ind_ >> (reldim++)) & 1);
    this -> intcoord_[j] = this -> baseintcoord_[j];
    this -> realcoord_[j] = this -> baserealcoord_[j];
    if (bit) {
      this -> intcoord_.increment(j, this -> bwidth_);
      this -> realcoord_[j] += brealwidth_;
    }
  }
}

Base3 
QMG::MG::ActiveBox::Loop_over_vertices::rindex() const {
  UInt32 rind = 0;
  for (int j = 0; j < this -> bdim_; ++j) {
    if ((this -> ind_ >> j) & 1)
      rind += Base3::power3(j);
  }
  return (Base3) rind;
}


// ------------------------------------------------------------------
// Loop_over_incidences
// Class for looping over incidences of a box.



QMG::MG::ActiveBox::Loop_over_incidences::Loop_over_incidences(const ActiveBox& ab,
                                                               const PatchTable& patchtable,
                                                               const Brep::Face_Spec& fspec) :

ab_(ab),
patchtable_(patchtable),
fspec_(fspec) {
  for (this -> relnum_ = 0; 
  this -> relnum_ < (this -> ab_).num_patch(); 
  ++(this -> relnum_)) {
    PatchTable::Index p = (this -> ab_).patch(this -> relnum_).patchind();
    if ((this ->patchtable_).owner(p) == (this -> fspec_) && 
      (this -> ab_).patch(this -> relnum_).num_incidence() > 0)
      break;
  }
  this -> ii_ = 0;
}


void 
QMG::MG::ActiveBox::Loop_over_incidences::operator++() {
  this -> ii_++;
  if (this -> ii_ < (this -> ab_).patch(this -> relnum_).num_incidence())
    return;
  for (++(this ->relnum_); 
  (this -> relnum_) < (this -> ab_).num_patch(); 
  ++(this -> relnum_)) {
    PatchTable::Index p = (this -> ab_).patch((this -> relnum_)).patchind();
    if ((this -> patchtable_).owner(p) == (this -> fspec_) &&
      (this -> ab_).patch(this -> relnum_).num_incidence() > 0)
      break;
  }
  this -> ii_ = 0;
}

QMG::MG::ActiveBox 
QMG::MG::ActiveBoxVec::get_last_box_() const {
  return ActiveBox(*this, this -> boxvec_.size() - 1);
}



 

void QMG::MG::boxstack_init_static_data(int di) {
  ActiveBoxVec::init_static_data(di);
  ActiveBox::init_static_data(di);
  BoxAddress::init_static_data(di);
}




#ifdef NSF_DEMO
namespace QMG {
  namespace Nsf_Demo {
    using namespace QMG;
    using namespace QMG::MG;
    extern void show_box(const ActiveBox& b, int code);
  }
}
#endif


// ------------------------------------------------------------------
// push_top_box
// This routine computes the initial (top) box of the quadtree and
// pushes it into an ActiveBoxVec.  This is how mesh generation
// is initiated.

void QMG::MG::ActiveBoxVec::push_top_box(const Brep& g,
                                         IncidenceTable& inc_table,
                                         Create_Subbox_Workspace& wkspa) {

#ifdef DEBUGGING
  debug_ostream = &wkspa.logstr.str();
#endif

  int di = g.embedded_dim();
  int pvstart = patchvec_.size();
  double expa = this -> expa();

  Mask flatdim = 0;
  IntCoord intcoord;
  Point midpointreal;

  Patch_Data ptd;
  for (PatchTable::Index patchind = 0; patchind < wkspa.patchtable.numpatch(); ++patchind) {
    int ivstart = incidencevec_.size();
    inc_table.get_incidence(flatdim, intcoord, midpointreal, wkspa.patchtable, patchind, 
      incidencevec_, wkspa.logstr, wkspa.scfac, wkspa.tol, wkspa.rworkspace, wkspa.output_stack, wkspa.gui);
    ptd.patchnum = patchind;
    ptd.inc_begin = ivstart;
    ptd.num_inc = incidencevec_.size() - ptd.inc_begin;
    patchvec_.push_back(ptd);
  }
  
  Box_Data boxd;
  boxd.patchvec_begin = pvstart;
  boxd.num_patch = patchvec_.size() - pvstart;
  boxd.weights_begin = weights_.size();
  boxd.link_parent = -1;
  boxd.box_completely_interior = false;
  boxd.size_function_verified = false;
  boxd.curvature_verified = false;
  boxd.killed = false;
  boxd.has_close_face = false;
  boxd.is_front_surface = 2;
  boxd.bxa.flatdim() = 0;
  for (int j = 0; j < di; ++j)
    boxd.bxa.lowerleft()[j] = 0;
  Index newboxind = boxvec_.size();
  boxvec_.push_back(boxd);

  
  Point lowerbound;
  Point upperbound;
  Point realcoord;
  {
    for (int j = 0; j < di; ++j) {
      lowerbound[j] = BIG_REAL;
      upperbound[j] = -BIG_REAL;
    }
  }
  for (int i1 = 0; i1 < incidencevec_.size(); ++i1) {
    IncidenceTable::Index inc = incidencevec_[i1];
    realcoord = inc_table.real_coord(inc);
    for (int j = 0; j < di; ++j) {
       if (realcoord[j] < lowerbound[j])
         lowerbound[j] = realcoord[j];
       if (realcoord[j] > upperbound[j])
         upperbound[j] = realcoord[j];
    }
  }
  Real maxwidth = 0.0;
  {
    for (int j = 0; j < di; ++j) {
      if (upperbound[j] - lowerbound[j] > maxwidth)
        maxwidth = upperbound[j] - lowerbound[j];
    }
  }

  if (maxwidth <= 0.0)
    throw_error("Some problem computing maxwidth");
  {
    for (int j = 0; j < di; ++j) {
      lowerbound[j] -= (maxwidth * (static_cast<double>(j)) / 20);
    }
  }
  maxwidth += .1 * (di - 1) * maxwidth;
  if (wkspa.logstr.verbosity() >= 1) {
    wkspa.logstr.str() << " Global coordinate origin = " << lowerbound 
      << " width = " << maxwidth << '\n';
  }

  
  BoxAddress::init_global_box_coords(lowerbound, maxwidth);
  if (wkspa.logstr.verbosity() >= 2) {
    if (wkspa.logstr.verbosity() >= 4) {
      wkspa.logstr.str() << "Top box ";
      this -> getbox(newboxind).full_dump(wkspa.logstr.str());
      wkspa.logstr.str() << '\n';
    }
    else {
      wkspa.logstr.str() << "Top box " << this -> getbox(newboxind) << '\n';
    }
  }
  ++ActiveBox::active_box_count_;
#ifdef NSF_DEMO
  QMG::Nsf_Demo::show_box(this -> getbox(newboxind), 0);
#endif
}



void QMG::MG::ActiveBoxVec::killbox(Index i, Logstream& logstr) {
  if (logstr.verbosity() >= 2) {
    const ActiveBox& bx = this -> getbox(i);
    logstr.str() << "deleting box " << bx << '\n';
  }
  boxvec_[i].killed = true;
  ++num_killed_;
  --ActiveBox::active_box_count_;
}


QMG::MG::ActiveBoxVec::~ActiveBoxVec() {
#ifndef _MSC_VER
#ifdef DEBUGGING
#ifdef EXCEPTIONS_ENABLED
  if (!std::uncaught_exception()) {
#endif
    for (int s = 0; s < boxvec_.size(); ++s) {
      if (!boxvec_[s].killed)
        throw_error("Live box remains in destroyed active box vec");
    }
#ifdef EXCEPTIONS_ENABLED
  }
#endif
#endif
#endif
}


void QMG::MG::ActiveBoxVec::dump(ostream& os) const {
}

void QMG::MG::ActiveBoxVec::debug(const PatchTable& patchtable,
                                  const IncidenceTable& inc_table,
                                  const BoxFaceDag& dag) const {
  int liveboxcount = 0;
  int patchstart = 0;
  int incstart = 0;
  int weightstart = 0;
  for (Index bxind = 0; bxind < boxvec_.size(); ++bxind) {
    const Box_Data& boxdat = boxvec_[bxind];
    if (!boxdat.killed) ++liveboxcount;
    if (boxdat.link_parent < -1 || boxdat.link_parent >= dag.size())
      throw_error("Link parent out of range");
    if (!boxdat.box_completely_interior) {
      if (boxdat.patchvec_begin != patchstart)
        throw_error("Patchstart not contiguous");
      if (boxdat.num_patch < 0)
        throw_error("Numpatch < 0");
      if (boxdat.patchvec_begin + boxdat.num_patch > patchvec_.size())
        throw_error("Patches run out of range");
      for (int relnum = 0; relnum < boxdat.num_patch; ++relnum) {
        const Patch_Data& ptc = patchvec_[patchstart + relnum];
        PatchTable::Index p = ptc.patchnum;
        if (p < 0 || p >= patchtable.numpatch())
          throw_error("Patch index out of range");
        if (ptc.inc_begin != incstart)
          throw_error("Incstart not contiguous");
        if (ptc.num_inc < 0)
          throw_error("Numinc < 0");
        if (incstart + ptc.num_inc > incidencevec_.size())
          throw_error("Incidences run out of range");
        for (int rpi = 0; rpi < ptc.num_inc; ++rpi) {
          IncidenceTable::Index inc = incidencevec_[incstart + rpi];
          if (inc_table.patchind(inc) != p)
            throw_error("Incidence does not belong to patch");
        }
        incstart += ptc.num_inc;
      }
      patchstart += boxdat.num_patch;
    }
    int bdim = boxdat.bxa.flatdim().dim();
    if (bdim < di_) {
      if (boxdat.weights_begin != weightstart)
        throw_error("Weight start not contiguous");
      int numweight = Base3::power3(bdim);
      if (weightstart + numweight > weights_.size())
        throw_error("Weight start out of range");
      weightstart += Base3::power3(bdim);
    }
  }
  if (liveboxcount != boxvec_.size() - num_killed_)
    throw_error("Live box count is off");
}




void QMG::MG::ActiveBox::full_dump(ostream& os) const {
  os << *this << " ";
  Real rw = this -> real_width();
  Point ll = this -> real_lowerleft();
  Mask fd = this -> flatdim();
  for (int j = 0; j < di_; ++j) {
    os << '[' << ll[j] << ',' << ( fd[j]? ll[j] : ll[j]+rw ) << ']';
    if (j < di_ - 1) os << '*';
  }
  os << " expa = " << this -> expa() << " szok = ";
  if (size_function_verified()) 
    os << "t";
  else
    os << "f";
  os << " curvok = ";
  if (curvature_verified()) 
    os << "t";
  else
    os << "f";
  os << "\n patches = ";
  for (int relnum = 0; relnum < this -> num_patch(); ++relnum) {
    PatchTable::Index patchind = this -> patch(relnum).patchind();
    os << "pat#" << patchind << "#  (";
    for (int j = 0; j < this -> patch(relnum).num_incidence(); ++j) {
      IncidenceTable::Index inc = this -> patch(relnum).incidence(j);
      os << "incd#" << IncidenceTable::seqno(inc) << "# ";
    }
    os << ") ";
  }
  os << '\n';
  os << "weights = ";
  int bdim = this -> dim();
  for (Base3 ri = 0; ri < Base3::power3(bdim); ++ri) {
    ri.output(os, bdim);
    os << ':' << this -> weight(ri) << ' ';
  }
  os << '\n';
}


void QMG::MG::ActiveBox::full_dump2(ostream& os,
                                    const PatchTable& patchtable,
                                    const IncidenceTable& inctable) const {
  os << *this << " ";
  Real rw = this -> real_width();
  Point ll = this -> real_lowerleft();
  Mask fd = this -> flatdim();
  for (int j = 0; j < di_; ++j) {
    os << '[' << ll[j] << ',' << ( fd[j]? ll[j] : ll[j]+rw ) << ']';
    if (j < di_ - 1) os << '*';
  }
  os << " expa = " << this -> expa() << " szok = ";
  if (size_function_verified()) 
    os << "t";
  else
    os << "f";
  os << " curvok = ";
  if (curvature_verified()) 
    os << "t";
  else
    os << "f";
  os << this -> is_front_surface();
  os << "\n patches = ";
  for (int relnum = 0; relnum < this -> num_patch(); ++relnum) {
    PatchTable::Index patchind = this -> patch(relnum).patchind();
    os << "pat#" << patchind << "#  (";
    os << "orig " << patchtable.owner(patchind) << " # "
       << patchtable.orig_index(patchind) << ") ";
    for (int j = 0; j < this -> patch(relnum).num_incidence(); ++j) {
      IncidenceTable::Index inc = this -> patch(relnum).incidence(j);
      os << "incd#" << IncidenceTable::seqno(inc) << "# ";
      os << "with quadtree entity ";
      for (int jj = 0; jj < di_; ++jj) {
        if (inc -> first.flatdim[jj])
          os << std::hex << inc -> first.intcoord[jj] << std::dec << " ";
        else
          os << "-- ";
      }
    }
    os << ") ";
    os << '\n';
  }
  os << '\n';
  os << "weights = ";
  int bdim = this -> dim();
  for (Base3 ri = 0; ri < Base3::power3(bdim); ++ri) {
    ri.output(os, bdim);
    os << ':' << this -> weight(ri) << ' ';
  }
  os << '\n';
}



QMG::ostream& operator<<(ostream& os, const QMG::MG::ActiveBox& ab) {
  int di = ab.embedded_dim();
  os << "(ab" << ab.iwidth() << ';';
  for (int ii = 0; ii < di; ++ii)
    os << ( (ab.flatdim(ii))? 1 : 0 );
  os << ';';
  for (int jj = 0; jj < di; ++jj) {
    os << std::hex << ab.boxaddress().lowerleft(jj);
    if (jj < di - 1)
      os << ',';
  }
  if (ab.completely_interior())
    os << '%';
  os << std::dec << ')';
  return os;
}


void QMG::MG::Brep_Orbits::debug(const PatchTable& patchtable,
    const IncidenceTable& inc_table,
    const BoxFaceDag& dag,
    const Brep::Face_Spec& fspec) const {
  int level = 0;
  ActiveBoxVec* nextvecp = this -> find_next_nonempty_level(fspec, level);
  while (nextvecp) {
    level = nextvecp -> iwidth();
    nextvecp -> debug(patchtable, inc_table, dag);
    ++level;
    nextvecp = this -> find_next_nonempty_level(fspec, level);
  }
}

void QMG::MG::Brep_Orbits::dump(ostream& os,
                                const Brep::Face_Spec& fspec) const {
  int level = 0;
  ActiveBoxVec* nextvecp = this -> find_next_nonempty_level(fspec, level);
  while (nextvecp) {
    level = nextvecp -> iwidth();
    os << "Beginning dump of level " << level << "\n";
    for (ActiveBoxVec::Loop_over_boxes loop(*nextvecp); loop.notdone(); ++loop) {
      ActiveBox bx = loop.getbox();
      bx.full_dump(os);
    }
    ++level;
    nextvecp = this -> find_next_nonempty_level(fspec, level);
  }
  os << std::flush;
}





QMG::Real QMG::MG::BoxAddress::multiplier_;
QMG::MG::Point QMG::MG::BoxAddress::global_coordinate_origin_;
int QMG::MG::BoxAddress::di_;
int QMG::MG::ActiveBox::di_;
QMG::MG::Weight_t QMG::MG::ActiveBox::fulldim_weights_[MAXDIM_3EXP];
int QMG::MG::ActiveBox::active_box_count_;
int QMG::MG::ActiveBoxVec::di_;
