// ------------------------------------------------------------------
// qinctab.cpp
//
// This file contains member functions for IncidenceTable.
// ------------------------------------------------------------------
// Author: Stephen A. Vavasis
// Copyright (c) 1999 by Cornell University.  All rights reserved.
// 
// See the accompanying file 'Copyright' for authorship information,
// the terms of the license governing this software, and disclaimers
// concerning this software.
// ------------------------------------------------------------------
// This file is part of the QMG software.  
// Version 2.0 of QMG, release date September 3, 1999.
// ------------------------------------------------------------------

#ifdef _MSC_VER
#if _MSC_VER == 1200
#include "qmatvec.h"
#endif
#endif

#include "qinctab.h"
#include "qmatvec.h"
#include "qmg_gui.h"
#include <cstddef>


#ifdef DEBUGGING
namespace {
  int get_incidence_invocation_count;
}
#endif


// ------------------------------------------------------------------
// PatchMath_for_curve_on_triangle
// This routine allows us to treat a curve that sits on a patch as
// a separate Bezier curve.  The patch is described by two parameters
// alpha and beta which are the slopes of the curve.

class QMG::MG::IncidenceTable::PatchMath_for_curve_on_triangle : public QMG::PatchMath {
private:
  const PatchTable& patchtab_;
  PatchTable::Index patchind_;
  Real alpha_,beta_;
  int deg_;
  Real control_point_coord_(int cpnum, int d) const;
public:
  PatchMath_for_curve_on_triangle(const PatchTable& patchtab,
    PatchTable::Index patchind,
    Real alpha,
    Real beta) :
  PatchMath(BEZIER_TRIANGLE, IncidenceTable::get_patch_degree1_(patchtab, patchind), 
    0, 1, 3, true),
    patchtab_(patchtab),
    patchind_(patchind),
    alpha_(alpha),
    beta_(beta),
    deg_(IncidenceTable::get_patch_degree1_(patchtab,patchind))
  { }
  ~PatchMath_for_curve_on_triangle() { }
};


// ------------------------------------------------------------------
// The control point coordinates of the curve are interpolated
// from the patch's control points.
    
    
QMG::Real 
QMG::MG::IncidenceTable::PatchMath_for_curve_on_triangle::
control_point_coord_(int cpnum, int d) const {
  // Should use decasteljau here, but no obvious way to allocate workspace
  Real coef = 1.0;
  for (int jj = 0; jj < cpnum; ++jj)
    coef *= beta_;
  Real t = 0.0;
  int liftcp = (deg_ - cpnum) * (deg_ - cpnum + 1) / 2;
  int incliftcp = deg_ - cpnum + 2;
  for (int j = 0; j <= cpnum; ++j) {
    t += IncidenceTable::get_patch_control_point_coord_(patchtab_, patchind_, liftcp, d) 
      * coef;
    coef *= alpha_ / beta_ * (cpnum - j) / (j + 1);
    liftcp += incliftcp;
    ++incliftcp;
  }
  return t;
}




// ------------------------------------------------------------------
// PatchMath_for_curve_on_quad
// This routine allows us to treat a curve that sits on a quad patch as
// a separate Bezier curve.  The curve has a fixed v and varying u.
// The fixed value of v is parameter beta.

class QMG::MG::IncidenceTable::PatchMath_for_curve_on_quad : public QMG::PatchMath {
private:
  const PatchTable& patchtab_;
  PatchTable::Index patchind_;
  Real beta_;
  int deg1_, deg2_;
  Real control_point_coord_(int cpnum, int d) const;
public:
  PatchMath_for_curve_on_quad(const PatchTable& patchtab,
    PatchTable::Index patchind,
    Real beta) :
  PatchMath(BEZIER_TRIANGLE, IncidenceTable::get_patch_degree1_(patchtab, patchind), 
    0, 1, 3, true),
    patchtab_(patchtab),
    patchind_(patchind),
    beta_(beta),
    deg1_(IncidenceTable::get_patch_degree1_(patchtab,patchind)),    
    deg2_(IncidenceTable::get_patch_degree2_(patchtab,patchind))
  { }
  ~PatchMath_for_curve_on_quad() { }
};

    
// The control points of the curve are interpolated from
// control points of the patch.
    
QMG::Real 
QMG::MG::IncidenceTable::PatchMath_for_curve_on_quad::
control_point_coord_(int cpnum, int d) const {
  // Should use decasteljau here, but no obvious way to allocate workspace
  Real coef = 1.0;
  for (int jj = 0; jj < deg2_; ++jj)
    coef *= (1 - beta_);
  Real t = 0.0;
  int liftcp = cpnum;
  for (int j = 0; j <= deg2_; ++j) {
    t += IncidenceTable::get_patch_control_point_coord_(patchtab_, patchind_, liftcp, d)
      * coef;
    coef *= beta_ / (1 - beta_) * (deg2_ - j) / (j + 1);
    liftcp += deg1_ + 1;
  }
  return t;
}


// Functions to expose otherwise hidden patchtable fields.

int
QMG::MG::IncidenceTable::get_patch_degree1_(const PatchTable& patchtable,
                                           PatchTable::Index patchind) {
  return patchtable.patches_[patchind].degree1;
}


int
QMG::MG::IncidenceTable::get_patch_degree2_(const PatchTable& patchtable,
                                           PatchTable::Index patchind) {
  return patchtable.patches_[patchind].degree2;
}



QMG::PatchType
QMG::MG::IncidenceTable::get_patch_type_(const PatchTable& patchtable,
                                           PatchTable::Index patchind) {
  return patchtable.patches_[patchind].ptype;
}



QMG::Real 
QMG::MG::IncidenceTable::get_patch_control_point_coord_(const PatchTable& patchtable,
                                                    PatchTable::Index patchind,
                                                    int cpnum,
                                                    int d)  {
  int cpstart = patchtable.patches_[patchind].control_point_start;
  Brep::ControlPointIndex cpi = patchtable.control_point_list_[cpstart + cpnum];
  return patchtable.brep_.control_point(cpi, d);
}



bool QMG::MG::IncidenceTable::Key::operator<(const Key& other) const {
  if (patchind < other.patchind)
    return true;
  if (patchind > other.patchind)
    return false;
  if ((UInt32) flatdim < (UInt32) other.flatdim)
    return true;
  if ((UInt32) flatdim > (UInt32) other.flatdim)
    return false;
  int sgndif = IntCoord::compare(intcoord, other.intcoord);
  if (sgndif < 0)
    return true;
  if (sgndif > 0)
    return false;
  if (seqno < other.seqno)
    return true;
  return false;
}


int QMG::MG::IncidenceTable::compare_index(const Index& i1, const Index& i2) {
  return (i1 -> first.seqno) - (i2 -> first.seqno);
}


QMG::MG::IncidenceTable::IncidenceTable() : sequence_number_(0) {
#ifdef DEBUGGING
  get_incidence_invocation_count = 0;
#endif
}

// This the main routine for IncidenceTable: it computes
// the incidence (intersection) between a box entity
// and a brep entity.  The box entity is described by
// flatdim and midpoint coord, and midpointcoord_real
// is midpoint_coord converted to real coordinates.
// The box entity is patchind.
// The result is placed in output_vec.


void     
QMG::MG::IncidenceTable::get_incidence(Mask flatdim, 
                                       const IntCoord& midpointcoord, 
                                       const Point& midpointcoord_real,
                                       const PatchTable& patchtable,
                                       PatchTable::Index patchind,
                                       vector<Index>& output_vec, // output variable
                                       Logstream& logstr, 
                                       double scfac,
                                       double tol,
                                       PatchMath::Workspace& workspace,
                                       vector<PatchMath::IntersectRecord>& os_workspace,
                                       Meshgen_gui& gui) {
#ifdef DEBUGGING
  ++get_incidence_invocation_count;
  int invoc_count = get_incidence_invocation_count;
#endif


  int di = patchtable.embedded_dim();
  int patchdim = patchtable.patches_[patchind].gdim;
  int affinedim = flatdim.dim();
  const int GUI_UPDATE_FREQUENCY = 20;

  // If the dimension is too low, no intersection.

  if (affinedim + patchdim < di)
    return;

  // Look up the incidence; see if already computed.

  Key key;
  key.intcoord = midpointcoord;
  key.flatdim = flatdim;
  key.patchind = patchind;
  key.seqno = 0;
  for (int j = 0; j < di; ++j) {
    if (!key.flatdim[j]) {
      key.intcoord[j] = 0;
    }
  }
  typedef map<Key,Data> TTTYPE;
  typedef TTTYPE::iterator TTTYPE_IT;

  Index rangestart = table_.lower_bound(key);
  key.seqno = sequence_number_;
  Index rangeend = table_.upper_bound(key);
  if (rangestart != rangeend) {
    // Check if the first record indicates "no intersection."
    if (rangestart -> second.param_coord[0] == code_for_no_intersection()) 
      return;

    // Otherwise return all the incidence records.

    for (Index i1 = rangestart; i1 != rangeend; ++i1) {
      output_vec.push_back(i1);
    }
    return;
  }

  // Get here if no intersection is already in the table.  In this case, compute
  // the intersection.

  os_workspace.resize(0, PatchMath::IntersectRecord());

  if (affinedim == 0) {
    throw_error("get_incidence reached with affine dim = 0");
  }
  else if (affinedim < di) {
    if (affinedim == 1) {
      Real matrixspace[MAXDIM*MAXDIM];
      Matrix lineeq(di - 1, di, matrixspace);
      Point linerhs;
      int matrixrow = 0;
      for (int j = 0; j < di; ++j) {
        if (flatdim[j]) {
          for (int j1 = 0; j1 < di; ++j1)
            lineeq(matrixrow,j1) = 0.0;
          lineeq(matrixrow,j) = 1.0;
          linerhs[matrixrow++] = midpointcoord_real[j];
        }
      }
      if (matrixrow != di - 1)
        throw_error("Inconsistent dimension");
      patchtable.patchmath(patchind).intersect_line(os_workspace, workspace,
        lineeq, linerhs, scfac, tol, true);
      
    }
    else { // affinedim == 2, di == 3
      
      Point plane_normal;
      Real plane_rhs;
      for (int j = 0; j < di; ++j) {
        if (flatdim[j]) {
          plane_normal[j] = 1.0;
          plane_rhs = midpointcoord_real[j];
        }
        else
          plane_normal[j] = 0.0;
      }
      if (patchdim == 1) {
        patchtable.patchmath(patchind).intersect_plane(os_workspace, workspace,
          plane_normal, plane_rhs, scfac, tol, true);
      }
      else { // patchdim == 2, affinedim == 2, di == 3 (the hardest case)

        // Loop over extremepoints with the right orientation.
        
        Key key2;
        key2.flatdim = 0;
        key2.patchind = patchind;
        for (int j = 0; j < di; ++j) {
          key2.intcoord[j] = 0;
        }
        key2.seqno = 0;
        rangestart = table_.lower_bound(key2);
        key2.seqno = sequence_number_;
        rangeend = table_.upper_bound(key2);
        if (rangestart != rangeend &&
          rangestart -> second.param_coord[0] != code_for_no_intersection()) {
          for (Index i1 = rangestart; i1 != rangeend; ++i1) { 
            if (i1 -> second.match_flatdim != flatdim) continue;   
            if (get_patch_type_(patchtable, patchind) == BEZIER_TRIANGLE) {       
              Real alpha = i1 -> second.param_coord[0];
              Real beta = i1 -> second.param_coord[1];
              Real sum = alpha + beta;
              if (sum == 0) {
                alpha = .5;
                beta = .5;
              }
              else {
                alpha /= sum;
                beta /= sum;
              }
              if (beta == 0.0) beta = MACHINE_EPS;
              PatchMath_for_curve_on_triangle pm(patchtable, patchind, alpha, beta);
              int os_init_size = os_workspace.size();
              pm.intersect_plane(os_workspace, workspace, plane_normal, 
                plane_rhs, scfac, tol, true);
              for (int j = os_init_size; j < os_workspace.size(); ++j) {
                Real t = os_workspace[j].paramcoord[0];
                os_workspace[j].paramcoord[0] = alpha * t;
                os_workspace[j].paramcoord[1] = beta * t;
                os_workspace[j].realcoord = 
                  patchtable.patchmath(patchind).get_real_point(os_workspace[j].paramcoord);
              }
            }
            else { 
              Real beta = i1 -> second.param_coord[1];
              if (beta >= 1.0) 
                beta = 1.0 - 2 * MACHINE_EPS;
              PatchMath_for_curve_on_quad pm(patchtable, patchind, beta);
              int os_init_size = os_workspace.size();
              pm.intersect_plane(os_workspace, workspace, plane_normal,
                plane_rhs,scfac, tol, true);
              
              for (int j = os_init_size; j < os_workspace.size(); ++j) {
                Real t = os_workspace[j].paramcoord[0];
                os_workspace[j].paramcoord[1] = beta;
                os_workspace[j].realcoord = 
                  patchtable.patchmath(patchind).get_real_point(os_workspace[j].paramcoord);
              }
            }
          }
        }
      }
    }
   
    
    
    int sz1 = (os_workspace.size() == 0)? 1 : os_workspace.size();
    
    for (int osi = 0; osi < sz1; ++osi) {
      Data d;
      bool intersection_exists = (os_workspace.size() > 0);
      if (intersection_exists) {
        d.param_coord[0] = os_workspace[osi].paramcoord[0];
        d.param_coord[1] = os_workspace[osi].paramcoord[1];
        d.real_coord = os_workspace[osi].realcoord;
        d.param_uncertainty = os_workspace[osi].param_uncertainty;
        for (int j = 0; j < di; ++j) {
          if (flatdim[j]) {
#ifdef DEBUGGING
            if (fabs(d.real_coord[j] - midpointcoord_real[j]) > 
              scfac * d.param_uncertainty * 10) {
              throw_error("Computed intersection point is too far from plane");
            }
#endif
            d.real_coord[j] = midpointcoord_real[j];
          }
        }
      }
      else {
        d.param_coord[0] = code_for_no_intersection();
      }
      
      key.seqno = sequence_number_++;
      if (sequence_number_ % GUI_UPDATE_FREQUENCY == 0) {
        gui.check_update();
      }
      pair<TTTYPE_IT,bool> rval =  table_.insert(pair<Key,Data>(key,d)); 
      if (!rval.second)
        throw_error("Insert failed unexpectedly");
      Index newit = rval.first;
      if (logstr.verbosity() >= 5) {
        logstr.str() << "New incd#" << key.seqno << "# between ";         
        for (int ii = 0; ii < di; ++ii)
          logstr.str() << ( (flatdim[ii])? 1 : 0 );
        logstr.str() << ';';
        for (int jj = 0; jj < di; ++jj) {
          if (flatdim[jj])
            logstr.str() << std::hex << midpointcoord[jj];
          else
            logstr.str() << '*';
          if (jj < di - 1)
            logstr.str() << ',';
        } 
        logstr.str()  <<  std::dec  << " = ";
        for (int j3 = 0; j3 < di; ++j3) {
          if (flatdim[j3])
            logstr.str() << midpointcoord_real[j3] << " ";
          else
            logstr.str() << "* ";
        }
        logstr.str() << "\n and pat#" << patchind << "#\n";
        if (intersection_exists) {
          logstr.str() <<" intersection real coordinates = " << d.real_coord 
            << " param coordinates = ";
          for (int ll = 0; ll < patchdim; ++ll)
            logstr.str() << d.param_coord[ll] <<  " ";
          logstr.str()  << " uncertainty = " << d.param_uncertainty << "\n";
        }
        else {
          logstr.str() << " no intersection \n";
        }
      }
      if (intersection_exists)
        output_vec.push_back(newit);
    }
  }
    
  else {  //get here if affine_dim=di  In this case, compute extreme points
    // in several directions.

    

    // Positive odd integer -- controls how many directions to compute.

    const int NUMDIREC = 5;

    int numdirec2 = (NUMDIREC - 1) / 2;
    double dnumdirec2 = static_cast<double>(numdirec2);
    int p3ub = (di == 3)? NUMDIREC : 1;

    Point splitdirec;
    {
      for (int j = 0; j < di; ++j)
        splitdirec[j] = sqrt(static_cast<double>(j) + 2.0);
    }

    Point direc;
    int direci[3];

    for (direci[0] = 0; direci[0] < NUMDIREC; ++(direci[0])) {
      for (direci[1] = 0; direci[1] < NUMDIREC; ++(direci[1])) {
        for (direci[2] = 0; direci[2] < p3ub; ++(direci[2])) {
          Mask matchmask = 0;
          {
            int ex0ct = 0;
            int bdrct = 0;
            int nonzpos;
            for (int jj = 0; jj < di; ++jj) {
              direc[jj] = (direci[jj] - dnumdirec2) / dnumdirec2;
              if (direci[jj] == 0 || direci[jj] == NUMDIREC - 1)
                ++bdrct;
              if (direci[jj] == NUMDIREC / 2)
                ++ex0ct;
              else
                nonzpos = jj;
            }
            
            if (bdrct == 0)
              continue;
            if (ex0ct == di - 1)
              matchmask = (1 << nonzpos);
          }
          if (Point::inner_product(splitdirec, direc) <= 0) continue;
          os_workspace.resize(0, PatchMath::IntersectRecord());
          patchtable.patchmath(patchind).get_extreme_points(os_workspace,
            workspace, direc, scfac, tol);
          for (int osi = 0; osi < os_workspace.size(); ++osi) {
            if (os_workspace[osi].degen) {
              if (logstr.verbosity() >= 5) {
                logstr.str() << "Incidence between interior and pat#" << patchind
                  << " at real coord " << os_workspace[osi].realcoord 
                  << "\n  param coordinates = ";
                for (int ll = 0; ll < patchdim; ++ll)
                  logstr.str() << os_workspace[osi].paramcoord[ll] <<  " ";
                logstr.str() << " uncertainty = " 
                  << os_workspace[osi].param_uncertainty << " not created because of degen\n";
              }
              continue;
            }
            Data d;
            d.param_coord[0] = os_workspace[osi].paramcoord[0];
            d.param_coord[1] = os_workspace[osi].paramcoord[1];
            d.real_coord = os_workspace[osi].realcoord;
            d.param_uncertainty = os_workspace[osi].param_uncertainty;
            key.seqno = sequence_number_++;
            if (sequence_number_ % GUI_UPDATE_FREQUENCY == 0) {
              gui.check_update();
            }
            d.match_flatdim = matchmask;
            pair<TTTYPE_IT,bool> rval1 = table_.insert(pair<Key,Data>(key,d)); 
            if (!rval1.second) 
              throw_error("Insert failed unexpectedly");
            Index newit = rval1.first;
            if (logstr.verbosity() >= 5) {
              logstr.str() << "New incd#" << key.seqno << "# between interior and pat#" 
                << patchind << "#\n";
              logstr.str() <<" intersection real coordinates = " << d.real_coord 
                << " param coordinates = ";
              for (int ll = 0; ll < patchdim; ++ll)
                logstr.str() << d.param_coord[ll] <<  " ";
              logstr.str() << " uncertainty = " 
                << d.param_uncertainty << "\n";
              logstr.str() << " match flatdim = " << d.match_flatdim << '\n';
              
            }
            output_vec.push_back(newit);
          }
        }
      }
    }
  }
}

