// ------------------------------------------------------------------
// qpatchtab.cpp
//
// This file contains routines for the patch table in mesh generation.
// ------------------------------------------------------------------
// Copyright (c) 1999 by Cornell University.  All rights reserved.
// 
// See the accompanying file 'Copyright' for authorship information,
// the terms of the license governing this software, and disclaimers
// concerning this software.
// ------------------------------------------------------------------
// This file is part of the QMG software.  
// Version 2.0 of QMG, release date September 3, 1999
// ------------------------------------------------------------------



#ifdef _MSC_VER
#if _MSC_VER == 1200
#include "qmatvec.h"
#endif
#endif

#include "qpatchtab.h"
#include "qancestor.h"
#include "qmatvec.h"
#include "qerr.h"


namespace {
  using namespace QMG;
  using namespace QMG::MG;

#ifdef DEBUGGING
  ostream* debug_str;
#endif

}


// Static member function to apply three-d hh transform
// to a vector, and then shift it over.  Used to project
// from 3D to 2D.

void QMG::MG::PatchTable::apply_3d_hh_and_shift(Real beta, 
                                                const Point& hhtransform,
                                                Point& vec) {
  Real ip = 0.0;
  {
    for (int j = 0; j < 3; ++j) {
      ip += hhtransform[j] * vec[j];
      }
  }
  {
    for (int j = 1; j < 3; ++j) {
      vec[j - 1] = vec[j] - beta * ip * hhtransform[j];
    }
  }
  vec[2] = 0.0;
}



// member function get_outward_tan_in_parent computes a tangent
// to the specified parent patch of the given patch, that
// points outward from the specified point on the patch.

QMG::Point 
QMG::MG::PatchTable::get_outward_tan_in_parent(const Point& paramcoord,
                                               Index patchind, 
                                               Loop_over_parents& loop_in_progress) const {

  Index parent_index = loop_in_progress.parent_index();
  Point liftcoords = lift_parameters(paramcoord, patchind, loop_in_progress.listind());
  Point paramdirec;
  if (di_ == 2) {
    if (loop_in_progress.lift_index(0) == 0)
      paramdirec[0] = 1.0;
    else
      paramdirec[0] = -1.0;
  } 
  else { // di_ == 3

    int lift0 = loop_in_progress.lift_index(0);
    int lift1 = loop_in_progress.lift_index(1);
    if (patches_[parent_index].ptype == BEZIER_TRIANGLE) {
      if ((lift0 == 2 || lift0 == 3) && (lift1 == 0)) {
        paramdirec[0] = -0.5;
        paramdirec[1] = 1.0;
      }
      else if (lift0 == 0 && (lift1 == 2 || lift1 == 3)) {
        paramdirec[0] = 1.0;
        paramdirec[1] = -0.5;
      }
      else if (lift0 + lift1 == 5) {
        paramdirec[0] = -0.5;
        paramdirec[1] = -0.5;
      }
      else {
        throw_error("Unknown lift index for triangular patch");
      }
    }
    else { //patchtype == BEZIER_QUAD
      if (lift0 == 0) {
        paramdirec[0] = 1.0;
      }
      else if (lift0 == 1) {
        paramdirec[0] = -1.0;
      }
      else {
        paramdirec[0] = 0.0;
      }
      if (lift1 == 0) {
        paramdirec[1] = 1.0;
      }
      else if (lift1 == 1) {
        paramdirec[1] = -1.0;
      }
      else {
        paramdirec[1] = 0.0;
      }
    }
  }
  return patchmath(parent_index).direc_deriv(liftcoords, paramdirec);
}


// member function vtxnbd_patchcone_bdry_tans_
// takes a 2D patch containing a 0D patch, and
// computes the tangents to the two boundary
// edges at the point.

#ifndef BUG_IN_USING
using QMG::pair;
#endif

using QMG::Triple;


Triple<QMG::Point,QMG::Point,QMG::Point>
QMG::MG::PatchTable::
vtxnbd_patchcone_bdry_tans_(Index patchind,
                            Index grandparent_index) const {
  Point bdry_tan_1, bdry_tan_2, gpliftcoord, bdry_tan;
  Point paramdirec, parent_paramcoord;
  bool first_found = false;
  bool second_found = false;
  for (Loop_over_children loop3(*this, grandparent_index);
       loop3.notdone();
       ++loop3) {
    Index uncle_index = loop3.child_index();
    for (Loop_over_children loop4(*this, uncle_index);
         loop4.notdone();
         ++loop4) {
      Index grandchild_index = loop4.child_index();
      if (grandchild_index == patchind) {
        
        int liftind = loop4.lift_index(0);
        if (liftind == 0) 
          paramdirec[0] = 1.0;
        else
          paramdirec[0] = -1.0;
        parent_paramcoord[0] = static_cast<double>(liftind);
        bdry_tan = patchmath(uncle_index).direc_deriv(parent_paramcoord, paramdirec);
        if (!first_found) {
          bdry_tan_1 = bdry_tan;
          first_found = true;
        }
        else if (!second_found) {
          bdry_tan_2 = bdry_tan;
          second_found = true;
          break;
        }
        gpliftcoord = lift_parameters_from_childloop(parent_paramcoord,
                                                     grandparent_index,
                                                     loop3.listind());
        
      }
    }
    if (first_found && second_found) break;
  }
  if (!first_found || !second_found)
    throw_error("Parent/child patch inconsistency");
  bdry_tan_1.normalize();
  bdry_tan_2.normalize();
  return Triple<Point, Point,Point>(bdry_tan_1, bdry_tan_2, gpliftcoord);
}


// member function vtxnbd_select_patchcone_randpoint
// takes a 2D patch containing a 0D patch, and
// selects a random point in the cone defined by the 
// the patch (extending the boundary tangents
// from the point out to infinity).
// Only used for 3D.

QMG::Point
QMG::MG::PatchTable::
vtxnbd_select_patchcone_randpoint(Index patchind,
                                  Index ancestor_index) const {

  Point endpoint2;
  if (gdim(ancestor_index) == 2) {
    Real wt1 = 0.2 + 0.6 * random_real();
    Real wt2 = 1.0 - wt1;
    Triple<Point, Point, Point> rval =
      vtxnbd_patchcone_bdry_tans_(patchind, ancestor_index);
    for (int j = 0; j < 3; ++j)
      endpoint2[j] = wt1 * rval.first[j] + wt2 * rval.second[j];
  }
  else {
    bool found = false;
    Point parent_paramcoord, paramdirec;
    for (Loop_over_children loop3(*this, ancestor_index);
         loop3.notdone();
         ++loop3) {
      if (loop3.child_index() == patchind) {
        int liftind = loop3.lift_index(0);
        if (liftind == 0) 
          paramdirec[0] = 1.0;
        else
          paramdirec[0] = -1.0;
        parent_paramcoord[0] = static_cast<double>(liftind);
        endpoint2 = 
          patchmath(ancestor_index).direc_deriv(parent_paramcoord, paramdirec);
        found = true;
        break;
      }
    }
    if (!found)
      throw_error("Illegal arguments to vtxnbd_select_patchcone_randpoint");
  }
  return endpoint2;
}


// member function vtxnbd_patchcone_meets_seg
// takes a 2D patch containing a 0D patch, and
// computes the intersection of the
// the patch with a segment.  It returns a code and the distance.

pair<QMG::MG::PatchTable::ConeIntersectCode, QMG::Real>
QMG::MG::PatchTable::
vtxnbd_patchcone_meets_seg(Index patchind,
                           Index grandparent_index,
                           const Point& endpoint1,
                           const Point& endpoint2,
                           const Point& segtan,
                           double degen_cutoff) const {

  // Get the two boundary tangents.

  Triple<Point, Point, Point> rval =
    vtxnbd_patchcone_bdry_tans_(patchind, grandparent_index);
  Point gpnormal = 
    patchmath(grandparent_index).normal(rval.third);
  Real denom = Point::inner_product(segtan, gpnormal);
  Real numer = Point::inner_product(endpoint1,gpnormal);
  if (fabs(numer) < degen_cutoff || fabs(denom) < degen_cutoff) {
    return pair<ConeIntersectCode, Real>(DEGEN, 0.0);
  }
  Real newdist = numer / denom;
  if (newdist < 0)
    return pair<ConeIntersectCode, Real>(MISSES_PATCH, 0.0);
  
  // Check if seg goes thru patch here
  
  Point ipoint;
  {
    for (int j = 0; j < 3; ++j) {
      ipoint[j] = (1.0 - newdist) * endpoint1[j] + newdist * endpoint2[j];
    }
  }
  
  Point cross1 = Point::cross_product(rval.first, ipoint);
  Point cross2 = Point::cross_product(ipoint, rval.second);
  Point cross3 = Point::cross_product(rval.first, rval.second);
  Real ip1 = Point::inner_product(cross1, cross3);
  Real ip2 = Point::inner_product(cross2, cross3);
  if (fabs(ip1) < degen_cutoff) {
    Real ip3 = Point::inner_product(ipoint, rval.second);
    if (ip3 > 0)
      return pair<ConeIntersectCode, Real>(HITS_EDGE, newdist);
    else 
      return pair<ConeIntersectCode, Real>(MISSES_PATCH, 0.0);
  }
  if (ip1 < 0) {
    return pair<ConeIntersectCode, Real>(MISSES_PATCH, 0.0);
  }
  if (fabs(ip2) < degen_cutoff) {
    Real ip3 = Point::inner_product(ipoint, rval.first);
    if (ip3 > 0)
      return pair<ConeIntersectCode, Real>(HITS_EDGE, newdist);
    else 
      return pair<ConeIntersectCode, Real>(MISSES_PATCH, 0.0);
  }
  if (ip2 < 0) {
    return pair<ConeIntersectCode, Real>(MISSES_PATCH, 0.0);
  }
  if (denom > 0)
    return pair<ConeIntersectCode, Real>(HITS_PATCH_FRONT, newdist);
  return pair<ConeIntersectCode, Real>(HITS_PATCH_BACK, newdist);
}


// member function determine_ray_front_face takes a ray
// coming out of a patch and determines which side of a full-dimensional
// patch sees the ray.  This is complicated if the input patch
// is low dimensional, because then we have to determine
// sidedness from a computation involving the
// patch's ancestors.


pair<QMG::MG::PatchTable::Index, QMG::MG::PatchTable::ConeIntersectCode>
QMG::MG::PatchTable::determine_ray_front_face(Index patchind, 
                                              const Point& param_coord,
                                              const Point& ray_direction,
                                              double tol) const {
#ifdef DEBUGGING
  bool trace_on = false;

  if (trace_on) {
    *debug_str << "Reached determine ray_front_face patchind = pat#" << patchind
      << "#\nray_direction = " << ray_direction << " param_coord = ";
    for (int j = 0; j < this -> gdim(patchind); ++j)
      *debug_str << param_coord[j]  << " ";
    *debug_str << "\n";
  }
#endif

  typedef pair<Index, ConeIntersectCode> ReturnvalType;

  Real degen_cutoff = sqrt(tol);

  Point normalized_direction = ray_direction;
  normalized_direction.normalize();
  Index result_patchind;
  ConeIntersectCode result_code;
  bool prev_empty;
  int gdim1 = gdim(patchind);


  // If the patch has dimension di-1, then the side is determined by
  // whether the inner product of the normal to the patch with the ray
  // is positive or negative.

  if (gdim1 == di_ - 1) {
    Point normal = patchmath(patchind).normal(param_coord);
    Real ip = Point::inner_product(normal, ray_direction);


#ifdef DEBUGGING
    if (trace_on) {
      *debug_str << "Case 1, normal = " << normal << " ip = " << ip << "\n";
    }
#endif

    if (fabs(ip) < degen_cutoff) {
#ifdef DEBUGGING
      if (trace_on)
        *debug_str << " returning degen\n";
#endif
      return ReturnvalType(0,DEGEN);
    }
    else if (ip > 0) {
#ifdef DEBUGGING
      if (trace_on)
        *debug_str << " returning " << front_side(patchind) << "\n";
#endif
      return ReturnvalType(patchind, HITS_PATCH_FRONT);
    }
    else {
#ifdef DEBUGGING
      if (trace_on)
        *debug_str << " returning " << back_side(patchind) << "\n";
#endif
      return ReturnvalType(patchind, HITS_PATCH_BACK);
    }
  }

  // If we failed that first test but the patch has no parents, then
  // it is an isolated internal boundary and we already classified the
  // region containing it.

  if (patches_[patchind].num_parents == 0) {
#ifdef DEBUGGING
    if (trace_on) 
      *debug_str << "Case 2 returning " << front_side(patchind) << "\n";
#endif
    return ReturnvalType(patchind, HITS_PATCH_FRONT);
  }

  // If the patch is dimension di-2 then it is either a segment between
  // two or more patches in R^3 or a point between two or more curves
  // in R^2.  In either case, we determine sidedness by finding the closest
  // counterclockwise patch of higher dimension to the ray, and using the
  // front or back classification of that higher dimensional patch.

  if (gdim1 == di_ - 2) {
    Point hhtransform;
    Real newcutoff = degen_cutoff;
    Real beta;
    if (di_ == 3) {
      Point direc;
      direc[0] = 1.0;
      Point tangent = patchmath(patchind).direc_deriv(param_coord, direc);
      beta = Matrix::compute_hh_transform(&hhtransform[0], &tangent[0], 1, 3);
      // Apply the HH transform to the ray direction; it's now 2d
      
      apply_3d_hh_and_shift(beta, hhtransform, normalized_direction);
      
      
      Real l2n = normalized_direction.l2norm();
      if (l2n < degen_cutoff) {
#ifdef DEBUGGING
        if (trace_on) 
          *debug_str << "newtol too small -- returning degen\n";
#endif
        return ReturnvalType(0, DEGEN);
      }
      newcutoff /= l2n;
    }


    Real ray_angle = atan2(normalized_direction[1], normalized_direction[0]);
    
#ifdef DEBUGGING
    if (trace_on)
      *debug_str << "Case 3 newcutoff = " << newcutoff << " ray_angle = " 
      << ray_angle << '\n';
#endif

    // Now figure out the closest counterclockwise angle of the tangents of the parents to this
    // face.
    Real closest_angle = BIG_REAL;
    for (Loop_over_parents loop(*this,patchind); 
    loop.notdone();
    ++loop) {
      Index parent_index = loop.parent_index();
      Point liftcoords = lift_parameters(param_coord, patchind, loop.listind());
      Point parent_normal = patchmath(parent_index).normal(liftcoords);
      Point parent_direc_deriv = 
        get_outward_tan_in_parent(param_coord, patchind, loop);
      if (di_ == 3) {
        apply_3d_hh_and_shift(beta, hhtransform, parent_direc_deriv);
        apply_3d_hh_and_shift(beta, hhtransform, parent_normal);
      }
      
      Real tanang = atan2(parent_direc_deriv[1],parent_direc_deriv[0]);
#ifdef DEBUGGING
      if (trace_on) 
        *debug_str << "parent = pat#" << parent_index << "# direc deriv = " 
                   << parent_direc_deriv[0] << " " << parent_direc_deriv[1]
          << " tanang = " << tanang << '\n';
#endif
      if (fabs(sin(tanang - ray_angle)) < newcutoff) {
#ifdef DEBUGGING
        if (trace_on)
          *debug_str << " returning because of degeneracy 1\n";
#endif
        return ReturnvalType(0, DEGEN);
      }
      if (tanang < ray_angle)
        tanang += 2 * PI;
      if (fabs(sin(tanang - ray_angle)) < newcutoff) {
#ifdef DEBUGGING
        if (trace_on)
          *debug_str << " returning because of degeneracy 2\n";
#endif
        return ReturnvalType(0, DEGEN);
      }
      bool select_tanang = false;
      if (fabs(tanang - closest_angle) < tol) {
        // Get here if two patches have almost the
        // same tangent.  In this case,
        // favor the patch that has the inside of
        // the domain facing the ray.

        if (prev_empty) {
          select_tanang = true;
        }
      }
      else if (tanang < closest_angle) {
        select_tanang = true;
      }

      if (select_tanang) {
        closest_angle = tanang;
        result_patchind = parent_index;
#ifdef DEBUGGING
        if (trace_on)
          *debug_str << " result patchind = " << result_patchind;
#endif
        if (parent_normal[1] * parent_direc_deriv[0] - 
          parent_normal[0] * parent_direc_deriv[1] < 0) {
          result_code = HITS_PATCH_FRONT;
          prev_empty = front_side(result_patchind).faceind() < 0;
#ifdef DEBUGGING
          if (trace_on)
            *debug_str << " frontside \n";
#endif
        }
        else {  
          result_code = HITS_PATCH_BACK;
          prev_empty = back_side(result_patchind).faceind() < 0;
#ifdef DEBUGGING
          if (trace_on)
            *debug_str << " backside \n";
#endif
        }
      }
    }
    return ReturnvalType(result_patchind, result_code);
  }
  // Get here only if gdim = 0 and di = 3.
  // This is the most complicated case.
  // In this case we pick a patch arbitrarily and extend its cone to
  // infinity and pick a test point in this cone.  Then we drop 
  // a segment from a point on the ray to this test point and find
  // which patches it crosses.  Whichever is the closest patch that
  // it crosses determines the chamber.


  // First find a 2D parent.

#ifdef DEBUGGING
  if (trace_on)
    *debug_str << "Case 4\n";
#endif


  Point segtan;
  Point endpoint1 = normalized_direction;
  Point endpoint2;

  {
    bool grandparent_found = false;
    for (Loop_over_parents loop1(*this, patchind);
         loop1.notdone();
         ++loop1) {
      Index parent_index = loop1.parent_index();
      for (Loop_over_parents loop2(*this, parent_index);
           loop2.notdone();
           ++loop2) {
        Index grandparent_index = loop2.parent_index();
        endpoint2 = 
          vtxnbd_select_patchcone_randpoint(patchind, 
                                            grandparent_index);
        segtan = Point::subtract(endpoint1, endpoint2);
        grandparent_found = true;
        break;
      }
      if (grandparent_found) break;
    }
    if (!grandparent_found) {
      for (Loop_over_parents loop(*this, patchind);
           loop.notdone();
           ++loop) {
        Index parent_index = loop.parent_index();
#ifdef DEBUGGING
        if (trace_on)
          *debug_str << "no two 2d parent; using 1-dparent pat#"
                     << parent_index << "# and code " << HITS_PATCH_FRONT << '\n';
#endif
        return ReturnvalType(parent_index, HITS_PATCH_FRONT);
      }
    }
  }


  Real bestdist = BIG_REAL;
  {
    for (Loop_over_parents loop1(*this, patchind);
         loop1.notdone();
         ++loop1) {
      Index parent_index = loop1.parent_index();
      for (Loop_over_parents loop2(*this, parent_index);
           loop2.notdone();
           ++loop2) {
        Index grandparent_index = loop2.parent_index();
        pair<ConeIntersectCode, Real> rval = 
          vtxnbd_patchcone_meets_seg(patchind, grandparent_index,
                                     endpoint1, endpoint2,segtan, 
                                     degen_cutoff);
        if (rval.first == DEGEN || rval.first == HITS_EDGE) 
          return ReturnvalType(0, DEGEN);
        if (rval.first == MISSES_PATCH)
            continue;
        Real newdist = rval.second;
        bool select_newdist = false;
        if (fabs(newdist - bestdist) < tol) {
          if (prev_empty) {
            select_newdist = true;
          }
        } else if (newdist < bestdist) {
            select_newdist = true;
        }
        if (select_newdist) {
          bestdist = newdist;
          result_code = rval.first;
          result_patchind = grandparent_index;
          Brep::Face_Spec chamb = (result_code == HITS_PATCH_FRONT)?
            front_side(result_patchind) : back_side(result_patchind);
          if (chamb.faceind() < 0)
            prev_empty = true;
        }
      }
    }
  }

  if (bestdist > 1.1)
    throw_error("Inconsistent state of bestdist");
#ifdef DEBUGGING
  if (trace_on)
    *debug_str << "returning result = " << result_patchind 
    << " " << result_code << '\n';
#endif
  return ReturnvalType(result_patchind, result_code);
}

// lift_parameters lifts parametric coordinates from a child
// patch to a parent patch.


QMG::MG::Point
QMG::MG::PatchTable::lift_parameters(const Point& lowdim_param, 
                                     Index child_index,
                                     int listind) const {

#ifdef RANGECHECK
  if (child_index < 0 || child_index >= patches_.size())
    throw_error("patchind out of range");
  if (listind < 0 || listind >= patches_[child_index].num_parents)
    throw_error("Child ind out of bounds");
#endif
  Point returnval;
  int gdim1 = patches_[child_index].gdim + 1;
  Real tmp[4];
  tmp[0] = 0.0;
  tmp[1] = 1.0;
  Real t = 1.0;
  {
    for (int i = 0; i < gdim1 - 1; ++i)
      t -= (tmp[i + 2] = lowdim_param[i]);
  }
  tmp[gdim1 + 1] = t;

  {
    int cs = patches_[child_index].parent_start + listind;
    for (int i = 0; i < gdim1; ++i) {
      int j = parentlists_[cs].lift_index[i];
      returnval[i] = tmp[j];
    }
  }
  return returnval;
}


QMG::MG::Point
QMG::MG::PatchTable::lift_parameters_from_childloop(const Point& lowdim_param, 
                                     Index parent_index,
                                     int listind) const {

#ifdef RANGECHECK
  if (parent_index < 0 || parent_index >= patches_.size())
    throw_error("patchind out of range");
  if (listind < 0 || listind >= patches_[parent_index].num_children)
    throw_error("Child ind out of bounds");
#endif
  Point returnval;
  int gdim1 = patches_[parent_index].gdim;
  Real tmp[4];
  tmp[0] = 0.0;
  tmp[1] = 1.0;
  Real t = 1.0;
  {
    for (int i = 0; i < gdim1 - 1; ++i)
      t -= (tmp[i + 2] = lowdim_param[i]);
  }
  tmp[gdim1 + 1] = t;

  {
    int cs = patches_[parent_index].child_start + listind;
    for (int i = 0; i < gdim1; ++i) {
      int j = childlists_[cs].lift_index[i];
      returnval[i] = tmp[j];
    }
  }
  return returnval;
}

// Determines if a patch is an ancestor of another.


bool QMG::MG::PatchTable::is_ancestor(Index child, Index ancestor) const {
#ifdef RANGECHECK
  if (child < 0 ||  child >= patches_.size() || 
    ancestor < 0 || ancestor >= patches_.size())
    throw_error("Range error in is_ancestor");
#endif
  if (child == ancestor) return true;
  for (Loop_over_children loop(*this, ancestor);
  loop.notdone();
  ++loop) {
    Index subind1 = loop.child_index();
    if (subind1 == child) return true;
    for (Loop_over_children loop2(*this, subind1);
    loop2.notdone();
    ++loop2) {
      Index subind2 = loop2.child_index();
      if (subind2 == child) return true;
    }
  }
  return false;
}



// patch's number of control points

int
QMG::MG::PatchTable::PatchInfo::num_ctl_pts() const {
  if (gdim == 0)
    return 1;
  else if (gdim == 1)
    return degree1 + 1;
  else if (gdim == 2 && ptype == BEZIER_TRIANGLE) 
    return (degree1 + 1) * (degree1 + 2) / 2;
  else if (gdim == 2 && ptype == BEZIER_QUAD)
    return (degree1 + 1) * (degree2 + 1);
  return -1;
}



// This class helps construct the patch table.
// We use a helper class because otherwise the
// patchtable constructor would be a huge routine.


class QMG::MG::PatchTable::Constructor_Helper {
private:
  struct PatchKey {
    int num_vert;
    Brep::ControlPointIndex vertex[4];
    bool operator<(const PatchKey& other) const;
    bool operator==(const PatchKey& other) const;
    void initialize_from_patchinfo(const PatchInfo& rec,
      const vector<Brep::ControlPointIndex>& control_point);
  };
  
  typedef map<PatchKey, Index> PatchLookup;
  typedef PatchLookup::iterator PatchLookupI;

  struct ParentLink {
    Index parent_index;
    int listind;
  };

  typedef multimap<Index, ParentLink> ParentTableType;
  typedef ParentTableType::iterator ParentTableTypeI;

  typedef pair<Index, bool> InsertReturnval;
  
  PatchTable& pt_;
  const Brep& brep_;
  Brep::Ancestor_Lookup lca_;
  PatchLookup patch_lookup_; 
  ParentTableType parent_table_;
  

  vector<Index> parent1_;
  vector<Index> parent2_;
  vector<My_bool> parent1_is_forward_;
  vector<My_bool> parent2_is_forward_;
  vector<My_bool> orientation_is_computed_;
  vector<Brep::Face_Spec> parent_fspec_;
  
  pair<Index,bool> insert_one_patch_and_recurse_(PatchInfo& patchinfo, 
    Logstream& logstr,
    const vector<Brep::ControlPointIndex>& control_points,
    const Brep::Face_Spec& owner,
    Brep::PatchIndex patchind);
  
  StatVector<bool, 33> lookup_orientation_;
  Brep::Face_Labeling<Index> patchind_begin_;
  Brep::Face_Labeling<Index> patchind_end_;
  Constructor_Helper(PatchTable& pt, const Brep& g) : pt_(pt), brep_(g), lca_(g),
    patchind_begin_(g, 0), patchind_end_(g, 0) {
    lookup_orientation_[0] = true;
    lookup_orientation_[2] = true;
    lookup_orientation_[3] = false;
    lookup_orientation_[10] = false;
    lookup_orientation_[12] = false;
    lookup_orientation_[13] = true;
    lookup_orientation_[20] = false;
    lookup_orientation_[21] = true;
    lookup_orientation_[23] = true;
    lookup_orientation_[30] = true;
    lookup_orientation_[31] = false;
    lookup_orientation_[32] = false;
  }
  ~Constructor_Helper() { }
  void put_errmsg1_(ostream& os, 
    const Brep::Face_Spec& fspec,
    Brep::PatchIndex patchind);
  void put_errmsg2_(ostream& os, 
    const Brep::Face_Spec& fspec);
  void put_errmsg3_(ostream& os, 
    const Brep::Face_Spec& fspec);

  void insert_patches(const Brep::Face_Spec& fspec, Logstream& logstr);
  void compute_orientation(const Brep::Face_Spec& fspec);
  void compute_is_root();
  void compute_all_parents();
  void assign_null_front_back_face();
  
  struct LineHitRecord {
    Real dist;
    bool front_hit;
    Index patchind;
    bool operator<(const LineHitRecord& other) const;
  };
  LineHitRecord lhr_default;
  
  void compute_front_back_face(double scfac, double tol, Logstream& logstr);
  void verify_grandchild_parity();
  friend PatchTable::PatchTable(const Brep& brep, Logstream& logstr, double scfac, double tol);
};


void 
QMG::MG::PatchTable::dump_patch(Index i, 
                                ostream& os) const {
  const PatchInfo& pi = patches_[i];
  os << "index = pat#" << i <<  "# owner = " << brep_.face_name(pi.owner) << " type = " << pi.ptype
    << " dim = " << pi.gdim << " degrees = " << pi.degree1 << " " << pi.degree2
    << "\n  orig index = " << pi.orig_index << " is_forward = " << pi.is_forward
    << " is_root = " << pi.is_root  << " front, back side = ";
  if (pi.is_root)
    os << pi.front_side << " " << pi.back_side;
  else
    os << "----";
  os  << "#\n  children = ";
  for (int j = 0; j < pi.num_children; ++j) {
    int j1 = j + patches_[i].child_start;
    os << "pat#" << childlists_[j1].childnum << "# (" << childlists_[j1].lift_index[0]
      << " " << childlists_[j1].lift_index[1] << ")  ";
  }
  os << "  parents = ";
  for (int m = 0; m < pi.num_parents; ++m) {
    int m1 = m + patches_[i].parent_start;
    os << " pat#" << parentlists_[m1].parentnum << "# (" << parentlists_[m1].lift_index[0]
      << " " << parentlists_[m1].lift_index[1] << ")  ";
  }
  os << "\n control points = ";
  for (int k = 0; k < pi.num_ctl_pts(); ++k) {
    int k1 = patches_[i].control_point_start + k;
    Brep::ControlPointIndex p = control_point_list_[k1];
    os << p << " (";
    for (int l1 = 0; l1 < brep_.embedded_dim(); ++l1) 
      os << brep_.control_point(p, l1) << " ";
    os << ") ";
  }
  if (pi.gdim == di_ - 1)
    os << "\nnormal at barycenter = " << patchmath(i).normal_at_barycenter()
    << " points toward " << pi.front_side << " away from " << pi.back_side << '\n';
  os << std::endl;
}


// patch table constructor invokes helper routines.
// If tol <= 0, then front-back face not computed.


QMG::MG::PatchTable::PatchTable(const Brep& brep, 
                                Logstream& logstr,
                                double scfac, 
                                double tol) :
brep_(brep), 
// serial_number_(1),
di_(brep.embedded_dim())
{ 
#ifdef DEBUGGING
  debug_str = &logstr.str();
#endif

  Constructor_Helper ch(*this, brep);
  ch.verify_grandchild_parity();
  for (Brep::Face_Spec_Loop_Over_Faces fspec(brep);
  fspec.notdone();
  ++fspec) {
    ch.patchind_begin_(fspec) = patches_.size();
    ch.insert_patches(fspec, logstr);
    ch.patchind_end_(fspec) = patches_.size(); 
    ch.compute_orientation(fspec);
  }
  ch.compute_is_root();
  if (tol > 0)
    ch.compute_front_back_face(scfac, tol, logstr);
  else
    ch.assign_null_front_back_face();
  ch.compute_all_parents();
  if (logstr.verbosity() >= 2) {
    logstr.str() << "Dump of patch table ";
    logstr.logdate();
    logstr.str() << std::endl;
    for (Index i = 0; i < patches_.size(); ++i) {
      dump_patch(i, logstr.str());
      logstr.str() << '\n';
    }
    logstr.str() << std::flush;
  }
}

bool QMG::MG::PatchTable::Constructor_Helper::PatchKey::
operator<(const PatchKey& other) const {
  if (num_vert < other.num_vert) return true;
  if (num_vert > other.num_vert) return false;
  for (int jj = 0; jj < num_vert; ++jj) {
    if (vertex[jj] < other.vertex[jj]) return true;
    if (vertex[jj] > other.vertex[jj]) return false;
  }
  return false;
}


bool QMG::MG::PatchTable::Constructor_Helper::PatchKey::
operator==(const PatchKey& other) const {
  if (num_vert != other.num_vert) return false;
  for (int jj = 0; jj < num_vert; ++jj)
    if (vertex[jj] != other.vertex[jj]) return false;
  return true;
}

bool QMG::MG::PatchTable::Constructor_Helper::LineHitRecord::
operator<(const LineHitRecord& other) const {
  if (dist < other.dist) return true;
  if (dist > other.dist) return false;
  if (front_hit && !other.front_hit) return true;
  if (!front_hit && other.front_hit) return false;
  if (patchind < other.patchind) return true;
  return false;
}




void QMG::MG::PatchTable::Constructor_Helper::PatchKey::
initialize_from_patchinfo(const PatchInfo& rec,
                          const vector<Brep::ControlPointIndex>& control_point) {
  if (rec.gdim == 0) {
    vertex[0] = control_point[0];
    num_vert = 1;
  }
  else if (rec.gdim == 1) {
    num_vert = 2;
    vertex[0] = control_point[0];
    vertex[1] = control_point[rec.degree1];
  }
  else if (rec.gdim == 2 && rec.ptype == BEZIER_TRIANGLE) {
    num_vert = 3;
    vertex[0] = control_point[0];
    vertex[1] = control_point[rec.degree1 * (rec.degree1 + 1) / 2];
    vertex[2] = control_point[(rec.degree1 + 1) * (rec.degree1 + 2) / 2 - 1];
  }
  else if (rec.gdim == 2 && rec.ptype == BEZIER_QUAD) {
    num_vert = 4;
    vertex[0] = control_point[0];
    vertex[1] = control_point[rec.degree1];
    vertex[2] = control_point[rec.degree2 * (rec.degree1 + 1)];
    vertex[3] = control_point[(rec.degree2 + 1) * (rec.degree1 + 1) - 1];
  }
  // Insertion sort to put vertices in increasing order.
  
  for (int j1 = num_vert - 1; j1 > 0; --j1) {
    for (int j2 = 0; j2 < j1; ++j2) {
      if (vertex[j2] > vertex[j2 + 1]) {
        Brep::ControlPointIndex tmp = vertex[j2];
        vertex[j2] = vertex[j2 + 1];
        vertex[j2 + 1] = tmp;
      }
    }
  }
}


// This routine makes sure that each topological face contains
// its grandchildren with even parity.  This condition is required
// for brep correctness.

void 
QMG::MG::PatchTable::Constructor_Helper::verify_grandchild_parity() {
  Brep::Face_Labeling<int> occ_count(brep_);
  int di = brep_.embedded_dim();
  for (int fdim = 2; fdim <= di; ++fdim) {
    for (Brep::Face_Spec_Loop_Over_Faces_Of_Dim fspec(brep_, fdim);
    fspec.notdone();
    ++fspec) {
      {
        for (Brep::Face_Spec_Loop_Over_Face_Subfaces subfspec(brep_, fspec);
        subfspec.notdone();
        ++subfspec) {        
          for (Brep::Face_Spec_Loop_Over_Face_Subfaces subsubfspec(brep_, subfspec);
          subsubfspec.notdone();
          ++subsubfspec) {
            occ_count(subsubfspec) = 0;
          }
        }
      }
     {
        for (Brep::Face_Spec_Loop_Over_Face_Subfaces subfspec(brep_, fspec);
        subfspec.notdone();
        ++subfspec) {        
          if (subfspec.is_internal_boundary()) continue;
          for (Brep::Face_Spec_Loop_Over_Face_Subfaces subsubfspec(brep_, subfspec);
          subsubfspec.notdone();
          ++subsubfspec) {
            ++occ_count(subsubfspec);
            if (subsubfspec.is_internal_boundary())
              ++occ_count(subsubfspec);
          }
        }
      }

     {
        for (Brep::Face_Spec_Loop_Over_Face_Subfaces subfspec(brep_, fspec);
        subfspec.notdone();
        ++subfspec) {        
          if (subfspec.is_internal_boundary()) continue;
          for (Brep::Face_Spec_Loop_Over_Face_Subfaces subsubfspec(brep_, subfspec);
          subsubfspec.notdone();
          ++subsubfspec) {
            if (occ_count(subsubfspec) % 2) {
              ostringstream ost;
              ost << "Topological entity " << brep_.face_name(fspec)
                << " contains topological entity " << brep_.face_name(subsubfspec)
                << " as a grandchild\nan odd number of times.  This indicates that "
                << "the boundary of " << brep_.face_name(fspec) << " does not close properly.";
              throw_error(ost.str());
            }
          }
        }
      }
    }
  }
}


void 
QMG::MG::PatchTable::Constructor_Helper::put_errmsg1_(ostream& os, 
                                                      const Brep::Face_Spec& fspec,
                                                      Brep::PatchIndex patchind) {
  os << "\nError occurred while processing geometric entity #" << patchind
    << " of brep face " << brep_.face_name(fspec) << '\n';
}



// This routine inserts all patches for a topological face
// into the patch table.


void  
QMG::MG::PatchTable::Constructor_Helper::insert_patches(const Brep::Face_Spec& fspec,
                                                        Logstream& logstr) {
  PatchInfo patchinfo;
  vector<Brep::ControlPointIndex> control_points;
  // patchinfo.tmp_label_serial_number = 0;
  for (Brep::Loop_over_patches_of_face patch(brep_, fspec);
  patch.notdone();
  ++patch) {
    patchinfo.ptype = patch.ptype();
    patchinfo.gdim = fspec.fdim();
    patchinfo.degree1 = patch.degree1();
    if (patchinfo.gdim == 2 && patchinfo.ptype == BEZIER_QUAD)
      patchinfo.degree2 = patch.degree2();
    else
      patchinfo.degree2 = 0;
    patchinfo.orig_index = patch.index();
    patchinfo.owner = fspec;
    // patchinfo.normal_deriv_estimate = 0.0;
    patchinfo.is_forward = true;
    int numpt = patchinfo.num_ctl_pts();
    control_points.resize(numpt);
    {
      for (int j = 0; j < numpt; ++j) {
        control_points[j] = patch.control_point(j);
      }
    }
    // Check for repeated control points.
    {
      for (int j = 1; j < numpt; ++j) {
        for (int j1 = 0; j1 < j; ++j1) {
          if (control_points[j] == control_points[j1]) {
            ostringstream ost;
            ost << "Control point " << control_points[j] << " occurs more than once in the patch";
            put_errmsg1_(ost, fspec, patch.index());
            throw_error(ost.str());
          }
        }
      }
    }

    insert_one_patch_and_recurse_(patchinfo, logstr, control_points, 
      fspec, patch.index());
  }
}

// This routine does the work for inserting patches into the patch
// table. It is recursive: when we try to insert a patch, we first
// check whether its subpatches have been inserted, and if not,
// insert the subpatches first.

QMG::MG::PatchTable::Constructor_Helper::InsertReturnval
QMG::MG::PatchTable::Constructor_Helper::
insert_one_patch_and_recurse_(PatchInfo& patchinfo,
                              Logstream& logstr,
                              const vector<Brep::ControlPointIndex>& control_points,                              
                              const Brep::Face_Spec& owner1,
                              Brep::PatchIndex patchind1) {
  
  PatchInfo subpatchinfo;
  vector<Brep::ControlPointIndex> sub_control_points;
  unsigned int num_children = 0;
  int gdim = patchinfo.gdim;
  if (gdim == 1) {
    num_children = 2;
  }
  else if (gdim == 2 && patchinfo.ptype == BEZIER_TRIANGLE) {
    num_children = 3;
  }
  else if (gdim == 2 && patchinfo.ptype == BEZIER_QUAD) {
    num_children = 4;
  }
  patchinfo.num_children = num_children;
  
  // Insert the children of the patch recursively.
  
  subpatchinfo = patchinfo;
  subpatchinfo.gdim = gdim - 1;
  Index children[4];
  int lift_index[4][2];
  
  int degree1 = patchinfo.degree1;
  int degree2 = patchinfo.degree2;
  
  for (int childcount = 0; childcount < num_children; ++childcount) {
    if (gdim == 1) {
      sub_control_points.resize(1);
      int ind = (childcount == 0)? 0 : degree1;
        sub_control_points[0] = control_points[ind];
      lift_index[childcount][0] = (childcount == 0)? 0 : 1;
      lift_index[childcount][1] = 0;
    }
    else if (gdim == 2 && patchinfo.ptype == BEZIER_TRIANGLE) {
      sub_control_points.resize(degree1 + 1);
      if (childcount == 0) {
        for (int i = 0; i <= degree1; ++i) {
          sub_control_points[i] = control_points[i * (i + 1) / 2];
        }
        lift_index[childcount][0] = 0;
        lift_index[childcount][1] = 3;
      }
      else if (childcount == 1) {
        for (int i = 0; i <= degree1; ++i) {
          sub_control_points[i] = control_points[(i + 1) * (i + 2) / 2 - 1];
        }
        lift_index[childcount][0] = 2;
        lift_index[childcount][1] = 3;
      }
      else {  // if (childcount == 2) 
        for (int i = 0; i <= degree1; ++i) {
          sub_control_points[i] = control_points[degree1 * (degree1 + 1) / 2 + i];
        }
        lift_index[childcount][0] = 2;
        lift_index[childcount][1] = 0;
      }
    }
    else {  // patchinfo.gdim == 2 && patchinfo.ptype == Brep::QUAD
      if (childcount < 2) {
        sub_control_points.resize(degree1 + 1);
        int offset = childcount * degree2 * (degree1 + 1);
        for (int i = 0; i <= degree1; ++i) {
          sub_control_points[i] = control_points[i + offset];
        }
        subpatchinfo.degree1 = degree1;
        lift_index[childcount][0] = 2;
        lift_index[childcount][1] = childcount;
      }
      else {
        sub_control_points.resize(degree2 + 1);
        int offset = (childcount - 2) * degree1;
        for (int i = 0; i <= degree2; ++i) {
          sub_control_points[i] = control_points[i * (degree1 + 1) + offset];
        }
        subpatchinfo.degree1 = degree2;
        lift_index[childcount][0] = childcount - 2;
        lift_index[childcount][1] = 2;
      }
    }
    InsertReturnval rval = insert_one_patch_and_recurse_(subpatchinfo, logstr,
      sub_control_points, owner1, patchind1);
    Index ci = rval.first;
    children[childcount] = ci;
    
    // If the childpatch is not new and this patch is 2-dimensional, we may
    // have to flip lift_index.
    
    if (gdim == 2 && !rval.second) {
      Brep::ControlPointIndex  firstcp = 
        pt_.control_point_list_[pt_.patches_[ci].control_point_start];
      if (firstcp != sub_control_points[0]) {
#ifdef DEBUGGING
        if (firstcp != sub_control_points.back()) {
          throw_error("Some problem inserting children into table.");
        }
#endif
        for (int ii = 0; ii < 2; ++ii) {
          int m = lift_index[childcount][ii];
          if (m == 2 || m == 3)
            lift_index[childcount][ii] = 5 - m;
        }
      }
    }
  }
  // Now all the children are inserted.  Insert the patch itself.
  PatchKey key;
  key.initialize_from_patchinfo(patchinfo, control_points);
  pair<PatchKey, Index> newitem(key, pt_.patches_.size());
  pair<PatchLookupI, bool> ival = patch_lookup_.insert(newitem);


  if (ival.second) {
    // New item was actually inserted into the lookup table.
    // This means we go ahead and create the patch.


    Index newpatchind = pt_.patches_.size();
    int cpstart = pt_.control_point_list_.size();
    patchinfo.control_point_start = cpstart;
    int chstart = pt_.childlists_.size();
    patchinfo.child_start = chstart;
    pt_.patches_.push_back(patchinfo);

    // Write out all the control points.

    pt_.control_point_list_.resize(cpstart + control_points.size());
    for (int i = 0; i < control_points.size(); ++i)
      pt_.control_point_list_[cpstart + i] = control_points[i];
    
    // Write out all the children.
    ;
    PatchChildInfo pci;
    for (int j = 0; j < num_children; ++j) {
      pci.childnum = children[j];
      pci.lift_index[0] = lift_index[j][0];
      pci.lift_index[1] = lift_index[j][1];
      pt_.childlists_.push_back(pci);
    }


    // Write out the parents of the children into the multimap.

    ParentLink pr;

    for (int k = 0; k < num_children; ++k) {
      Index ci = children[k];
      pr.parent_index = newpatchind;
      pr.listind = k;
      parent_table_.insert(pair<Index,ParentLink>(ci,pr));
    }

    return pair<Index, bool>(newitem.second, true);
  }
  
  // Get here if the patch is already in the table.  In this case, 
  // return the index to the patch.

  // Check that the other owner of the patch is a subface of this face.

  Index oldind = ival.first -> second;


  Brep::Face_Spec otherowner = pt_.patches_[oldind].owner;
  Brep::Face_Spec fspec = patchinfo.owner;
  if (otherowner.fdim() > fspec.fdim() || 
    (otherowner.fdim() == fspec.fdim() && otherowner != fspec)) {
    ostringstream ost;
    ost << "Patch owned by conflicting brep faces " << brep_.face_name(fspec) << " and " 
      << brep_.face_name(otherowner) << "\n";
    put_errmsg1_(ost, owner1, patchind1);
    throw_error(ost.str());
  }

  if (otherowner == fspec && gdim == fspec.fdim()) {
    ostringstream ost;
    ost << "Patch occurs twice in brep topological face";
    put_errmsg1_(ost, owner1, patchind1);
    throw_error(ost.str());
  }

  if (otherowner.fdim() < fspec.fdim()) {
    bool found = false;
    for (Brep::Ancestor_Lookup::Loop_over_ancestors ancloop(lca_,otherowner);
    ancloop.notdone();
    ++ancloop) {
      if (ancloop.ancestor_fspec() == fspec) {
        found = true;
        break;
      }
    }
    if (!found) {
      ostringstream ost;
      ost << "Patch occurs in brep faces " << brep_.face_name(fspec) << " and "
        << brep_.face_name(otherowner) << " which do not have an ancestor/descendent relation";
      put_errmsg1_(ost, owner1, patchind1);
      throw_error(ost.str());
    }
  }

  return pair<Index, bool>(oldind, false);
}


// This routine computes the parent field for all patches.
// It does so by inverting the child ifnromation, which
// was already computed.

void 
QMG::MG::PatchTable::Constructor_Helper::compute_all_parents() {
  PatchTable::PatchParentInfo ppi;
  for (Index patchind = 0; patchind < pt_.patches_.size(); ++patchind) {
    pt_.patches_[patchind].parent_start = pt_.parentlists_.size();
    pt_.patches_[patchind].num_parents = 0;
    pair<ParentTableTypeI, ParentTableTypeI> rval = parent_table_.equal_range(patchind);
    if (rval.first != parent_table_.end()) {
      for (ParentTableTypeI it = rval.first; it != rval.second; ++it) {
        ppi.parentnum = (it -> second).parent_index;
        int chpos = pt_.patches_[ppi.parentnum].child_start + (it -> second).listind;
        ppi.lift_index[0] = pt_.childlists_[chpos].lift_index[0];
        ppi.lift_index[1] = pt_.childlists_[chpos].lift_index[1];
        pt_.parentlists_.push_back(ppi);
        ++(pt_.patches_[patchind].num_parents);
      }
    }
  }
}


void 
QMG::MG::PatchTable::Constructor_Helper::put_errmsg2_(ostream& os, 
                                                      const Brep::Face_Spec& fspec) {
  os << "\nError occurred while computing orientation"
    << " of brep face " << brep_.face_name(fspec) << '\n';
}



// This routine makes sure all the patches of a face
// have the same orientation.  Those that have the
// wrong orientation get a flag flipped so that their
// normals still come out right.  Orientation is
// computed by a graph search on neighboring patches.

void 
QMG::MG::PatchTable::Constructor_Helper::
compute_orientation(const Brep::Face_Spec& fspec) {

  if (fspec.fdim() == 0 || fspec.fdim() == pt_.di_)
    return;

  Index lastind = patchind_end_(fspec);
  parent1_.resize(lastind);
  parent2_.resize(lastind);
  orientation_is_computed_.resize(lastind, My_bool(false));
  parent1_is_forward_.resize(lastind, My_bool(false));
  parent2_is_forward_.resize(lastind, My_bool(false));
  parent_fspec_.resize(lastind, Brep::Face_Spec(-1,-1));

  // Loop over all child patches to initialize arrays.
  Index startind = patchind_begin_(fspec);
  Index endind = patchind_end_(fspec);


  // Determine all the descendents of fspec.  Use a graph search.

  Brep::Face_Labeling<bool> is_descendent(brep_,false);
  {
    vector<Brep::Face_Spec> searchstack;
    searchstack.resize(1, fspec);
    while (searchstack.size() > 0) {
      Brep::Face_Spec subspec = searchstack.back();
      searchstack.pop_back();
      is_descendent(subspec) = true;
      for (Brep::Face_Spec_Loop_Over_Face_Subfaces subsubfspec(brep_, subspec);
      subsubfspec.notdone();
      ++subsubfspec) {
        searchstack.push_back(subsubfspec);
      }
    }
  }

  // Determine which children are internal boundaries.
  
  Brep::Face_Labeling<int> occurrence_count(brep_,0);
  {
    for (Brep::Face_Spec_Loop_Over_Face_Subfaces subfspec(brep_,fspec);
    subfspec.notdone();
    ++subfspec)
      occurrence_count(subfspec) += (subfspec.is_internal_boundary())? 2 : 1;
  }
  


  {
    for (Index patchind = startind; patchind < lastind; ++patchind) {
      if (pt_.gdim(patchind) == fspec.fdim())
        orientation_is_computed_[patchind] = false;
    }
  }

  // Fill in the parent fields.

  {
    for (Index patchind = startind; patchind < lastind; ++patchind) {
      if (pt_.gdim(patchind) < fspec.fdim())
        continue;  
      for (Loop_over_children loop(pt_, patchind); loop.notdone(); ++loop) {
        Index childind = loop.child_index();
        if (!is_descendent(pt_.owner(childind))) {
          ostringstream ost;
          ost << "Patch " << pt_.orig_index(patchind) << " of topological face " << brep_.face_name(fspec)
            << " has a subpatch lying on topological face " << brep_.face_name(pt_.owner(childind))
            << "\nbut the latter is not a descendent of the former.";
          put_errmsg2_(ost, fspec);
          throw_error(ost.str());
        }
        int whichparent;
        if (parent_fspec_[childind] != fspec) {
          parent_fspec_[childind] = fspec;
          parent1_[childind] = patchind;
          parent2_[childind] = -1;
          whichparent = 1;
        }
        else {
          if (parent2_[childind] != -1) {
            ostringstream ost;
            ost << "Patch " << pt_.orig_index(patchind) << " of topological face " << brep_.face_name(fspec)
              << " has a subpatch owned by two other patches of this topological face, namely,\n"
              << "patches" << parent1_[childind] << " and " << parent2_[childind]
              << ", which is an illegal topological construction.";
            put_errmsg2_(ost, fspec);
            throw_error(ost.str());
          }
          parent2_[childind] = patchind;
          whichparent = 2;
        }
        // Determine orientation with respect to parent.

        int listind = loop.listind();
        int li0 = pt_.childlists_[pt_.patches_[patchind].child_start + listind].lift_index[0];
        int li1 = pt_.childlists_[pt_.patches_[patchind].child_start + listind].lift_index[1];
        int code = li0 * 10 + li1;
        bool ori = !lookup_orientation_[code];
        if (whichparent == 1)
          parent1_is_forward_[childind] = ori;
        else
          parent2_is_forward_[childind] = ori;
      }
    }
  }

  // Check that each subpatch is covered the right number of times.
  
  {
    for (Index patchind = startind; patchind < lastind; ++patchind) {
      if (pt_.gdim(patchind) < fspec.fdim())
        continue;  
      for (Loop_over_children loop(pt_, patchind); loop.notdone(); ++loop) {
        Index childind = loop.child_index();
        Brep::Face_Spec owner = pt_.owner(childind);
        if (parent2_[childind] == -1) {
          if (owner.fdim() != fspec.fdim() - 1 ||
            !is_descendent(owner) ||
             occurrence_count(owner) & 1 == 0) {
            ostringstream ost;
            ost << "Patch " << pt_.orig_index(patchind) << " of topological face " <<
              brep_.face_name(fspec) << " has one of its boundaries exposed but the exposed boundary\n"
              << "is not owned by a topological subface.\nThe control points of the exposed boundary are:";
            int cps = pt_.patches_[childind].control_point_start;
            for (int jj = 0; jj < pt_.patches_[childind].num_ctl_pts(); ++jj) {
              ost << pt_.control_point_list_[cps + jj] << ' ';
            }
            put_errmsg2_(ost, fspec);
            throw_error(ost.str());
          }
        }
        else {
          if (owner.fdim() == fspec.fdim() - 1 &&
            occurrence_count(owner) & 1 == 1) {
            ostringstream ost;
            ost << "Patch " << pt_.orig_index(patchind) << " of topological face "
              <<   brep_.face_name(fspec) << " has a subpatch interior to the face and yet is owned\n" 
              <<  "by a proper topological subface " << brep_.face_name(owner);
            put_errmsg2_(ost, fspec);
            throw_error(ost.str());
          }
        }
      }
    }
  }

  // Check that each topological subface has all of its patches covered.

  {
    for (Brep::Face_Spec_Loop_Over_Face_Subfaces subspec(brep_,fspec);
    subspec.notdone();
    ++subspec) {
      if (subspec.fdim() != fspec.fdim() - 1) continue;
      for (Index subind = patchind_begin_(subspec); 
      subind < patchind_end_(subspec);
      ++subind) {
        if (pt_.gdim(subind) != subspec.fdim()) continue;
        if (parent_fspec_[subind] != fspec) {
          ostringstream ost;
          ost << "Topological subface " << brep_.face_name(subspec) << " patch " 
            << pt_.orig_index(subind) << " is not adjacent to any patch of\n"
            << " topological face " << brep_.face_name(fspec);
          put_errmsg2_(ost, fspec);
          throw_error(ost.str());
        }
      }
    }
  }

  // Now finally assign orientations using a graph search.

  {
    for (Index patchind = startind; patchind < endind; ++patchind) {
      if (pt_.gdim(patchind) < fspec.fdim()) continue;
      if (orientation_is_computed_[patchind]) continue;

      // Start a search from this patch.

      orientation_is_computed_[patchind] = true;
      pt_.patches_[patchind].is_forward = true;

      vector<Index> stack;
      stack.resize(1);
      stack[0] = patchind;

      while (stack.size() > 0) {
        Index patchind1 = stack.back();
        stack.pop_back();
        bool this_ori = pt_.patches_[patchind1].is_forward;
        for (Loop_over_children loop(pt_, patchind1); loop.notdone(); ++loop) {
          Index childind = loop.child_index();
          Index otherparent;
          if (parent1_[childind] == patchind1) {
            otherparent = parent2_[childind];
          }
          else if (parent2_[childind] == patchind1) {
            otherparent = parent1_[childind];
          }
          else {
            ostringstream ost;
            ost << "Inconsistent internal data structure in compute_orientation";
            put_errmsg2_(ost, fspec);
            throw_error(ost.str());
          }
          if (otherparent == -1) continue;
          bool nbr_correct_ori = (parent1_is_forward_[childind] == parent2_is_forward_[childind])?
            !this_ori : this_ori;
          if (orientation_is_computed_[otherparent]) {
            if (pt_.patches_[otherparent].is_forward != nbr_correct_ori) {
              ostringstream ost;
              ost << "Unable to come up with consistent orientation -- possible moebius strip";
              put_errmsg2_(ost, fspec);
              throw_error(ost.str());
            }
          } 
          else {
            orientation_is_computed_[otherparent] = true;
            pt_.patches_[otherparent].is_forward = nbr_correct_ori;
            stack.push_back(otherparent);
          }
        }
      }
    }
  }
}

// This routine determines which patches are roots, namely, those with
// no parents.

void QMG::MG::PatchTable::Constructor_Helper::
compute_is_root() {
  {
    for (Index patchind = 0; patchind < pt_.patches_.size(); ++patchind) {
      pt_.patches_[patchind].is_root = true;
    }
  }
  { 
    for (Index patchind = 0; patchind < pt_.patches_.size(); ++patchind) {
      for (Loop_over_children loop(pt_, patchind); loop.notdone(); ++loop) {
        pt_.patches_[loop.child_index()].is_root = false;
      }
    }
  }
}


#ifdef DEBUGGING
namespace QMG {
  namespace Local_to_PatchMath_CPP {
    using namespace QMG;
    extern ostream* debug_str;
    extern bool dump_everything;
  }
}
#endif


// This routine initializes the patch table front-back faces to all -1's.
// This routine is called when we need a patchtable without this info.

void QMG::MG::PatchTable::Constructor_Helper::
assign_null_front_back_face() {
  for (Index patchind1 = 0;
  patchind1 < pt_.patches_.size();
  ++patchind1) {
    pt_.patches_[patchind1].front_side = -1;
    pt_.patches_[patchind1].back_side = -1;
  }
}


void 
QMG::MG::PatchTable::Constructor_Helper::put_errmsg3_(ostream& os, 
                                                      const Brep::Face_Spec& fspec) {
  os << "\nError occurred while determining region in front and behind of "
    << " of brep face " << brep_.face_name(fspec) << '\n';
}

// This routine computes the chamber in the front and back
// of each patch of dimension di - 1.  We do this by noticing
// that the brep itself gives information about which chambers
// are bounded by the patches, so it's only a question of determining
// sidedness.  We determine sidedness by drawing a line and seeing
//what patches are hit.  We use the knowledge that the line starts
// and ends outside the brep.
// Invoked only if tol>0.


void QMG::MG::PatchTable::Constructor_Helper::
compute_front_back_face(double scfac, 
                        double tol,
                        Logstream& logstr) {

#ifdef DEBUGGING
  using QMG::Local_to_PatchMath_CPP::debug_str;
  using QMG::Local_to_PatchMath_CPP::dump_everything;
  
  debug_str = &logstr.str();
  dump_everything = false;

#endif

 
  // Determine for each face the chambers of which it is a direct subface.
  
  int di = brep_.embedded_dim();
  Brep::Face_Labeling<Brep::Face_Spec> chamber1(brep_, Brep::Face_Spec(-1,-1));
  Brep::Face_Labeling<Brep::Face_Spec> chamber2(brep_, Brep::Face_Spec(-1,-1));
  Brep::Face_Labeling<bool> is_direct_subface(brep_, false);
  Error_Message errmsg;
  {
    for (Brep::Face_Spec_Loop_Over_Faces fspec(brep_); fspec.notdone(); ++fspec) {
      int fdim = fspec.fdim();
      if (fdim == di) continue;
      if (fdim == di - 1)
        is_direct_subface(fspec) = true;
      else {
        is_direct_subface(fspec) = true;
        for (Brep::Ancestor_Lookup::Loop_over_ancestors ancloop(lca_, fspec);
        ancloop.notdone();
        ++ancloop) {
          int afdim = ancloop.ancestor_fspec().fdim();
          if (afdim > fdim && afdim < di) {
            is_direct_subface(fspec) = false;
            break;
          }
        }
      }
      bool has_chamber_ancestor = false;
      if (!is_direct_subface(fspec)) {
        // Make sure there is a chamber among the ancestors of this face.
        for (Brep::Ancestor_Lookup::Loop_over_ancestors ancloop(lca_, fspec);
        ancloop.notdone();
        ++ancloop) {
          if (ancloop.ancestor_fspec().fdim() == di) {
            has_chamber_ancestor = true;
            chamber1(fspec) = ancloop.ancestor_fspec();
            chamber2(fspec) = ancloop.ancestor_fspec();
            break;
          }
        }
      }
      else {
        for (Brep::Ancestor_Lookup::Loop_over_ancestors ancloop(lca_, fspec);
        ancloop.notdone();
        ++ancloop) {
          Brep::Face_Spec afspec = ancloop.ancestor_fspec();
          if (afspec.fdim() != di) continue;
          has_chamber_ancestor = true;
          int count = ancloop.occurrence_count();
          if (count % 2) {
            if (chamber1(fspec).fdim() >= 0) {
              if (chamber2(fspec).fdim() >= 0) {
                ostringstream ost;
                ost << "Topological face " << brep_.face_name(fspec) << " is a descendent of three\n"
                  << " regions, namely " << brep_.face_name(chamber1(fspec)) << ", " 
                  << brep_.face_name(chamber2(fspec)) << " and " << brep_.face_name(afspec)
                  << " which is not valid topologically";
                put_errmsg3_(ost, fspec);
                throw_error(ost.str());
              }
              chamber2(fspec) = afspec;
            } 
            else {
              chamber1(fspec) = afspec;
            }
          }
          else {
            if (chamber1(fspec).fdim() >= 0) {
              ostringstream ost;
              ost << "Topological face " << brep_.face_name(fspec) << " is a descendent of three\n"
                << " regions, namely " << brep_.face_name(chamber1(fspec)) << ", " 
                << " and " << brep_.face_name(afspec)
                << " (twice) which is not valid topologically";
              put_errmsg3_(ost, fspec);
              throw_error(ost.str());

            }
            chamber1(fspec) = afspec;
            chamber2(fspec) = afspec;
          }
        }
      }
      if (!has_chamber_ancestor) {
        ostringstream ost;
        ost << "Topological face " << brep_.face_name(fspec) << " is not a descendent of any region.";
        put_errmsg3_(ost, fspec);
        throw_error(ost.str());
      }
 //     std::cerr << " for face " << fspec.fdim() << ":" << fspec.faceind() << "  chamber1 = "
   //     << chamber1(fspec).fdim() << ":" << chamber1(fspec).faceind() << "  chamber2 = "
   //     << chamber2(fspec).fdim() << ":" << chamber2(fspec).faceind() << std::endl;
      
    }
  }
  
  // Now draw segments to compute front and back face.
  // Find the maximum control point.
  
  Point upper_right, lower_left;
  {
    for (int i = 0; i < di; ++i) {
      upper_right[i] = -BIG_REAL;
      lower_left[i] = BIG_REAL;
    }
   
  }
  {
    for (int cpind = 0; cpind < brep_.num_control_points(); ++cpind) {
      for (int i = 0; i < di; ++i) {
        Real t = brep_.control_point(cpind,i);
        if (t > upper_right[i])
          upper_right[i] = t;
        if (t < lower_left[i])
          lower_left[i] = t;
      }
    }
  }

  
  PatchMath::Workspace workspace1;

  Brep::Face_Labeling<bool> front_back_computed(brep_,false);
  
  {
    for (Brep::Face_Spec_Loop_Over_Faces fspec(brep_);
    fspec.notdone();
    ++fspec) {
      if (!is_direct_subface(fspec)) continue;
      if (front_back_computed(fspec)) continue;
      // Find patchind0, a full-dimensional patch of the face.

// ostringstream ost2;

      Index patchind0 = patchind_begin_(fspec);
      while (patchind0 < patchind_end_(fspec) && pt_.gdim(patchind0) < fspec.fdim())
        ++patchind0;
      if (patchind0 >= patchind_end_(fspec)) continue;

// ost2 << " processing " << fspec << " patchind0 = " << patchind0 << " bary_param = " << bary_param
// << " midpoint = " << midpoint << '\n';

      Point startpoint;
      Point endpoint;
      Point normal;
      Point bary_param = pt_.patchmath(patchind0).barycenter_param();
      vector<LineHitRecord> hitlist;
      LineHitRecord rec;
      vector<PatchMath::IntersectRecord> output_stack;
      bool degen = true;
      int trycount = 0;
      while (degen && trycount++ < 400) {
        Point perturb_param;
        {
          for (int ii = 0; ii < di - 1; ++ii)
            perturb_param[ii] = bary_param[ii] + random_real() / (2 * di);
        }
        Point midpoint = 
          pt_.patchmath(patchind0).get_real_point(perturb_param);
        for (int i = 0; i < di; ++i) {
          Real delta = (1.0 + random_real()) * (upper_right[i] - lower_left[i]);
          startpoint[i] = midpoint[i] - delta;
          endpoint[i] = midpoint[i] + delta;
        }
        hitlist.resize(0, lhr_default);
        Point segtangent = Point::subtract(endpoint,startpoint);
        degen = false;

#ifdef DEBUGGING
        if (trycount >= 398) {
          *debug_str << " shooting ray from " << fspec.fdim() 
                     << ":" << fspec.faceind() << " patch " << patchind0
                     << " bary_param = " << bary_param 
                     << " perturb_param = " << perturb_param 
                     << " midpoint = " << midpoint << '\n'
                     << " startpoint = " << startpoint << '\n';

          *debug_str << " ray shoot tangent = " << segtangent 
                     << " endpoint = " << endpoint << '\n'
                     << " scfac = " << scfac << " tol = " << tol << '\n';
        }
#endif
        for (Index patchind = 0; patchind < pt_.patches_.size(); ++patchind) {
          if (pt_.patches_[patchind].gdim != di - 1) continue;
          output_stack.resize(0, PatchMath::IntersectRecord());
          pt_.patchmath(patchind).intersect_seg(output_stack, workspace1,  
            startpoint, endpoint, scfac, tol, false);

          int degen_count = 0;
#ifdef DEBUGGING
          if (trycount >= 398) {
            *debug_str << " on patch " << patchind 
                       << " #hits = " << output_stack.size() << " \n";
          }
#endif
            
          for (int ii = 0; ii < output_stack.size(); ++ii) {
            if (output_stack[ii].degen) {
              degen = true;
            }
            rec.dist = output_stack[ii].dist;
            normal = pt_.patchmath(patchind).normal(output_stack[ii].paramcoord);
            Real ip = Point::inner_product(segtangent,normal);
            rec.front_hit = ip > 0;
            Real dist = output_stack[ii].dist;
            rec.patchind = patchind;

#ifdef DEBUGGING
            if (trycount >= 398) {
              *debug_str << " hit " << ii << " at distance " << rec.dist 
                         << " front hit = " << rec.front_hit << '\n'
                         << "  hit on patch " << rec.patchind 
                         << " front = " << rec.front_hit << " dist = "
                         << rec.dist 
                         << "degen = " << output_stack[ii].degen << '\n';
            }
#endif
            hitlist.push_back(rec);
          }
          if (degen) break;
        }
      }
      if (degen) {
        ostringstream ost;
        ost << "Intractable degeneracy in patchtable initialization";
        put_errmsg3_(ost, fspec);
        throw_error(ost.str());
      }
      
      sort(hitlist.begin(), hitlist.end());
      
      //  Now label all the front and backs of the chambers encountered.
      
      Brep::Face_Spec prev_chamber(-1,-1);
      Brep::Face_Spec next_chamber(-1,-1);
      
      bool zero_found = false;

// ost2 << " hitlistsize = " << hitlist.size() << '\n';
      
      for (int ii = 0; ii < hitlist.size(); ++ii) {
        Index patchind = hitlist[ii].patchind;
        Brep::Face_Spec owner = pt_.owner(patchind);

// ost2 << " hit " << ii << " on patch " << patchind << " owner " << owner << '\n';

        if (chamber1(owner) == prev_chamber) {
          next_chamber = chamber2(owner);
        }
        else if (chamber2(owner) == prev_chamber) {
          next_chamber = chamber1(owner);
        }
        else {
          ostringstream ost;
          ost << "Topological face " << brep_.face_name(owner) << " incident upon chamber ";
          if (prev_chamber.fdim() < 0) 
            ost << " [exterior of domain] ";
          else 
            ost << brep_.face_name(prev_chamber);
          ost << "\nbut " << brep_.face_name(owner) << " is not a descendent of that chamber";
          put_errmsg3_(ost, fspec);
          throw_error(ost.str());
        }

// ost2 << "next chamber = " << next_chamber << " prev chamber = " << prev_chamber  << "\n";

        Brep::Face_Spec front_face = 
          (hitlist[ii].front_hit)?  next_chamber : prev_chamber;
        Brep::Face_Spec back_face =
          (hitlist[ii].front_hit)?  prev_chamber : next_chamber;
 //       std::cerr << " for face " << owner.fdim() << ":" << owner.faceind() << " patch " << patchind << " front_face = "
  //        << front_face.fdim() << ":" << front_face.faceind() << " back_face = "
   //       << back_face.fdim() << ":" << back_face.faceind() << std::endl;

        if (front_back_computed(owner)) {
          if (front_face.faceind() != pt_.patches_[patchind].front_side
            || back_face.faceind() != pt_.patches_[patchind].back_side) {
            ostringstream ost;
            ost << "Inconsistent frontface/backface for topological face " <<
              brep_.face_name(owner) << " possibly indicating that\n"
            "patches cross illegally";
            put_errmsg3_(ost, fspec);
            throw_error(ost.str());
          }
        }
        else {
          front_back_computed(owner) = true;
          for (Index patchind1 = patchind_begin_(owner); 
          patchind1 < patchind_end_(owner);
          ++patchind1) {
            if (pt_.gdim(patchind1) == di - 1) {
              pt_.patches_[patchind1].front_side = front_face.faceind();
              pt_.patches_[patchind1].back_side = back_face.faceind();
            }
          }
        }
        if (ii > 0 && fspec.fdim() < di - 1 && hitlist[ii].dist > 0.5 &&
          hitlist[ii - 1].dist < 0.5) {
          zero_found = true;
          front_back_computed(fspec) = true;
          for (Index patchind1 = patchind_begin_(fspec);
          patchind1 < patchind_end_(fspec);
          ++patchind1) {
            pt_.patches_[patchind1].front_side = prev_chamber.faceind();
            pt_.patches_[patchind1].back_side = prev_chamber.faceind();
            if (prev_chamber != chamber1(fspec)) {
              ostringstream ost;
              ost << "Internal boundary topological face " << brep_.face_name(fspec)
                << " listed as lying in chamber " << brep_.face_name(chamber1(fspec))
                << "\nbut seems to lie in chamber ";
              if (prev_chamber.fdim() < 0) 
                ost << "[domain exterior]";
              else
                ost    << brep_.face_name(prev_chamber);
              put_errmsg3_(ost, fspec);
              throw_error(ost.str());
            }
          }
        }
        prev_chamber = next_chamber;
      }

// ost2 << " at loop end next_chamber = " << next_chamber << '\n';

      if (fspec.fdim() < di - 1 && !zero_found) {
        ostringstream ost;
        ost << "Internal boundary topological face " << brep_.face_name(fspec) 
          << " is external to all chambers.";
        put_errmsg3_(ost, fspec);
        throw_error(ost.str());
      }

#ifdef _MSC_VER
#if _MSC_VER == 1200
      Brep::Face_Spec tmpfspec = Brep::Face_Spec(-1,-1);
      if (next_chamber != tmpfspec) 
#else
      if (next_chamber != Brep::Face_Spec(-1,-1)) 
#endif
#else
      if (next_chamber != Brep::Face_Spec(-1,-1)) 
#endif
        {
        ostringstream ost;

// ost << ost2.str();

        ost << "Boundary fails to close or crosses itself.";
        put_errmsg3_(ost, fspec);
        throw_error(ost.str());
      }
    }
  }
  {
    for (Brep::Face_Spec_Loop_Over_Faces fspec(brep_);
    fspec.notdone();
    ++fspec) {
      if (is_direct_subface(fspec) && !front_back_computed(fspec)) {
        ostringstream ost;
        ost << brep_.face_name(fspec) << " has no patches.";
        put_errmsg3_(ost, fspec);
        throw_error(ost.str());
      }
    }
  }
}




bool QMG::MG::PatchInBrepAddress::operator<(const PatchInBrepAddress& ip) const {
  if (fspec < ip.fspec) return true;
  if (fspec > ip.fspec) return false;
  return patchind < ip.patchind;
}

bool 
QMG::MG::PatchInBrepAddress::operator==(const PatchInBrepAddress& ip) const {
  return fspec == ip.fspec && patchind == ip.patchind;
}
