// ------------------------------------------------------------------
// gm_vizp.cpp
//
// This file contains the driver function and worker function for 
// for gmvizp, a routine for processing a brep or simplicial
// complex and extracting points, edges or triangles for graphics.
// ------------------------------------------------------------------
// Author: Stephen A. Vavasis
// Copyright (c) 1999 by Cornell University.  All rights reserved.
// 
// See the accompanying file 'Copyright' for authorship information,
// the terms of the license governing this software, and disclaimers
// concerning this software.
// ------------------------------------------------------------------
// This file is part of the QMG software.  
// Version 2.0 of QMG, release date September 3, 1999.
// ------------------------------------------------------------------
#ifdef _MSC_VER
#if _MSC_VER == 1200
#include "qmatvec.h"
#endif
#endif

#include "qbrep.h"
#include "qsimpcomp.h"
#include "qmatvec.h"
#include "qfrontend.h"

namespace QMG {
  extern istream& eatwhite(istream&);
}
  

namespace {
  using namespace QMG;
  const int MAXDIM = 3;

  // ------------------------------------------------------------------
  // SimpKey: a data structure for keeping track of which simplices
  // have already been generated so that we don't do them twice, since
  // rendering may be slow.

  struct SimpKey {
    static int dim;
    SimpComplex::VertexOrdinalIndex verts[MAXDIM + 1];
    bool operator<(const SimpKey&) const;
    bool operator==(const SimpKey&) const;
    void sort();
  };

  int SimpKey::dim;

  bool SimpKey::operator<(const SimpKey& o) const {
    for (int j = 0; j < dim + 1; ++j) {
      if (verts[j] < o.verts[j]) return true;
      if (verts[j] > o.verts[j]) return false;
    }
    return false;
  }


  bool SimpKey::operator==(const SimpKey& o) const {
    for (int j = 0; j < dim + 1; ++j) {
      if (verts[j] != o.verts[j]) return false;
    }
    return true;
  }

  void SimpKey::sort() {
    for (int m = dim; m >= 1; --m) {
      for (int k = 0; k < m; ++k) {
        if (verts[k] > verts[k+1]) {
          SimpComplex::VertexOrdinalIndex tmp = verts[k];
          verts[k] = verts[k+1];
          verts[k+1] = tmp;
        }
      }
    }
  }


  struct Vizp_data {
    Matrix coords;
    IntMatrix simps;
    Matrix colors;
    Vizp_data(Matrix& coor1, IntMatrix& si1, Matrix& col1) :
      coords(coor1), simps(si1), colors(col1) { }

    // Destructive copy semantics.

  private:
    class Vizp_data_returnval {
    private:
      friend struct Vizp_data;
      Vizp_data& dref_;
      Vizp_data_returnval(Vizp_data& v, int) : dref_(v) { }
    };
  public:
    Vizp_data(Vizp_data& o) :
      coords(o.coords), simps(o.simps), colors(o.colors) { }
    Vizp_data(const Vizp_data_returnval& x) :
      coords(x.dref_.coords), simps(x.dref_.simps), colors(x.dref_.colors) { }
    operator Vizp_data_returnval() {
      return Vizp_data_returnval(*this, 0);
    }
  };


  // ------------------------------------------------------------------
  // This routine extracts coordinates and simplices from a simplicial
  // complex.

  Vizp_data get_pts_simps(const SimpComplex& sc, 
    const vector<Real>& default_color,
    int dim) {

    int di = sc.embedded_dim();
    Point::init_static_data(di);
    int gdim = sc.gdim();
    if (dim > gdim || dim < 0)
      throw_error("dim argument out of range in get_pts_simps");
    if (default_color.size() != 4)
      throw_error("Wrong color length in get_pts_simps");
    int numnodes = sc.num_nodes();
    int numelem = sc.num_elements();

    // Get the nodes
    Matrix vc(numnodes, di);
    for (SimpComplex::VertexOrdinalIndex vnum_o = 0; vnum_o < numnodes; ++vnum_o) {
      for (int i = 0; i < di; ++i)
        vc(vnum_o, i) = sc.real_coord_o(vnum_o,i);
    }

    // Get the simplices. Loop over subsets of indices with
    // exactly the right number of entries.

    vector<int> selected_verts;

    map<SimpKey, int> faces_drawn;
    SimpKey key;
    SimpKey::dim = dim;
    vector<SimpComplex::VertexOrdinalIndex> outputsimps;

    for (unsigned int subsetmask = 0; 
    subsetmask < (1 << (gdim + 1));
    ++subsetmask) {
      selected_verts.resize(0);
      for (int j = 0; j < gdim + 1; ++j) {
        if (subsetmask & (1<<j)) selected_verts.push_back(j);
      }
      int slvs = selected_verts.size();
      if (slvs != dim + 1) continue;
      
      // Get all the subfaces with the selected vertices.
      int levsize = sc.brep_levelsize(gdim);
      for (Brep::FaceIndex faceind = 0; faceind < levsize; ++faceind) {
        Brep::Face_Spec fspec(gdim, faceind);
        int nmf = sc.num_meshface_on_face(fspec);
        for (int seqno = 0; seqno < nmf; ++seqno) {
          for (int l = 0; l < dim + 1; ++l) {
            SimpComplex::VertexGlobalIndex vnum =
              sc.node_of_meshface_on_face(fspec, seqno, selected_verts[l]);
            key.verts[l] = sc.global_to_ordinal(vnum);
          }
          key.sort();
          if (faces_drawn[key] == 0) {
            faces_drawn[key] = 1;
            for (int l1 = 0; l1 < dim + 1; ++l1) {
              outputsimps.push_back(key.verts[l1]);
            }
          }
        }
      }
    }
    
    
    int outputcount = outputsimps.size() / (dim + 1);
    IntMatrix faces(outputcount, dim + 1);
    Matrix colors(outputcount, 4);
    int pos = 0;
    for (int ocount = 0; ocount < outputcount; ++ocount) {
      for (int l = 0; l < dim + 1; ++l) {
        faces(ocount, l) = outputsimps[pos++];
      }
      for (int cc = 0; cc < 4; ++cc) {
        colors(ocount, cc) = default_color[cc];
      }
    }
    return Vizp_data(vc, faces, colors);
  }
  
  struct Colorspec {
    Real co[4];
  };
  
  
  // ------------------------------------------------------------------
  // This routine is for parsing the "color" property of a brep face.
  // The corresponding value should be a four-tuple of reals (r g b alpha)

  Colorspec parse_color_propval(const Brep& g, const Brep::Face_Spec& fspec) {
    Colorspec returnval;
    string colorstring = g.face_lookup_prop_val(fspec, "color");
    if (colorstring.length() == 0) {
      returnval.co[0] = -1;
      return returnval;
    }
    istringstream ist(colorstring);
    char c = 0;
    char c2 = 0;
    ist >> eatwhite >> c >> eatwhite >> returnval.co[0] >>
      eatwhite >> returnval.co[1] >> eatwhite >> returnval.co[2] >> eatwhite
      >> returnval.co[3] >> eatwhite >> c2;
    if (!ist || c != '(' || c2 != ')') {
      ostringstream os;
      os << "Value associated with \"color\" property of face " << g.face_name(fspec)
        << " incorrectly formatted.\n"
        << "Should be (r g b alpha) where r,g,b,alpha are real numbers";
      throw_error(os.str());
    }
    return returnval;
  }
  
  
  // ------------------------------------------------------------------
  // This routine is for extracting simplices from a simplicial complex
  // in which we look up the color specs in the corresponding brep.

  
  Vizp_data get_pts_simps(const SimpComplex& sc,
    const Brep& g,
    const vector<Real>& default_color,
    int dim) {

    int di = sc.embedded_dim();
    Point::init_static_data(di);
    int gdim = sc.gdim();
    if (dim > gdim || dim < 0)
      throw_error("dim argument out of range in get_pts_simps");
    if (gdim > g.gdim() || di != g.embedded_dim())
      throw_error("mesh dimension does not match brep dimension");
    if (compare_nocase(g.lookup_brep_propval(Brep::global_id_propname()),
      sc.lookup_prop_val(Brep::global_id_propname())) != 0)
      throw_error("mesh's brep id does not match brep's id");

    if (default_color.size() != 4)
      throw_error("Wrong color length in get_pts_simps");
    int numnodes = sc.num_nodes();
    int numelem = sc.num_elements();

    // Make a table holding all the colors

    map<Brep::Face_Spec, Colorspec> colortable;

    for (Brep::Face_Spec_Loop_Over_Faces fspec(g); fspec.notdone(); ++fspec) {
      Colorspec facecolor = parse_color_propval(g,fspec);
      if (facecolor.co[0] < 0) {
        for (int j = 0; j < 4; ++j)
          facecolor.co[j] = default_color[j];
      }
      colortable[fspec] = facecolor;
    }
   
    SimpKey::dim = dim;

    // Get the nodes
    Matrix vc(numnodes, di);
    for (SimpComplex::VertexOrdinalIndex vnumo = 0; vnumo < numnodes; ++vnumo) {
      for (int i = 0; i < di; ++i)
        vc(vnumo, i) = sc.real_coord_o(vnumo,i);
    }

    // Get the simplices. Loop over subsets of indices with
    // exactly the right number of entries.

    vector<int> selected_verts;

    map<SimpKey, int> faces_drawn;
    SimpKey key;
    vector<SimpComplex::VertexOrdinalIndex> outputsimps;
    vector<Colorspec> colors;
    
    
    for (int fdim = 0; fdim <= gdim; ++fdim) {
      int levsize = g.level_size(fdim);
      if (levsize != sc.brep_levelsize(fdim))
        throw_error("Mismatch in brep and sc level sizes");
      
      
      for (unsigned int subsetmask = 0; 
      subsetmask < (1 << (fdim + 1));
      ++subsetmask) {
        selected_verts.resize(0);
        for (int j = 0; j < fdim + 1; ++j) {
          if (subsetmask & (1<<j)) selected_verts.push_back(j);
        }
        if (selected_verts.size() != dim + 1) continue;


        for (Brep::FaceIndex faceind = 0; faceind < levsize; ++faceind) {
          Brep::Face_Spec fspec(fdim, faceind);
          int nmf = sc.num_meshface_on_face(fspec);
          
          
          // Get all the subfaces with the selected vertices.
          
          for (int seqno = 0; seqno < nmf; ++seqno) {
            for (int l = 0; l < dim + 1; ++l) {
              SimpComplex::VertexGlobalIndex vnum = 
                sc.node_of_meshface_on_face(fspec, seqno, selected_verts[l]);
              key.verts[l] = sc.global_to_ordinal(vnum);
            }
            key.sort();
            if (faces_drawn[key] == 0) {
              faces_drawn[key] = 1;
              Colorspec cspec = colortable[fspec];
              for (int l1 = 0; l1 < dim + 1; ++l1) {
                outputsimps.push_back(key.verts[l1]);
              }
              colors.push_back(cspec);
            }
          }
        }
      }
    }
  

    int outputcount = outputsimps.size() / (dim + 1);
    IntMatrix faces(outputcount, dim + 1);
    Matrix colorso(outputcount, 4);
    int pos = 0;
    for (int ocount = 0; ocount < outputcount; ++ocount) {
      for (int l = 0; l < dim + 1; ++l) {
        faces(ocount, l) = outputsimps[pos++];
      }
      for (int cc = 0; cc < 4; ++cc) {
        colorso(ocount, cc) = colors[ocount].co[cc];
      }
    }
    return Vizp_data(vc, faces, colorso);
  }
    

  // ------------------------------------------------------------------
  // This routine is for extracting coordinates and simplices
  // from a brep.  The routine generates only straight segments
  // and triangles, so curved objects are approximated via subdivision.
  // The argument beziersub determines how many pieces to
  // in which to subdivide.


  Vizp_data get_pts_simps(const Brep& g, 
    const vector<Real>& default_color,
    int dim, 
    int beziersub) {

    int di = g.embedded_dim();
    Point::init_static_data(di);
    int gdim = g.gdim();
    if (dim > gdim || dim < 0)
      throw_error("dim argument out of range in get_pts_simps");
    if (default_color.size() != 4)
      throw_error("Wrong color length in get_pts_simps");
    if (beziersub < 0 || beziersub > 10000)
      throw_error("Beziersub out of range in get_pts_simps");
    Point realcoord;
    Point paramcoord;
    int pcount = 0;
    vector<Real> pt_coords;
    vector<int> simp_vtx;
    vector<Colorspec> colors;
    for (Brep::Face_Spec_Loop_Over_Faces_Of_Dim fspec(g, dim);
    fspec.notdone();
    ++fspec) {

      Colorspec facecolor = parse_color_propval(g,fspec);
      if (facecolor.co[0] < 0) {
        for (int j = 0; j < 4; ++j)
          facecolor.co[j] = default_color[j];
      }

      for (Brep::Loop_over_patches_of_face patch(g, fspec);
      patch.notdone();
      ++patch) {
        int pbase = pcount;
        if (dim == 0) {
          Brep::ControlPointIndex cpi = patch.control_point(0);  
          for (int i = 0; i < di; ++i)
            pt_coords.push_back(g.control_point(cpi,i));
          ++pcount;      
          simp_vtx.push_back(pbase);
          colors.push_back(facecolor);

        }
        else if (dim == 1) {
          int subdiv = (patch.degree1() == 1)? 1 : beziersub;
          for (int vcount = 0; vcount <= subdiv; ++vcount) {
            paramcoord[0] = static_cast<double>(vcount) /
              static_cast<double>(subdiv);
            realcoord = g.patchmath(fspec, patch.index()).
              get_real_point(paramcoord);
            for (int i = 0; i < di; ++i)
              pt_coords.push_back(realcoord[i]);
            ++pcount;
          }
          for (int sub = 0; sub < subdiv; ++sub) {
            simp_vtx.push_back(pbase + sub);
            simp_vtx.push_back(pbase + sub + 1);
            colors.push_back(facecolor);
          }
        }
        else if (dim == 2) {
          if (patch.ptype() == BEZIER_TRIANGLE) {
            int subdiv = (patch.degree1() == 1)? 1 : beziersub;
            {
              for (int rownum = 0; rownum <= subdiv; ++rownum) {
                for (int colnum = 0; colnum <= rownum; ++colnum) {
                  paramcoord[0] = static_cast<double>(colnum) / 
                    static_cast<double>(subdiv);
                  paramcoord[1] = static_cast<double>(rownum-colnum) /
                    static_cast<double>(subdiv);
                  realcoord = g.patchmath(fspec, patch.index()).
                    get_real_point(paramcoord);
                  for (int i = 0; i < di; ++i)
                    pt_coords.push_back(realcoord[i]);
                  ++pcount;
                }
              }
            }
            int sub = 0;
            {
              for (int rownum = 0; rownum < subdiv; ++rownum) {
                for (int colnum = 0; colnum <= rownum; ++colnum) {
                  simp_vtx.push_back(pbase + sub);
                  simp_vtx.push_back(pbase + sub + rownum + 1);
                  simp_vtx.push_back(pbase + sub + rownum + 2);
                  colors.push_back(facecolor);
                  if (rownum < subdiv - 1) {
                    simp_vtx.push_back(pbase + sub + rownum + 2);
                    simp_vtx.push_back(pbase + sub + rownum + 1);
                    simp_vtx.push_back(pbase + sub + 2 * rownum + 4);
                    colors.push_back(facecolor);
                  }
                  ++sub;
                }
              }
            }
          }
          else { //patchtype == QUAD
            int subdiv1 = beziersub;
            int subdiv2 = beziersub;
            {
              for (int rownum = 0; rownum <= subdiv2; ++rownum) {
                for (int colnum = 0; colnum <= subdiv1; ++colnum) {
                  paramcoord[0] = static_cast<double>(colnum) / 
                    static_cast<double>(subdiv1);
                  paramcoord[1] = static_cast<double>(rownum) /
                    static_cast<double>(subdiv2);
                  realcoord = g.patchmath(fspec, patch.index()).
                    get_real_point(paramcoord);
                  for (int i = 0; i < di; ++i)
                    pt_coords.push_back(realcoord[i]);
                  ++pcount;
                }
              }
            }
            {
              for (int rownum = 0; rownum < subdiv2; ++rownum) {
                for (int colnum = 0; colnum < subdiv1; ++colnum) {
                  int startpoint = colnum + (rownum * (subdiv1 + 1));
                  simp_vtx.push_back(pbase + startpoint);
                  simp_vtx.push_back(pbase + startpoint + subdiv1 + 2);
                  simp_vtx.push_back(pbase + startpoint + subdiv1 + 1);
                  simp_vtx.push_back(pbase + startpoint);
                  simp_vtx.push_back(pbase + startpoint + 1);
                  simp_vtx.push_back(pbase + startpoint + subdiv1 + 2);
                  for (int c0 = 0; c0 < 2; ++c0) {
                    colors.push_back(facecolor);
                  }
                }
              }
            }
          }
        }
      }
    }
    int numsimp = simp_vtx.size() / (dim + 1);
#ifdef DEBUGGING
    if (pt_coords.size() != pcount * di ||
      simp_vtx.size() != numsimp * (dim + 1) ||
      colors.size() != numsimp)
      throw_error("Array size error in gm_vizp");
#endif

    int pos = 0;
    Matrix coord(pcount,di);
    for (int j1 = 0; j1 < pcount; ++j1)
      for (int d = 0; d < di; ++d)
        coord(j1,d) = pt_coords[pos++];
    int pos1 = 0;
    IntMatrix simpv(numsimp, dim + 1);
    Matrix colorm(numsimp, 4);
    for (int j2 = 0; j2 < numsimp; ++j2) {
      for (int d = 0; d < dim + 1; ++d)
        simpv(j2, d) = simp_vtx[pos1++];
      for (int c = 0; c < 4; ++c)
        colorm(j2, c) = colors[j2].co[c];
    }

    return Vizp_data(coord, simpv, colorm);
  }
  

#define GM_ROUTINE_NAME gm_vizp
#define GM_ROUTINE_NAME_Q "gm_vizp"


  // ------------------------------------------------------------------
  // This is the driver for gm_vizp and is a standard driver.
  
  static void worker(const QMG::FrontEnd::ArgValType& argvalhandle,
    QMG::FrontEnd::ReturnValType& returnvalhandle,
    QMG::FrontEnd::Interpreter& interp) {
    
    using namespace QMG;
    using namespace QMG::FrontEnd;
    
    argvalhandle.verify_nargin(4, 5, GM_ROUTINE_NAME_Q);
    returnvalhandle.verify_nargout(3,3, GM_ROUTINE_NAME_Q);

    // possible calling sequences:

    // gm_vizp(brep, default_color, dim, bezier_sub)
    // gm_vizp(simpcomp, default_color, dim, unused)
    // gm_vizp(simpcomp, brep, default_color, dim, unused)

    if (argvalhandle.nargin() == 4) {
      Object_Type_Code code = argvalhandle.determine_argument_type(0);
      if (code == UNKNOWN)
        throw_error("Argument 0 could not be interpreted as brep or simpcomplex");      
      if (code == BREP) {      
        Vizp_data vizpdata = get_pts_simps(argvalhandle.get_brep(0), 
          argvalhandle.get_vector_double(1),
          argvalhandle.get_int(2), 
          argvalhandle.get_int(3));
        returnvalhandle.put_matrix(0, vizpdata.coords);
        returnvalhandle.put_intmatrix(1, vizpdata.simps);
        returnvalhandle.put_matrix(2, vizpdata.colors);
        return;
      }
      else {      
        Vizp_data vizpdata = get_pts_simps(argvalhandle.get_simpcomp(0), 
          argvalhandle.get_vector_double(1),
          argvalhandle.get_int(2)); 
        returnvalhandle.put_matrix(0, vizpdata.coords);
        returnvalhandle.put_intmatrix(1, vizpdata.simps);
        returnvalhandle.put_matrix(2, vizpdata.colors);
        return;
      }
    }
 
    else { // nargin = 5
      Vizp_data vizpdata = get_pts_simps(argvalhandle.get_simpcomp(0),
        argvalhandle.get_brep(1),
        argvalhandle.get_vector_double(2),
        argvalhandle.get_int(3));
      returnvalhandle.put_matrix(0, vizpdata.coords);
      returnvalhandle.put_intmatrix(1, vizpdata.simps);
      returnvalhandle.put_matrix(2, vizpdata.colors);
    }
    return;
  }
}

#include "gm_entrypoint.cpp"
