// ------------------------------------------------------------------
// polytri.cpp
//
// This file contains a function for triangulating a planar polygon
// or PSLG using CDT.
// ------------------------------------------------------------------
// Author: Stephen A. Vavasis
// Copyright (c) 1999 by Cornell University.  All rights reserved.
// 
// See the accompanying file 'Copyright' for authorship information,
// the terms of the license governing this software, and disclaimers
// concerning this software.
// ------------------------------------------------------------------
// This file is part of the QMG software.  
// Version 2.0 of QMG, release date September 3, 1999.
// ------------------------------------------------------------------

#ifdef _MSC_VER
#if _MSC_VER == 1200
#include "qmatvec.h"
#endif
#endif

#include "qpolytri.h"
#include "qmatvec.h"

#include <fstream>


namespace QMG {
  namespace PolyTri {
    Triangulation
      pslg_constrained_delaunay_triangulate(VertexList& vtxlist,
      const vector<Edge>& elist);
    Triangulation poly_cdt(const Matrix& vtxlist, 
      const vector<Edge>& elist,
      double tol);
  }
}

namespace {
  using namespace QMG;
  using namespace QMG::PolyTri;


  // The triangulation algorithm maintains a front, which is a
  // heap of labeled distances.  The label is an int.

  struct Labeled_dist {
    Real dist;
    int label;
    Labeled_dist(Real d, int l) : dist(d), label(l) { }
    Labeled_dist(const Labeled_dist& o) : dist(o.dist), label(o.label) { }
    Labeled_dist() : dist(0), label(-1) { }
    bool operator==(const Labeled_dist& other) const;
    bool operator<(const Labeled_dist& other) const;
  };



  bool Labeled_dist::operator==(const Labeled_dist& other) const {
    if (dist == other.dist && label == other.label) return true;
    return false;
  }

  bool Labeled_dist::operator<(const Labeled_dist& other) const {
    if (dist < other.dist || (dist == other.dist && label < other.label))
      return true;
    return false;
  }
  Labeled_dist null_labeled_dist;

  // This routine projects three-dimensional vertices into 2D.
  // orthogonally to the plane containing the vertices.
  // The plane is computed via QR factorization with pivoting.


  void make_2d_vertices(const Matrix& vtxlist, VertexList& vtxlist2,
    double tol) {
    int m = vtxlist.numrows();
    int n = vtxlist.numcols();
    if (n < 2 || m < 3) {
      throw_error("Range error in make_2d_vertices");
    }
    if (n == 2) {
      for (int k = 0; k < m; ++k) {
        vtxlist2.add_vertex(vtxlist(k,0), vtxlist(k,1));
      }
      return;
    }
    Matrix A(n,m - 1);
    vector<Real> hhtrans(n);
    Real Afrob = 0.0;
    {
      for (int i = 1; i < m; ++i) {
        for (int j = 0; j < n; ++j) {
          A(j, i - 1) = vtxlist(i,j) - vtxlist(0,j);
          Afrob += A(j,i-1) * A(j,i-1);
        }
      }

    }
    Afrob = sqrt(Afrob);

    {
      for (int l = 0; l < 2; ++l) {
        Real maxnorm = 0.0;
        int pivcol = 0;
        for (int p = 0; p < m - 1; ++p) {
          Real thiscolnorm = 0.0;
          for (int i = l; i < n; ++i) {
            thiscolnorm += A(i,p) * A(i,p);
          }
          if (thiscolnorm > maxnorm) {
            maxnorm = thiscolnorm;
            pivcol = p;
          }
        }
        Real beta = Matrix::compute_hh_transform(&hhtrans[l], 
          &A(l,pivcol), m - 1, n - l);


        {
          for (int j = 0; j < m - 1; ++j) {
            Real ip = 0.0;
            {
              for (int i = l; i < n; ++i) {
                ip += hhtrans[i] * A(i,j);
              }
            }
            Real beta1 = beta * ip;
            {
              for (int i = l; i < n; ++i) {
                A(i,j) -= beta1 * hhtrans[i];
              }
            }
          }
        }
      }
    }
    {
      Real resdfro = 0.0;
      for (int j = 0; j < m - 1; ++j) {
        for (int i = 2; i < n; ++i) {
          resdfro += A(i,j) * A(i,j);
        }
      }
      resdfro = sqrt(resdfro);
      if (resdfro >= tol * Afrob)
        throw_error("Polygon vertices exceed non-coplanarity tolerance");
    }
    vtxlist2.add_vertex(0.0, 0.0);
    {
      for (int j = 0; j < m - 1; ++j) {
        vtxlist2.add_vertex(A(0,j), A(1,j));
      }
    }
  }

  // These two structs are used by search tri for walking from
  // triangle to triangle.

  enum Triangle_Label {NOT_YET_SEARCHED, KEEP, DELETE};
  
  struct Two_Int {
    int first;
    int second;
    Two_Int() : first(-1), second(-1) { }
    Two_Int(const Two_Int& other) : first(other.first), second(other.second) { }
    Two_Int& operator=(const Two_Int& o) {
      first = o.first;
      second = o.second;
      return *this;
    }
    ~Two_Int() { }
  };

  
  // This routine takes as input a triangulation and labels
  // the triangles according to whether they are inside
  // or outside the polygon.  This is determined by
  // starting from the outer boundary (whose triangles
  // are not in the polygon) and then walking from
  // triangle to neighboring triangle.  Each time
  // a boundary is crossed, the disposition changes.


  void search_tri(const vector<Edge>& elist, 
    const Triangulation& init_tri, 
    vector<Triangle_Label>& disposition) {

    map<Edge,int> countmap;
    int nume = elist.size();

    {
      for (int k = 0; k < nume; ++k) {
        ++countmap[elist[k].sort()];
      }
    }


    int nt = init_tri.numtri();

    // Map edges to the two triangles that contain them.
    map<Edge,Two_Int> tri_occur;
    {
      for (int l = 0; l < nt; ++l) {
        for (int j = 0; j < 3; ++j) {
          Edge e(init_tri.tri_vertex(l,j), init_tri.tri_vertex(l, (j+1) % 3));
          e = e.sort();
          Two_Int twoint = tri_occur[e];
          if (twoint.first == -1)
            twoint.first = l;
          else if (twoint.second == -1)
            twoint.second = l;
          else
            throw_error("edge adjacent to >2 triangles");
          tri_occur[e] = twoint;
        }
      }
    }

    // Pick a starting triangle that was labeled by the caller.

    int starttri;
    for (starttri = 0; starttri < nt; ++starttri) {
      if (disposition[starttri] != NOT_YET_SEARCHED)
        break;
    }
    if (starttri == nt)
      throw_error("No starting triangle found");

    vector<int> search_stack;
    search_stack.push_back(starttri);

    // Now search all triangles with a stack.

    while (search_stack.size() > 0) {
      int curtri = search_stack.back();
      search_stack.pop_back();
      Triangle_Label thislab = disposition[curtri];
      if (thislab == NOT_YET_SEARCHED)
        throw_error("Stack inconsistency");
      for (int j = 0; j < 3; ++j) {
        Edge e(init_tri.tri_vertex(curtri, j), 
          init_tri.tri_vertex(curtri, (j+1) % 3));
        e = e.sort();
        Two_Int twoint = tri_occur[e];
        int nbrtri = -1;
        if (twoint.first == curtri)
          nbrtri = twoint.second;
        else if (twoint.second == curtri)
          nbrtri = twoint.first;
        if (nbrtri < 0)
          throw_error("Inconsistent map state");
        int occ_cnt = countmap[e];
        Triangle_Label newlab;
        if ((occ_cnt % 2) ^ (thislab == KEEP)) {
          newlab = KEEP;
        }
        else {
          newlab = DELETE;
        }
        if (disposition[nbrtri] == NOT_YET_SEARCHED) {
          disposition[nbrtri] = newlab;
          search_stack.push_back(nbrtri);
        }
        else {
          if (disposition[nbrtri] != newlab)
            throw_error("Illegal polygon -- boundary does not close");
        }
      }
    }

    {
      for (int l = 0; l < nt; ++l) {
        if (disposition[l] == NOT_YET_SEARCHED)
          throw_error("Disconnected triangulation?");
      }
    }
  }
}


namespace QMG {
  namespace PolyTri {
    ostream& operator<<(ostream& os, const Edge& e) {
      os << "Ed:v" << e.endpoint(0) << "/v" << e.endpoint(1) << '/';
      return os;
    }
  }
}


// #define POLYTRILOG

// ------------------------------------------------------------------
// This routine computes the constrained Delaunay triangulation (CDT)
// of a planar straight line graph (PSLG).

QMG::PolyTri::Triangulation 
QMG::PolyTri::pslg_constrained_delaunay_triangulate(VertexList& vtxlist,
                                                    const vector<Edge>& elist) {


  // Make a list of all constraint edges.
  // Use a map to prevent duplicates.

  vector<Edge> constraint;
  vector<bool> top_done;
  vector<bool> bottom_done;
  map<Edge,int> checkedge;

  using std::advance;



#ifdef POLYTRILOG

  using std::endl;
  std::ofstream dbstream("pslglog");
  ostream* debug_str = &dbstream;
  for (int nv1 = 0; nv1 < vtxlist.size(); ++nv1) {
    *debug_str << "vertex v" << nv1 << "/ at " << vtxlist.coord(nv1,0) 
      << ',' << vtxlist.coord(nv1,1) << endl;
  }
#endif


  {
    int numv = vtxlist.size();
    for (int j = 0; j < elist.size(); ++j) {
      Edge e = elist[j];
      for (int k = 0; k < 2; ++k) {
      if (e.endpoint(k) < 0 || e.endpoint(k) >= numv)
        throw_error("Entry in edge list out of range");
      }
      if (e.endpoint(0) == e.endpoint(1)) continue;
      if (checkedge.find(e.sort()) == checkedge.end()) {

        int edgeind = constraint.size();
        constraint.push_back(e);
        top_done.push_back(false);
        bottom_done.push_back(false);
        checkedge[e.sort()] = edgeind;
#ifdef POLYTRILOG 
        *debug_str << " adding edge " << e << " at pos " << edgeind << endl;
#endif
      }
      else {
#ifdef POLYTRILOG
        *debug_str << " skipping edge " << e << endl;
#endif
      }
    }
  }

  // Add the bounding box to the stack, and four new vertices.

  {
    Real minx = BIG_REAL;
    Real maxx = -BIG_REAL;
    Real miny = BIG_REAL;
    Real maxy = -BIG_REAL;
    for (int j = 0; j < vtxlist.size(); ++j) {
      Real x = vtxlist.coord(j,0);
      if (x < minx)
        minx = x;
      if (x > maxx)
        maxx = x;
      Real y = vtxlist.coord(j,1);
      if (y < miny)
        miny = y;
      if (y > maxy)
        maxy = y;
    }

    if (vtxlist.size() == 0) {
      minx = 0;
      maxx = 1;
      miny = 0;
      maxy = 1;
    }

    Real deltax = maxx - minx;
    if (deltax == 0.0)
      deltax = fabs(minx) + 1;
    Real deltay = maxy - miny;
    if (deltay == 0.0)
      deltay = fabs(miny) + 1;
    Real lbx = minx - deltax;
    Real ubx = maxx + deltax;
    Real lby = miny - deltay;
    Real uby = maxy + deltay;
    VertexIndex newv1 = vtxlist.add_vertex(lbx,lby);
    VertexIndex newv2 = vtxlist.add_vertex(ubx,lby);
    VertexIndex newv3 = vtxlist.add_vertex(ubx,uby);
    VertexIndex newv4 = vtxlist.add_vertex(lbx,uby);
#ifdef POLYTRILOG
    *debug_str << " lbx, ubx, lby, uby = " << lbx << ',' << ubx << ',' 
      << lby << ',' <<  uby << endl;
    *debug_str << " new verts = v" << newv1 << "/ v" << newv2 << "/ v"
      << newv3 << "/ v" << newv4 << "/" << endl;
#endif
    int sz = constraint.size();
    constraint.push_back(Edge(newv1,newv2));
    top_done.push_back(false);
    bottom_done.push_back(true);
    checkedge[Edge(newv1,newv2).sort()] = sz++;
    constraint.push_back(Edge(newv2,newv3));
    top_done.push_back(false);
    bottom_done.push_back(true);
    checkedge[Edge(newv2,newv3).sort()] = sz++;
    constraint.push_back(Edge(newv3,newv4));
    top_done.push_back(false);
    bottom_done.push_back(true);
    checkedge[Edge(newv3,newv4).sort()] = sz++;
    constraint.push_back(Edge(newv4,newv1));
    top_done.push_back(false);
    bottom_done.push_back(true);
    checkedge[Edge(newv4,newv1).sort()] = sz++;
  }

  Matrix transf(2,2);
  Matrix smalleq(2,2);
  Real smallrhs[2];
  Real smallsoln[2];
  vector<Labeled_dist> vtxqueue;
  vector<Labeled_dist> edgequeue;
  VertexList tvtxlist;
  vector<Real> tan_point_x;
  vector<Real> tan_point_y;
  vector<Real> slope_c;
  vector<Real> slope_d;
  int numv = vtxlist.size();
  Triangulation tri;

  // Main loop: generate a triangle for each edge in the workstack.

  for (int eloop = 0; eloop < constraint.size() * 2; ++eloop) {
    int edgeind1 = eloop / 2;
    Edge e = constraint[edgeind1];
    if (eloop & 1) {
      if (top_done[edgeind1])
        continue;
      top_done[edgeind1] = true;
    }
    else {
      if (bottom_done[edgeind1])
        continue;
      bottom_done[edgeind1] = true;
      e = e.reverse();
    }

#ifdef POLYTRILOG
    *debug_str << "\n\n+++Main loop processing edge "<< e << endl;
#endif
    int nume = constraint.size();

    // Find a tranformation so that the two endpoints of the edge are at (-1,0) and (1,0).

    VertexIndex startv = e.endpoint(0);
    VertexIndex endv = e.endpoint(1);
    Real stx = vtxlist.coord(startv, 0);
    Real sty = vtxlist.coord(startv, 1);
    Real enx = vtxlist.coord(endv, 0);
    Real eny = vtxlist.coord(endv, 1);

    Real xdisp = enx - stx;
    Real ydisp = eny - sty;
    Real elength2 = xdisp * xdisp + ydisp * ydisp;

    if (elength2 == 0) continue;
    Real mult = 2.0 / elength2;

    transf(0,0) = xdisp * mult;
    transf(0,1) = ydisp * mult;
    transf(1,0) = -ydisp * mult;
    transf(1,1) = xdisp * mult;
    Real offsetx = (stx + enx) * 0.5;
    Real offsety = (sty + eny) * 0.5;

    tvtxlist.resize(0);
    {
      for (int j = 0; j < numv; ++j) {
        Real tmpx = vtxlist.coord(j,0) - offsetx;
        Real tmpy = vtxlist.coord(j,1) - offsety;
        Real newx = transf(0,0) * tmpx + transf(0,1) * tmpy;
        Real newy = transf(1,0) * tmpx + transf(1,1) * tmpy;
        if (j == startv || j == endv) {
          newy = 0.0;
          if (j == startv)
            newx = -1.0;
          else
            newx = 1.0;
        }
#ifdef POLYTRILOG
        *debug_str << " tvtx new vertex v" << j << "/ at "
          << newx << "," << newy << endl;
#endif
        tvtxlist.add_vertex(newx, newy);
      }
    }

    // For each vertex, compute its p-value and insert it into the
    // queue.

    vtxqueue.resize(0, null_labeled_dist);
    {
      for (int j = 0; j < numv; ++j) {
        Real y = tvtxlist.coord(j,1);
        if (y <= 0) continue;
        Real x = tvtxlist.coord(j,0);
        Real p = (x * x + y * y - 1) / (2 * y);
#ifdef POLYTRILOG
        *debug_str << " new vtxqueue entry v" << j << "/ with p = " << p << endl;
#endif
        vtxqueue.push_back(Labeled_dist(p,j));
      }
    }

    // For each constraint edge, compute its p-value and insert it into
    // the queue if its tangent point is in the segment's
    // relative interior.

    edgequeue.resize(0, null_labeled_dist);    
    tan_point_x.resize(nume);
    tan_point_y.resize(nume);
    slope_c.resize(nume);
    slope_d.resize(nume);
    {
      for (int j = 0; j < nume; ++j) {
#ifdef POLYTRILOG
        *debug_str << "Considering constraint " << constraint[j] << endl;
#endif
        VertexIndex startv1 = constraint[j].endpoint(0);
        VertexIndex endv1 = constraint[j].endpoint(1);
        if (startv == startv1 && endv == endv1) continue;
        if (startv == endv1 && endv == startv1) continue;
        Real stx1 = tvtxlist.coord(startv1,0);
        Real sty1 = tvtxlist.coord(startv1,1);
        Real enx1 = tvtxlist.coord(endv1,0);
        Real eny1 = tvtxlist.coord(endv1,1);
        if (startv == startv1 && eny1 <= 0.0) continue;
        if (endv == startv1 && eny1 <= 0.0) continue;
        if (startv == endv1 && sty1 <= 0.0) continue;
        if (endv == endv1 && sty1 <= 0.0) continue;
        smalleq(0,0) = stx1;
        smalleq(0,1) = sty1;
        smalleq(1,0) = enx1;
        smalleq(1,1) = eny1;
        smallrhs[0] = 1;
        smallrhs[1] = 1;
        Real condest = smalleq.linear_solve(smallrhs, smallsoln);
        if (condest == BIG_REAL) continue;
        Real c = smallsoln[0];
        Real d = smallsoln[1];
#ifdef POLYTRILOG
        *debug_str << " for this constraint, c, d = " << c << ',' <<  d << endl;
#endif
        Real cdlen = c * c + d * d;
        Real qa = c * c;
        Real qb = 2.0 * d;
        Real qc = cdlen - 1.0;
        Real discr;
        Real ynum;

        if (startv == startv1 || startv == endv1 || 
          endv == startv1 || endv == endv1) {
          discr = 0.0;
          ynum = 0;
        }
        else {
          ynum = 1.0 - c * c;
          discr = 4.0 * cdlen * ynum;
        }
#ifdef POLYTRILOG
        *debug_str << " for this constraint, ynum = " << ynum << endl;
#endif
        if (ynum < 0.0) continue;
        Real p;
        // numerically stable quadratic formula

        if (c == 0.0 && d < 0.0) continue;

        if (d <= 0.0)
          p = (-qb + sqrt(discr)) / (2 * qa);
        else
          p = 2.0 * qc / (-qb - sqrt(discr));
#ifdef POLYTRILOG
        *debug_str << " for this constraint p = " << p << endl;
#endif

        // compute y coord
        Real y = sqrt(ynum / cdlen);
        // Numerically stable way to recover x
        Real x;
        if (startv == startv1 || startv == endv1)
          x = -1.0;
        else if (endv == startv1 || endv == endv1)
          x = 1.0;
        else if (fabs(d) > fabs(c))
          x = (y - p) * c / d;
        else
          x = (1.0 - d * y) / c;

#ifdef POLYTRILOG
        *debug_str << " for this constraint tan point = " << x << ',' << y << " versus endpoints "
          << stx1 << ',' << sty1 << " and " << enx1 << ',' << eny1 << endl;
#endif

        //Check whether (x,y) is actually in the segment.

        if ((x - stx1) * (x - enx1) + (y - sty1) * (y - eny1) > 0) continue;

        // Put it in the queue.
        tan_point_x[j] = x;
        tan_point_y[j] = y;
        slope_c[j] = c;
        slope_d[j] = d;
        edgequeue.push_back(Labeled_dist(p,j));
#ifdef POLYTRILOG
        *debug_str << "new edgequeue entry " << constraint[j] << " tangent at point "
          << tan_point_x[j] << "," << tan_point_y[j] << " slope = " << slope_c[j] << ','
          << slope_d[j] << " p = " << p << endl;
#endif

      }
    }

    // Sort the two queues.
    
    sort(edgequeue.begin(), edgequeue.end());
    sort(vtxqueue.begin(), vtxqueue.end());
    
    // Now start growing the circle until we hit something.
    
    int vqptr = 0;
    int eqptr = 0;
    
    typedef multimap<Real,int> Front_type;
    typedef Front_type::iterator FTI;
    Front_type front;
#ifdef POLYTRILOG
    *debug_str << " \nbeginning front search " << endl;
#endif

    for (;;) {
      if (vqptr >= vtxqueue.size())
        throw_error("Unexpected end of vertex queue in polytri");
      
      // Check if the next thing to add is a vertex. 
      
      bool hit_on_vertex;
      Real x,y,p;
      VertexIndex thisvtxind;
      int thisedgeind;
      
      if (eqptr >= edgequeue.size() || 
        edgequeue[eqptr].dist >= vtxqueue[vqptr].dist) {
        hit_on_vertex = true;
        thisvtxind = vtxqueue[vqptr].label;
        p = vtxqueue[vqptr].dist;
        x = tvtxlist.coord(thisvtxind,0);
        y = tvtxlist.coord(thisvtxind,1);
#ifdef POLYTRILOG
        *debug_str << " processing front vertex v" << thisvtxind << "/ at x,y " << x << ','
          << y << " p = " << p << endl;
#endif
      }
      else {
        hit_on_vertex = false;
        thisedgeind = edgequeue[eqptr].label;
        p = edgequeue[eqptr].dist;
        x = tan_point_x[thisedgeind];
        y = tan_point_y[thisedgeind];
#ifdef POLYTRILOG
        *debug_str << " processing front edge" << constraint[thisedgeind] << " at x,y " << x << ','
          << y << " p = " << p << endl;
#endif
      }
      
      // Check if the point x,y is hidden by the front.
      // Find the angle that (x,y) makes w.r.t the center
      // of the circle at (0,p).
      
      Real ang = atan2(y - p,x);
#ifdef POLYTRILOG
      *debug_str << " ang = " << ang << endl;
#endif
      int checkcount;
      int checklist[2];
      if (front.size() == 0)
        checkcount = 0;
      else {
        FTI next = front.lower_bound(ang);
        FTI prev;
        if (next == front.begin()) {
          prev = front.end();
          advance(prev, -1);
        }
        else {
          prev = next;
          advance(prev, -1);
        }
        if (next == front.end()) {
          next = front.begin();
        }
        if (prev == next) {
          checkcount = 1;
          checklist[0] = prev -> second;
        }
        else {
          checkcount = 2;
          checklist[0] = prev -> second;
          checklist[1] = next -> second;
        }
      }
      bool hidden = false;
      for (int checki = 0; checki < checkcount; ++checki) {
        // Determine if (x,y) is hidden.
        // If the test point is a vertex of the edge we
        // are checking, then it is not hidden.
        
        int cedgeind = checklist[checki];
        if (hit_on_vertex &&
          (constraint[cedgeind].endpoint(0) == thisvtxind ||
          constraint[cedgeind].endpoint(1) == thisvtxind))
          continue;
#ifdef POLYTRILOG
        *debug_str << " checking against edge " << constraint[cedgeind] << endl;
#endif
        // Intersect the segment from the new point to the origin
        //  with the support line of the constraint.
        
        smalleq(0,0) = slope_c[cedgeind];
        smalleq(0,1) = slope_d[cedgeind];
        smallrhs[0] = 1.0;
        smalleq(1,0) = y;
        smalleq(1,1) = -x;
        smallrhs[1] = 0;
        Real condest = smalleq.linear_solve(smallrhs,smallsoln);
        if (condest > 1e16) continue;
#ifdef POLYTRILOG
        *debug_str << " smallsoln = " << smallsoln[0] <<',' << smallsoln[1]
          << " y = " << y << endl;
#endif
        if (fabs(y) > fabs(x)) {
          if (smallsoln[1] >= 0 && smallsoln[1] <= y && smallsoln[0]) {
            hidden = true;
          }
        }
        else {
          if (fabs(smallsoln[0]) <= fabs(x) && smallsoln[0] * x >= 0) {
            hidden = true;
          }
        }
        if (hidden) break;
      }
#ifdef POLYTRILOG
      *debug_str << "hidden = " << hidden << endl;
#endif
      if (hit_on_vertex) {
        if (!hidden) {
          // The hit is on a vertex, and the vertex is not hidden,
          // then we have found the sought-after triangle and we
          // can emit it.
          tri.add_triangle(startv, endv, thisvtxind);
#ifdef POLYTRILOG
          *debug_str << " adding triangle v" << startv << "/ v" << endv << "/ v"
            << thisvtxind << "/" << endl;
#endif
          // Add two new constraint edges, if not already present.
          for (int k = 0; k < 2; ++k) {
            VertexIndex ep1 = (k == 0)? endv : thisvtxind;
            VertexIndex ep2 = (k == 0)? thisvtxind : startv;
            Edge enew(ep1, ep2);
            if (checkedge.find(enew.sort()) == checkedge.end()) {
              int sz = constraint.size();
              constraint.push_back(enew);
              top_done.push_back(true);
              bottom_done.push_back(false);
              checkedge[enew.sort()] = sz;
#ifdef POLYTRILOG
              *debug_str << " adding constraint edge " << enew << " at position " << sz << endl;
#endif
            }
            else {
              int j1 = checkedge[enew.sort()];
              if (ep1 == constraint[j1].endpoint(0)) {
                if (top_done[j1]) 
                  throw_error("Top of edge done twice?");
                top_done[j1] = true;
              }
              else {
                if (bottom_done[j1])
                  throw_error("Bottom of edge done twice?");
                bottom_done[j1] = true;
              }
            }
          }
          break;  //end of front;
        }
        else { // hidden
          ++vqptr;
        }
      } 
      else { // hit is on edge
        if (!hidden) {
          // Add it to the front.
#ifdef POLYTRILOG
          *debug_str << " adding edge " << constraint[thisedgeind] << " to front" << endl;
#endif
          front.insert(pair<Real,int>(ang,thisedgeind));
        }
        ++eqptr;
      }
    } // Done with finding triangle 
  } // End of main work loop.
  return tri;
}






// ------------------------------------------------------------------
// poly_cdt computes the CDT of a polygon.  It first computes the
// CDT of the corresponding PSLG, then it calls searchtri to 
// get rid of unused triangles.

QMG::PolyTri::Triangulation 
QMG::PolyTri::poly_cdt(const Matrix& vtxlist, 
                       const vector<Edge>& elist,
                       double tol) {
  VertexList vl;
  make_2d_vertices(vtxlist, vl, tol);
  Triangulation init_tri = pslg_constrained_delaunay_triangulate(vl, elist);
  int nt = init_tri.numtri();
  vector<Triangle_Label> disposition(nt);
  {
    for (int k = 0; k < nt; ++k) {
      disposition[k] = NOT_YET_SEARCHED;
      for (int l = 0; l < 3; ++l) {
        if (init_tri.tri_vertex(k,l) >= vtxlist.numrows())
          disposition[k] = DELETE;
      }
    }
  }
  search_tri(elist, init_tri, disposition);
  Triangulation result;
  {
    for (int k = 0; k < nt; ++k) {
      if (disposition[k] == KEEP) {
        result.add_triangle(init_tri.tri_vertex(k,0), 
          init_tri.tri_vertex(k,1),
          init_tri.tri_vertex(k,2));
      }
    }
  }
  return result;
}

